package daikon.util;

import daikon.VarInfo;
import daikon.derive.Derivation;
import daikon.derive.binary.SequenceScalarSubsequence;
import daikon.derive.ternary.SequenceScalarArbitrarySubsequence;
import daikon.inv.OutputFormat;

public class ESCJSplitter {

	public static class ESCJArrayVarInfo {
		public ESCJArrayVarInfo(String array_name, String lower_bound,
				String upper_bound) {
			super();
			this.array_name = array_name;
			this.lower_bound = lower_bound;
			this.upper_bound = upper_bound;
		}

		public final String array_name;
		public final String lower_bound;
		public final String upper_bound;
	}

	public static ESCJArrayVarInfo getArrayVarInfo(VarInfo array_var_info) {
		String lower_bound;
		String upper_bound;
		String array_name;

		if (array_var_info.derived != null) {
			Derivation derived = array_var_info.derived;
			if (derived instanceof SequenceScalarSubsequence) {
				SequenceScalarSubsequence sequenceScalarSubsequence = (SequenceScalarSubsequence) derived;
				array_name = sequenceScalarSubsequence.get_array_var()
						.name_using(OutputFormat.JAVA);
				lower_bound = sequenceScalarSubsequence.get_lower_bound()
						.esc_name();
				upper_bound = sequenceScalarSubsequence.get_upper_bound()
						.esc_name();
			} else if (derived instanceof SequenceScalarArbitrarySubsequence) {
				SequenceScalarArbitrarySubsequence sequenceScalarArbitrarySubsequence = (SequenceScalarArbitrarySubsequence) derived;
				array_name = sequenceScalarArbitrarySubsequence.get_array_var()
						.name_using(OutputFormat.JAVA);
				
				VarInfo[] bases = sequenceScalarArbitrarySubsequence.getBases();
				
				
				lower_bound = bases[1].name() + (!sequenceScalarArbitrarySubsequence.left_closed ? "+1" : "");
				upper_bound = bases[2].name() + (!sequenceScalarArbitrarySubsequence.right_closed ? "-1" : "");

			} else {
				throw new ClassCastException("Implement this case! "
						+ derived.getClass().getName());
			}

		} else {
			array_name = array_var_info.name_using(OutputFormat.JAVA);
			lower_bound = "0";
			upper_bound = array_name + ".length-1";
		}

		String search_str = String.format("size(%s[..])",array_name);
		String replace_str = String.format("%s.length",array_name);
		
		lower_bound = lower_bound.replace(search_str, replace_str);
		upper_bound = upper_bound.replace(search_str, replace_str);
		
		ESCJArrayVarInfo retval = new ESCJArrayVarInfo(array_name, lower_bound,
				upper_bound);
		return retval;

	}
}
