#!/usr/bin/python

from gp import *
from ginpink import extension
import nondaemonicpool

import smtplib
from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email import Encoders


def mail(from_name, from_email, to, subject, text, attach=None):
	pass


# Order fixed statically to avoid nondeterministic selection of batches
java3_util3_all_methods = ['Arrays.binarySearch0', 'Arrays.fill1', 'Arrays.fill0', 'Arrays.quicksortPartition', 'Arrays.insertionSort1', 'Arrays.vecswap', 'Arrays.merge0', 'Arrays.insertionSort0', 'Arrays.equals0', 'Arrays.hashCode0', 'Arrays.hashCodeInteger', 'Arrays.fillInteger0', 'Arrays.fillInteger1'] + \
    ['ArrayDeque.removeFirstOccurrence', 'ArrayDeque.contains', 'ArrayDeque.removeLastOccurrence'] + \
    ['ArrayList.indexOf', 'ArrayList.lastIndexOf', 'ArrayList.remove0', 'ArrayList.clear'] + \
    ['Collections.sort', 'Collections.reverse', 'Collections.replaceAll'] + \
    ['Vector.removeAllElements', 'Vector.indexOf1', 'Vector.lastIndexOf1', 'Vector.setSize', 'Vector.removeRange0']
mbc_all_methods = ['GCD.gcd_Euclid_subtraction', 'GCD.gcd_Euclid_division'] + \
                  ['Maximum.max_one_way', 'Maximum.max_two_way'] + \
                  ['Searching.has_sequential_index', 'Searching.has_binary'] + \
                  ['Sorting.partition', 'Sorting.bubble_sort_basic', 'Sorting.bubble_sort_improved', 'Sorting.selection_sort', 'Sorting.insertion_sort', 'Sorting.comb_sort']

loops = {'Arrays': ['binarySearch0', 'fill1', 'fill0', 'quicksortPartition', 'insertionSort1', 'vecswap', 'merge0', 'insertionSort0', 'equals0', 'hashCode0', 'hashCodeInteger', 'fillInteger0', 'fillInteger1'], 
         'ArrayDeque': ['removeFirstOccurrence', 'contains', 'clear', 'removeLastOccurrence'], 
         'ArrayList': ['indexOf', 'lastIndexOf', 'remove0', 'clear', 'batchRemove'], 
         'Collections': ['sort', 'reverse', 'replaceAll', 'checked_rotate1', 'checked_rotate0'], 
         'Vector': ['removeAllElements', 'indexOf1', 'lastIndexOf1', 'setSize', 'copyOf', 'removeRange0'],
         'GCD': ['gcd_Euclid_subtraction', 'gcd_Euclid_division'],
         'Maximum': ['max_one_way', 'max_two_way'],
         'Searching': ['has_sequential_index', 'has_binary'],
         'Sorting': ['partition', 'bubble_sort_basic', 'bubble_sort_improved', 'selection_sort', 'insertion_sort', 'comb_sort']}

basic_mutations = ['mutations(n=1)', 'mutations(n=2)', 'mutations_and_aging(n=2, n_age=2)']
withold = ['add_oldless_post() ; add_old() ; mutations(n=1)']
distance = ["add_distance_expressions() ; add_constant_expression('j') ; mutations(n=3)"]
parametric = ['add_parametric_calls(1) ; mutations(n=2)']
full_mutations = ['mutations(n=3)']
post_pred = ['activate_post_predicates() ; predicate_mutations(n=1)', \
             'activate_post_predicates() ; predicate_mutations_and_aging(n=1, n_age=1, variants=True)', \
             'activate_post_predicates() ; predicate_mutations(n=2)', \
             'activate_post_predicates() ; predicate_mutations_and_aging(n=2, n_age=1)', \
             'activate_post_predicates() ; predicate_mutations(n=3)']
boolean_mutations = ['activate_theory_from_post(n=3) ; predicate_mutations(n=0, more=1)']
theory_pred = ['activate_theory_from_post(n=3) ; predicate_mutations_and_aging(n=0, more=1, n_age=1)', \
               'activate_theory_predicates(n=3) ; predicate_mutations(n=3)', \
               'activate_theory_predicates(n=3) ; predicate_mutations_and_aging(n=3, n_age=2, variants=False)']
discard = ["keep_only('jc') ; add_parametric_calls(1) ; activate_theory_from_post(n=3) ; predicate_mutations_and_aging(n=0, n_age=1, more=1, variants=True)"]

extras = ['activate_post_expressions() ; mutations(n=3)', \
          'add_resultless_post() ; add_distance_expressions() ; mutations_and_aging(n=2, n_age=1)', \
          'add_distance_expressions() ; mutations_and_aging(n=2, n_age=1)', \
          'activate_post_predicates() ; activate_weakening_predicates() ; predicate_mutations(n=1)', \
          'add_parametric_calls(1) ; activate_theory_from_post(n=3) ; predicate_mutations_and_aging(n=0, n_age=1, more=1, variants=True)', \
          'add_distance_expressions() ; activate_theory_predicates(n=3) ; predicate_mutations_and_aging(n=3, n_age=1, variants=False)', \
          'add_distance_expressions() ; activate_post_predicates() ; predicate_mutations_and_aging(n=2, n_age=1, variants=True)']



strategies = [basic_mutations, withold, distance, parametric, full_mutations, \
              post_pred, boolean_mutations, theory_pred, discard, extras]
flat_commands = [y for x in strategies for y in x]

# Order of command sequences
java3_util3_command_order = [1, 8, 10, 11, 2, 14, 4, 15, 12, 6, 17, 16, 3, 5, 7, 13] # , 9]
mbc_command_order = [1, 8, 2, 18, 7, 22, 19, 24, 23] #, 20, 21]

# Timeout for each command sequence, in seconds
timeout = {1: 5, 2: 500, 3: 50, 4: 15, 5: 700, 6: 1100, 7: 190, 8: 5, 9: 30, 10: 15, 11: 30, 12: 700, 13: 700, 14: 330, 15: 700, 16: 400, 17: 700, 18: 3, 19: 400, 20: 550, 21: 3, 22: 550, 23: 3000, 24: 900, \
0: 30}

scalar_command = "activate_weakening_predicates() ; scalar_mutations()"

# Methods for which quantifiers should be expanded
expand_quantifiers = ['Arrays.merge0', 'Arrays.insertionSort0']

# Methods for which triviality check should be nonstrict
nonstrict_simplification = ['Arrays.fill1', 'Arrays.quicksortPartition', 'Arrays.vecswap', 'Arrays.fillInteger1', 'Collections.sort', 'Collections.reverse', 'Collections.replaceAll', 'Collections.checked_rotate1', 'Collections.checked_rotate0', 'Vector.setSize', 'Vector.copyOf', 'Vector.removeRange0'] + \
                           ['Maximum.max_one_way', 'Maximum.max_two_way', 'Searching.has_sequential_index', 'Searching.has_binary']

# Directory where theories info files are located
theory_location = {'Arrays': '../theories/', 
                   'ArrayDeque': '../theories/', 
                   'ArrayList': '../theories/', 
                   'ArraysExplicit': '../theories/', 
                   'Collections': '../theories/', 
                   'Vector': '../theories/',
                   'GCD': '.',
                   'Maximum': '.',
                   'Searching': '.',
                   'Sorting': '.'}

# Number of separate batches (for different processes)
N_batches=7
# Email when done
default_email='caf@inf.ethz.ch'

from_name = 'SCOOP server'
from_email = 'noreply@scoop.inf.ethz.ch'

allmethods = {'java3.util3': java3_util3_all_methods, \
              'mbc': mbc_all_methods}
commandorder = {'java3.util3': java3_util3_command_order, \
                 'mbc': mbc_command_order}


def process_batch(batch_methods, timeout=None, commands=[], email=None, n_batch=-1, extension='', scalar=False):
    start = time.time()
    errors = False
    for cm in batch_methods:
        classname, methodname = cm.split('.')
        if not loops.has_key(classname) or methodname not in loops[classname]:
            print "Error: " + cm + " doesn't define a known class.method name"
            errors = True
            break
        PlainPrinter.ext = property(lambda self: extension)
        po = FilePrinter
        if not scalar:
            strict = cm not in nonstrict_simplification
        else:
            strict = False
        quantifiers = cm in expand_quantifiers
        res = process_with_commands((classname,), theory_location[classname], method_only=methodname, \
                                        printall=True, fullyqualified=True, expandquant=quantifiers, \
                                        PrintOut=po, debug=False, timeout=timeout, special=commands, \
                                    strict=strict)
    stop = time.time()
    return stop - start
    if email is not None:
        if errors:
            text = "There were errors in processing this batch"
        else:
            text = "Completed mutations batch #%d in %f seconds\n\n " +\
                "The batch consisted of the following methods:\n%s" % (n_batch, stop - start, cm.join("\n"))
        mail(from_name, from_email, email, 'Mutations batch #' + int(n_batch) + ' done', text)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-P', '--processes', dest='P', action='store', type=int, default=N_batches, 
                        help='split the work into %(dest)s process invocations (default: %(default)s)')
    parser.add_argument('--batch', dest='batch', action='store', type=int, 
                        help='ignore running processes and execute process batch # %(dest)s (numbering starts from 0)')
    parser.add_argument('-m', dest='method_name', action='store',
                        help='process only method with name %(dest)s, which must be of the form class.method')
    parser.add_argument('--email', dest='address', action='store', default=default_email, 
                        help='when done with the current invocation, send an email to %(dest)s (default: %(default)s)')
    parser.add_argument('--noemail', dest='email', action='store_false', 
                        help='do NOT send an email (overrides default)')
    parser.add_argument('--scalar', dest='scalar', action='store_true', 
                        help='perform a special iteration 0 with scalar invariants')
    parser.add_argument('package', nargs=1, metavar='PKG', action='store', \
                        help='process all known methods of package %(dest)s')
    args = parser.parse_args()

    pkg = args.package[0]
    if not allmethods.has_key(pkg):
        print "I don't know about methods of classes in package: " + pkg
        sys.exit(0)
    else:
        all_methods = allmethods[pkg]
    if args.scalar:
        ext = 'scalar'
        tos = [timeout[0]]
        commands = [[s.strip() for s in scalar_command.split(';')]]
    else:
        ext = 'mutations'
        command_order = commandorder[pkg]
        tos = [timeout[k] for k in command_order]
        commands = [map(lambda s: s.strip(), flat_commands[k - 1].split(';')) for k in command_order]
    if args.method_name is not None:
        sel_methods = [args.method_name]
    else:
        sel_methods = all_methods
    N_methods = len(sel_methods)
    N_procs = min(N_methods, args.P)
    batch_size = max(1, N_methods / N_procs)
    rem = N_methods % N_procs
    if args.method_name is not None:
        batches = [0]
    else:
        batches = range(N_procs)
        # # Determine the current batch by checking how many processes are running
        # cmd = 'ps -u ' + username + ' -C ' + args[0] + ' | wc -l'
        # cur_batch = int(cmd) # first batch is number 0
    if not args.email:
        em = None
    else:
        em = args.address
    pool = nondaemonicpool.NonDaemonicPool(processes=N_procs)
    procs = list()
    last = 0
    for b in batches:
        first = last
        last = first + batch_size
        if rem > 0 and b in batches[-rem:]:
            last += 1
        assert last <= N_methods
        if args.batch is not None and args.batch != b:
            procs.append(None)
            continue
        smet = sel_methods[first:last]
        procs.append(pool.apply_async(process_batch, \
                                          args=(smet,), \
                                          kwds={'timeout': tos, 'commands': commands, \
                                                    'email': em, 'n_batch': b, 'extension': ext, \
                                                    'scalar': args.scalar}))
    for b in batches:
        p = procs[b]
        if p is None:
            continue
        try:
            p.get()
        except:
            print "Error: some problem with the process on batch #" + str(b)
    pool.close()
    pool.join()
