from jmlvisitor import *
from JMLformulasParser import parseString
import JML_ast
import os

location = ''
extension = '.info'


def read_class(classname, location='.', theory=False):
    with open(os.path.join(location, classname + extension)) as f:
        filecontent = f.readlines()
    lines = [l.strip() for l in filecontent]
    in_pkg = False
    in_ints = False
    in_bools = False
    ints = list()
    bools = list()
    pkg = ""
    for cl in lines:
        if cl == '' or cl.startswith('//'):
            continue
        elif cl == '== Package ==':
            in_pkg = True
            in_ints = False
            in_bools = False
        elif cl == '== Integer expressions ==':
            in_ints = True
            in_pkg = False
            in_bools = False
        elif cl == '== Boolean expressions ==':
            in_bools = True
            in_pkg = False
            in_ints = False
        elif in_pkg:
            pkg = cl
        elif in_ints:
            ints.append(cl)
        elif in_bools:
            bools.append(cl)
        else:
            # Unknown content: skipping
            continue
    if not theory:
        cl = PinkType(classname, ints, bools, pkg=pkg)
    else:
        cl = PinkTheory(classname, ints, bools, pkg=pkg)
    return cl


def read_methods(classname, location='.'):
    with open(os.path.join(location, classname + extension)) as f:
        filecontent = f.readlines()
    lines = [l.strip() for l in filecontent]
    ms = list()
    name = ''
    pkg = ''
    in_pkg = False
    in_method = False
    in_static = False
    in_ints = False
    in_bools = False
    in_refs = False
    in_pre = False
    in_post = False
    in_invs = False
    in_cmd = False
    for cl in lines:
        if cl == '' or cl.startswith('//'):
            continue
        elif cl == '== Package ==':
            in_pkg = True
            in_static = False
            in_method = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_pre = False
            in_post = False
            in_invs = False
            in_cmd = False
        elif cl == '== Method ==':
            in_method = True
            in_pkg = False
            in_static = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_pre = False
            in_post = False
            in_invs = False
            in_cmd = False
            if name != '':
                ms += [(name, static, ints, bools, refs, posts, invs, cmds, pres)]
            name = ''
            static = False
            ints = list()
            bools = list()
            refs = {}
            pres = list()
            posts = list()
            invs = list()
            cmds = list()
        elif cl == '== Static ==':
            in_static = True
            in_pkg = False
            in_method = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_pre = False
            in_post = False
            in_invs = False
            in_cmd = False
        elif cl == '== Integer expressions ==':
            in_ints = True
            in_pkg = False
            in_method = False
            in_static = False
            in_bools = False
            in_refs = False
            in_pre = False
            in_post = False
            in_invs = False
            in_cmd = False
        elif cl == '== Boolean expressions ==':
            in_bools = True
            in_pkg = False
            in_method = False
            in_static = False
            in_ints = False
            in_refs = False
            in_pre = False
            in_post = False
            in_invs = False
            in_cmd = False
        elif cl == '== Ref expressions ==':
            in_refs = True
            in_pkg = False
            in_method = False
            in_static = False
            in_ints = False
            in_bools = False
            in_pre = False
            in_post = False
            in_invs = False
            in_cmd = False
        elif cl == '== Preconditions ==':
            in_pre = True
            in_pkg = False
            in_method = False
            in_static = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_post = False
            in_invs = False
            in_cmd = False
        elif cl == '== Postconditions ==':
            in_post = True
            in_pkg = False
            in_method = False
            in_static = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_pre = False
            in_invs = False
            in_cmd = False
        elif cl == '== Invariants ==':
            in_invs = True
            in_pkg = False
            in_method = False
            in_static = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_post = False
            in_cmd = False
        elif cl == '== Commands ==':
            in_cmd = True
            in_pkg = False
            in_method = False
            in_static = False
            in_ints = False
            in_bools = False
            in_refs = False
            in_pre = False
            in_post = False
            in_invs = False
        elif in_pkg:
            pkg = cl
        elif in_method:
            name = cl
        elif in_static:
            static = True if cl == 'Y' else False
        elif in_ints:
            ints.append(cl)
        elif in_bools:
            bools.append(cl)
        elif in_refs:
            key, ids = cl.split(':')
            ids = ids.split(',')
            refs[key] = [i.strip() for i in ids]
        elif in_pre:
            pres.append(cl)
        elif in_post:
            posts.append(cl)
        elif in_invs:
            invs.append(cl)
        elif in_cmd:
            cmds.append([x.strip() for x in cl.split(';')])
        else:
            # Unknown content: skipping
            continue
    # Last method read
    ms += [(name, static, ints, bools, refs, posts, invs, cmds, pres)]
    res = list()
    for m in ms:
        classlist = list()
        for c in m[4].keys() + [classname]:
            classlist.append(read_class(c))
        res.append((PinkMethod(m[0], classname, m[2], m[3], m[4], \
                               classlist, is_static=m[1], commands=m[7], pkg=pkg), m[5], m[6], m[8]))
    return res



def substitute(root, exps, n, subn):
    """Generate all substitutions of expressions `exps' in expression `root',
    combining exactly `n' substitutions in each mutation and using substitutor class
    with name `subn'."""
    if n <= 0:
        res = [root]
    else:
        res = list()
        for e in exps:
            for r in substitute(root, exps, n - 1, subn):
                for s in eval(subn)(e, r):
                    res.append(s)
    return list(set(res))  # Eliminate duplicates


def substitute_all(root, exps, n, subn):
    """Generate all substitutions of integer expressions `exps' in expression `root',
    combining up to `n' substitutions in each mutation and using substitutor class
    with name `subn'."""
    res = list()
    for x in range(n + 1):
        res += substitute(root, exps, x, subn)
    return list(set(res)) # Eliminate duplicates



class PinkType(object):

    # Name of the type
    name = None
    # Is this a class with all static members?
    static = False
    # Expressions of int type, grouped by number of arguments
    ints = None
    # Expressions of boolean type, grouped by number of arguments
    bools = None

    # Full package name (if available)
    package = None

    def __init__(self, class_name, intlist, boollist, pkg=""):
        """Takes a class_name (string), a list intlist of expressions of int type,
        and a list boollist of expressions of boolean type. All such expressions are strings,
        and use the syntax of unqualitied feature calls of the class, with # symbols replacing
        arguments (or array indexes)."""
        self.name = class_name
        self.package = pkg
        if self.static:
            pf = self.name + '.'
        else:
            pf = ''
        self.ints = {}
        for e in intlist:
            n = e.count('#')
            if not self.ints.has_key(n):
                self.ints[n] = list()
            self.ints[n].append(pf + e)
        self.bools = {}
        for e in boollist:
            n = e.count('#')
            if not self.bools.has_key(n):
                self.bools[n] = list()
            self.bools[n].append(pf + e)

    # Int expressions available on the current object type,
    # instantiating up to `n' arguments with the expressions `base'
    # on target reference called `obj'.
    def get_ints(self, obj='this', n=0, base=[]):
        res = list()
        t = obj + '.' if len(obj) > 0 else ''
        for m in range(n + 1):
            if self.ints.has_key(m):
                for e in self.ints[m]:
                    if e.find('$') != -1:
                        new_e = e.replace('$', obj)
                    else:
                        new_e = t + e
                    be = parseString(new_e.replace('#', 'ZZZZZZZZZZZZZZ') + ';')
                    res += [b for b in substitute(be, base, m, 'IntSubstitutions') \
                                if b.text().find('ZZZZZZZZZZZZZZ') == -1]
        return res
                
    # Boolean expressions available on the current object type,
    # instantiating up to `n' arguments with the expressions `base'
    # on target reference called `obj'.
    # If fixedArgs is an optional list of argument expression, to 
    # be used in sequence as arguments for the predicates being instantiated.
    def get_bools(self, obj='this', n=0, base=[], refs=[], fixedArgs=[]):
        rs = list()
        t = obj + '.' if len(obj) > 0 else ''
        for m in range(n + 1):
            if self.bools.has_key(m):
                for e in self.bools[m]:
                    for v in refs:
                        new_e = e.replace('$', v)
                        for a in fixedArgs:
                            new_e = new_e.replace('#', a, 1)
                        new_m = max(0, m - len(fixedArgs))
                        be = parseString(t + new_e.replace('#', 'ZZZZZZZZZZZZZZ') + ';')
                        rs += [b for b in substitute(be, base, new_m, 'IntSubstitutions') \
                               if b.text().find('ZZZZZZZZZZZZZZ') == -1]
        return rs
        
        

class PinkTheory(PinkType):

    def __init__(self, class_name, intlist, boollist, pkg=""):
        self.static = True
        super(PinkTheory, self).__init__(class_name, intlist, boollist, pkg=pkg)



class PinkMethod(object):

    # Dictionary: classname --> PinkType
    classes = None
    # PinkType of enclosing class
    enclosing = None
    # Package name (if available)
    package = None
    # Method name (string)
    name = None
    # Is the method static?
    static_method = False
    # How many parameters to substitute to get feature call expressions?
    arity = 0
    # List of primitive integer _parsed expressions_
    ints = []
    # List of primitive boolean _parsed expressions_
    bools = []
    # Dictionary: classname --> list of ref variable names
    refs = None

    # List of lists of commands to mutate
    commands = None

    def __init__(self, method_name, of_class, intlist, boollist, refdict, classlist, is_static=False, \
                 commands=[[]], pkg=""):
        """Taking a method name (string), of class (string), a list intlist of local expressions
        of int type, a list boollist of local expressions of boolean type, a mapping refdict:
        classname (string) --> refname (string), and a list of PinkType objects giving the context."""
        self.static_method = is_static
        self.name = method_name
        self.package = pkg
        self.ints = [parseString(e + ';') for e in intlist]
        self.bools = [parseString(e + ';') for e in boollist]
        self.refs = refdict
        self.classes = {}
        self.enclosing = None
        for c in classlist:
            self.classes[c.name] = c
            if c.name == of_class:
                self.enclosing = c
        if not is_static and self.enclosing is not None:
            # Add members reachable of this
            self.ints += self.enclosing.get_ints(obj='', n=0, base=list())
        self.commands = commands
        self.arity = 0

    # As available within the method's body
    def get_ints(self):
        res = self.ints
        for c in self.refs.keys():
            if self.classes.has_key(c):
                for r in self.refs[c]:
                    res += self.classes[c].get_ints(obj=r, n=self.arity, base=self.ints)
                    # using only integers as replacements
        return res

    # As available within the method's body
    def get_bools(self):
        res = self.bools
        for c in self.refs.keys():
            if self.classes.has_key(c):
                for r in self.refs[c]:
                    res += self.classes[c].get_ints(obj=r, n=self.arity, base=self.ints)
                    # using only integers as replacements
        return res

    # As available within the method's body, just names
    def get_refs(self):
        return [r for t, rs in self.refs.items() for r in rs]

    def set_arity(self, n):
        if n >= 0:
            self.arity = n
    


class CommonMutator(object):

    # PinkMethod object
    method = None
    # Dictionary: classname --> PinkTheory
    theories = None
    # List of postcondition _parsed expressions_
    posts = None

    _method_ints = []
    _active_preds = []
    _instantiated = False
    _weakening_preds = []
    _saved_ints = None

    resultQueue = None

    class fakeQueue(object):
        def put(self, arg):
            pass

    def __init__(self, method, theories, posts, resultQueue=None):
        self.method = method
        self.theories = {}
        for t in theories:
            self.theories[t.name] = t
        self.posts = [parseString(p) for p in posts]
        self._method_ints = self.method.get_ints()
        self._active_preds = list()
        self._weakening_preds = list()
        if resultQueue is not None:
            self.resultQueue = resultQueue
        else:
            # self.resultQueue = self.fakeQueue()
            pass

    def mutations(self, n=1):
        res = list()
        for p in self.posts:
            res += substitute_all(p, self._method_ints, n, 'IntSubstitutions')
        # self.resultQueue.put(res)
        return res

    def mutations_and_aging(self, n=1, n_age=1):
        bm = self.mutations(n=n)
        aging_plus = parseString('+ 1;')
        aging_minus = parseString('- 1;')
        res = list()
        for m in bm:
            res += substitute_all(m, [aging_plus, aging_minus], n_age, 'AgingSubstitutions')
        # self.resultQueue.put(bm + res)
        return bm + res

    def boolean_mutations(self, n=1):
        unneg = list()
        if not self._instantiated:
            for p in self._active_preds:
                unneg += substitute_all(p, self._method_ints, n, 'IntSubstitutions')
        else:
            unneg = self._active_preds
        neg = list()
        for m in unneg:
            ne = JML_ast.NotExpression()
            ne.item = m
            neg.append(ne)
        muts = unneg + neg
        res = list()
        for p in self.posts:
            res += substitute_all(p, muts, n, 'BooleanSubstitutions')
        # self.resultQueue.put(res)
        return res

    def add_distance_expressions(self, positive=False):
        # pairs of int expressions
        pairs = [(x, y) for x in self._method_ints for y in self._method_ints if y != x]
        for p in pairs:
            if not positive:
                de = JML_ast.SubExpression()
            else:
                de = JML_ast.AddExpression()
            de.left = p[0]
            de.right = p[1]
            self._method_ints.append(de)
        self._method_ints = list(set(self._method_ints))

    def add_parametric_calls(self, n):
        if n >= 0:
            old_arity = self.method.arity
            self.method.set_arity(n)
        self._method_ints = self.method.get_ints()
        self.method.set_arity(old_arity)

    def add_constant_expression(self, exp):
        e = parseString(exp + ';')
        self._method_ints.append(e)

    def keep_only(self, exp):
        # self._saved_ints = copy.deepcopy(self.method.ints)
        self.method.ints = list()
        ee = exp.split(',')
        for e in ee:
            p = parseString(e.strip() + ';')
            self.method.ints.append(p)

    def add_oldless_post(self):
        oldless = list()
        for p in self.posts:
            new_p = parseString(''.join(p.text().split('\\old')) + ';')
            if new_p.text() != p.text():
                oldless.append(new_p)
        self.posts += oldless

    def add_old(self, n=1):
        old = list()
        for p in self.posts:
            old += substitute_all(p, [p], n, 'OldAdder')
        self.posts += old

    def add_resultless_post(self):
        resultless = list()
        for p in self.posts:
            for s in ResultRemover(p):
                resultless.append(s)
        self.posts += resultless

    def activate_theory_predicates(self, n=3):
        ts = list()
        for tn in self.theories.keys():
            for p in self.posts:
                if p.text().find(tn + '.') != -1:
                    ts.append(self.theories[tn])
        for t in set(ts):
            self._active_preds += t.get_bools(n=n, obj='', base=self._method_ints, \
                                              refs=self.method.get_refs())
        self._instantiated = True

    def suggest(self, pred, more=1):
        """Activate theory predicates using the arguments of predicates in `pred' as initial instantiations,
        with up to `more' additional substitutions (excluding refs)"""
        ts = list()
        for tn in self.theories.keys():
            for p in self.posts:
                if p.text().find(tn + '.') != -1:
                    ts.append(self.theories[tn])
        suggestions = list()
        for p in PredicateExtractor(parseString(pred + ';')):
            if len(p.args) > 1:
                # skip first argument assuming it's a reference
                suggestions.append([a.text() for a in p.args[1:]])
        for sug in suggestions:
            for t in ts:
                self._active_preds += t.get_bools(n=len(sug)+more, obj='', base=self._method_ints, \
                                                      refs=self.method.get_refs(), fixedArgs=sug)
        self._instantiated = True

    def activate_theory_from_post(self, n=3):
        ps = list()  # Predicates extracted from post
        for p in self.posts:
            for pred in PredicateExtractor(p):
               ps.append(pred)
        arguments = list()
        for p in ps:
            # Shortcut: the first argument is assumed to be a ref, and is skipped
            arguments.append([a.text() for a in p.args[1:]])
        ts = list()
        for tn in self.theories.keys():
            for p in self.posts:
                if p.text().find(tn + '.') != -1:
                    ts.append(self.theories[tn])
        for fargs in arguments:
            for t in set(ts):
                self._active_preds += t.get_bools(n=n, obj='', base=self._method_ints, \
                                                      refs=self.method.get_refs(), fixedArgs=fargs)
        self._instantiated = True

    def activate_post_expressions(self):
        ps = list()
        for p in self.posts:
            for pred in PredicateExtractor(p):   # Predicate extractor actually extracts any method call
               ps.append(pred)
        self._method_ints += list(set(ps))

    def activate_post_predicates(self):
        ps = list()
        for p in self.posts:
            for pred in PredicateExtractor(p):
               ps.append(pred)
        self._active_preds += list(set(ps))

    def activate_weakening_predicates(self, fromInit=False):
        if fromInit:
            pass
        else:
            # pairs of int expressions
            pairs = [(x, y) for x in self._method_ints for y in self._method_ints if y != x]
            for p in pairs:
                le = JML_ast.LessExpression()
                leq = JML_ast.LessEqualExpression()
                # gr = JML_ast.GreaterExpression()
                # greq = JML_ast.GreaterEqualExpression()
                # eq = JML_ast.EqualsExpression()
                # neq = JML_ast.NEqualsExpression()
                es = [le, leq] # , eq, neq]
                for e in es:
                    e.left = p[0]
                    e.right = p[1]
                self._weakening_preds += es

    def predicate_mutations(self, n=1, more=0, variants=True):
        unneg = list()
        if not self._instantiated:
            for p in self._active_preds:
                unneg += substitute_all(p, self._method_ints, n, 'IntSubstitutions')
        else:
            for p in self._active_preds:
                unneg += substitute_all(p, self._method_ints, more, 'IntSubstitutions')
        if not variants:
            return unneg
        neg = list()
        for m in unneg:
            ne = JML_ast.NotExpression()
            ne.item = m
            neg.append(ne)
        muts = unneg + neg
        res = list()
        bools = self._weakening_preds + self.method.get_bools()
        for m in muts:
            for b in bools:
                we = JML_ast.ImpliesExpression()
                we.left = b
                we.right = m
                res.append(we)
                n_we = JML_ast.ImpliesExpression()
                n_we.left = JML_ast.NotExpression()
                n_we.left.item = b
                n_we.right = m
                res.append(n_we)
        # self.resultQueue.put(muts + res)
        return muts + res

    def predicate_mutations_and_aging(self, n=1, more=0, n_age=1, variants=False):
        # Removing duplicates
        bm = list(set(self.predicate_mutations(n=n, more=more, variants=variants)))
        aging_plus = parseString('+ 1;')
        aging_minus = parseString('- 1;')
        res = list()
        for m in bm:
            res += substitute_all(m, [aging_plus, aging_minus], n_age, 'AgingSubstitutions')
        # self.resultQueue.put(bm + res)
        return bm + res

    def scalar_mutations(self):
        bools = self._weakening_preds
        for b in self.method.get_bools():
            bools.append(b)
            negb = JML_ast.NotExpression()
            negb.item = b
            bools.append(negb)
        pairs = [(x, y) for x in self._method_ints for y in self._method_ints if y != x]
        res = list()
        for b in bools:
            for l, r in pairs:
                le = JML_ast.LessExpression()
                le.left = l
                le.right = r
                impl_le = JML_ast.ImpliesExpression()
                impl_le.left = b
                impl_le.right = le
                res.append(impl_le)
                leq = JML_ast.LessEqualExpression()
                leq.left = l
                leq.right = r
                impl_leq = JML_ast.ImpliesExpression()
                impl_leq.left = b
                impl_leq.right = leq
                res.append(impl_leq)
        return res
