#!/usr/bin/python

from ginpink import *
import multiprocessing
import sys
import time
import argparse
import jml2exec
import os, stat
import pickle
import logging

# def timeout(func, args=(), kwargs={}, timeout_duration=300, default=None, resultQueue=None):
#     """Execute func for up to timeout_duration seconds. Returns default if func times out,
#     otherwise a pair (running time, result).
#     """
#     import multiprocessing, time, Queue
#     it = multiprocessing.Process(target=func, args=args, kwargs=kwargs)
#     start = time.time()
#     it.start()
#     it.join(timeout_duration)
#     end = time.time()
#     if it.is_alive():
#         it.terminate()
#         it.join()
#     try:
#         res = resultQueue.get(block=False)
#     except Queue.Empty:
#         res = None
#     return (end - start, res)


def get_methods_and_theories(appclasses, theoryclasses, theorylocation):
    methods = []
    theories = []
    for c in appclasses:
        methods += read_methods(c)
    for t in theoryclasses:
        theories.append(read_class(t, theory=True, location=theorylocation))
    return (methods, theories)


def how_many_found(muts, invs):
    ms = [x.text().replace('+ -', '- ') for x in muts]
    iv = [x.text() for x in [parseString(i) for i in invs]]
    return set(ms) & set(iv)


def string_args(kwargs):
    res = []
    for key, value in kwargs.items():
        res.append(str(key) + '=' + str(value))
    return '(' + ', '.join(res) + ')'

def process(class_names):
    # resultQueue = multiprocessing.Queue()

    methods, theories = get_methods_and_theories(class_names, theoryclasses)

    for method, postconditions, invariants, preconditions in methods:
        print "*********************************************************"
        print "METHOD:", method.enclosing.name + '.' + method.name
        unique_invariants = set(invariants)
        for n_pre in range(len(preprocessings) + 1):
            for n_proc in range(len(processings)):
                mutator = CommonMutator(method, theories, postconditions, resultQueue=None)
                options = ''
                for p in preprocessings[:n_pre]:
                    options += p + ' + '
                    eval('mutator' + '.' + p)
                print "With options", options, "and command", processings[n_proc]
                start = time.time()
                print "Starting at:", time.strftime("%H:%M %S", time.localtime(start))
                # mutations = timeout(eval('mutator' + '.' + processings[n_proc][0]), \
                #                     kwargs=processings[n_proc][1], resultQueue=resultQueue,
                #                     timeout_duration=60)
                mutations = eval('mutator' + '.' + processings[n_proc][0] + \
                                 string_args(processings[n_proc][1]))
                stop = time.time()
                mutations = (stop - start, mutations)
                assert len(mutations) == 2
                if mutations[1] is None:
                    print "-----> TIMEOUT"
                    print ""
                    break
                else:
                    shared = how_many_found(mutations[1], unique_invariants)
                    print "-----> Generated", len(mutations[1]), \
                          "Found", len(shared), "invariants (syntactically) in", mutations[0]
                    for s in shared:
                        print s
                    print ""


class MutationWrapper(object):

    methodname = ""

    # List of sets of mutations generated in each iteration
    mutations = []

    # Set of all generated mutations
    all_mutations = {}

    # List of commands in each iteration
    commands = []

    # List of invariants found in each iteration
    found_invariants = []

    # Total number of invariants found so far
    how_many_found = 0

    cur_iteration = 0

    # List of number of generated
    generated = []

    # Set of all invariants (to know when to stop), in normalized text form
    invariants = {}

    # Time spent in each iteration
    times = []

    # List of number of mutations generated, or negative number to denote timeout expired
    outcome = []
    
    def __init__(self, methodname, invariants):
        self.methodname = methodname
        self.mutations = list()
        self.commands = list()
        self.all_mutations = set()
        self.found_invariants = list()
        self.cur_iteration = 0
        self.outcome = list()
        # Reparsing, to normalize text representation
        self.invariants = set([x.text() for x in [parseString(i + ';') for i in invariants]])
        self.how_many_found = 0
        self.generated = list()
        self.times = list()

    def _nit(self, iteration=None):
        if iteration is None:
            iteration = self.cur_iteration
        return self.methodname + '.' + 'iteration_' + str(iteration) + '.' + 'dat'

    def new_iteration(self, commands):
        """Initialize new iteration"""
        self.cur_iteration += 1
        self.mutations.append(commands)
        self.commands.append(None)
        self.outcome.append(None)
        self.found_invariants.append(None)
        self.times.append(None)
        self.generated.append(None)

    def new_mutations(self, mutations, generated):
        """Add mutations to the set of mutations generated in the current iteration"""
        news = set(mutations) - self.all_mutations
        self.all_mutations = self.all_mutations | news
        self.mutations[self.cur_iteration - 1] = news
        self.outcome[self.cur_iteration - 1] = len(news)
        self.found_invariants[self.cur_iteration - 1] = set([n for n in news if n[1].text() in self.invariants])
        self.how_many_found += len(self.found_invariants[self.cur_iteration - 1])
        self.generated[self.cur_iteration - 1] = generated
        return news

    def new_time(self, time):
        self.times[self.cur_iteration - 1] = time

    def timeout(self, timeout=0):
        """Set outcome to timeout for current iteration"""
        news = set()
        self.mutations[self.cur_iteration - 1] = news
        self.outcome[self.cur_iteration - 1] = -timeout
        self.found_invariants[self.cur_iteration - 1] = news
        self.generated[self.cur_iteration - 1] = 0

    def found(self):
        """List of invariants found in the current iteration"""
        return self.found_invariants[self.cur_iteration - 1]

    def done(self):
        """Have all invariants been found?"""
        return self.found_invariants == len(self.invariants)

    def dump(self):
        """Dump the content to file"""
        pickle.dump(self, open(self._nit(), 'wb'))

    def __str__(self):
        res = ''
        for i in range(self.cur_iteration):
            header = """*********************************************************
METHOD: %s
Iteration # %d
Processing with: %s""" % (self.methodname, i + 1, self.commands[i])
            if self.outcome[i] < 0:
                summary = "No mutations generated within a timeout of " + str(-self.outcome[i]) + " seconds"
            elif self.outcome[i] == 0:
                summary = "No mutations returned"
            else:
                summary =  "-----> Generated %d of which %d new and valid. " % (self.generated[i], self.outcome[i])
                summary += "Found %d new invariants (syntactically) out of %d" % \
                           (len(self.found_invariants[i]), len(self.invariants))
                summary += " in %f" % self.times[i]
                invs = ""
                for s in self.found_invariants[i]:
                    invs += s[1].text() + '\n'
            res += header + '\n' + summary + '\n' + invs
        return res



def process_commandsequence(command_sequence, mutator, debug):
    for cmd in command_sequence:
        if debug:
            logging.debug("Command: " + cmd)
        # Only the last command should return a list
        mutations = eval('mutator' + '.' + cmd)
        if debug:
            logging.debug("Command " + cmd + " returned with " + \
                          str(0 if mutations is None else len(mutations)) + " mutations")
    if debug:
        logging.debug("Command sequence terminated")
    return mutations



def executable_forms(mutations, expandquant, debug, strict):
    exec_mutations = list()
    for x in mutations:
        if debug:
            logging.debug("Computing executable form of: " + x.text())
        jex = jml2exec.ExecutableJML(x, parsed=True)
        jex.make_executable()
        # If conversion to executable succeeded
        if jex.is_executable():
            if debug:
                logging.debug("Executable form: " + jex.node.text())
            if len(jex.errors) == 0:
                logging.debug("Executable form produced without errors")
            else:
                logging.debug("Executable form produced with errors:")
                en = 0
                for err in jex.errors:
                    logging.debug("Error #" + str(en) + " " + err)
                    en += 1
            exec_form = jex.node
            jml_form = x
            if debug:
                logging.debug("Checking triviality of: " + exec_form.text())
            tr = jml2exec.RemoveTrivial(exec_form, strict=strict)
            if tr.trivial:
                if debug:
                    logging.debug("Trivial expression skipped")
                continue
            elif debug:
                logging.debug("Expression non trivial")
            if expandquant:
                if debug:
                    logging.debug("Computing quantified form of: " + x.text())
                # Expand quantification in JML form
                jex = jml2exec.ExecutableJML(copy.deepcopy(x), parsed=True)
                jex.make_quantification()
                if debug:
                    logging.debug("Quantified form: " + jex.node.text())
                    if len(jex.errors) == 0:
                        logging.debug("Quantified form produced without errors")
                    else:
                        logging.debug("Quantified form produced with errors:")
                        en = 0
                        for err in jex.errors:
                            logging.debug("Error #" + str(en) + " " + err)
                            en += 0
                jml_form = jex.node
            exec_mutations.append((exec_form, jml_form))
        else:
            # Not executable
            if debug:
                logging.debug("Could not produce executable form of: " + x.text())
            continue
    return exec_mutations
        

   
def process_method(method, postconditions, invariants, theories, theoryFullNames, \
                   PrintOut, expandquant, fullyqualified, printall, only, timeout, serialize, debug, \
                   skip=False, special=None, preconditions=None, strict=True):
    m_name = method.name
    if fullyqualified:
        pkg = method.package + '.' if method.package != '' else ''
        m_name = pkg + method.enclosing.name + '.' + m_name
    if debug:
        logging.debug("Processing method: " + m_name)
    unique_invariants = set(invariants)
    iteration = 0
    all_mutations = set()
    mw = MutationWrapper(m_name, invariants)
    po_open = False
    if special is not None:
        method.commands = special
    for command_sequence in method.commands:
        iteration += 1
        if only > 0 and iteration > only:
            # Do up to only iteration
            break
        strcmd = ' ; '.join(command_sequence)
        mw.new_iteration(strcmd)
        if serialize:
            # Trying to find a previous serialized state
            fn = mw._nit()
            if os.path.exists(fn):
                mw = pickle.load(open(fn, 'rb'))
                outoftime=False
                continue
        if skip:
            mw.timeout()
            mw.new_time(-666)
            continue
        if debug:
            logging.debug("Iteration #" + str(iteration))
            logging.debug("Command sequence: " + strcmd)
        po = PrintOut(m_name, iteration)
        po_open = True
        po.printout( "*********************************************************" )
        po.printout( "METHOD: " + m_name )
        method_bak = copy.deepcopy(method)
        mutator = CommonMutator(method_bak, theories, postconditions, resultQueue=None)
        po.printout( "Iteration # " + str(iteration) )
        po.printout( "Processing with: " + strcmd )
        start = time.time()
        # print "Starting at:", time.strftime("%H:%M %S", time.localtime(start))
        if debug:
            logging.debug("Starting at: " + str(time.strftime("%H:%M %S", time.localtime(start))))
        if type(timeout) is int and timeout > 0:
            tov = timeout
        elif type(timeout) is list and iteration <= len(timeout):
            tov = timeout[iteration - 1]
        else:
            tov = None
        if tov is not None:
            pool = multiprocessing.Pool(processes=1)
            presult = pool.apply_async(process_commandsequence, args=(command_sequence, mutator, debug))
            try:
                outoftime = False
                mutations = presult.get(timeout=tov)
            except multiprocessing.TimeoutError:
                pool.terminate()
                mutations = None
                outoftime = True
            pool.close()
            pool.join()
        else:
            # No timeout set
            outoftime = False
            mutations = process_commandsequence(command_sequence, mutator, debug)
        if debug:
            logging.debug("Command sequence terminated at: " + str(time.strftime("%H:%M %S", time.localtime(start))))
            if outoftime:
                logging.debug("Command sequence timed out")
            else:
                logging.debug("No timeout detected")
        if preconditions is not None and iteration == 1:
            mutations = preconditions + mutations
        if mutations is not None:
            generated = len(mutations)
            if debug:
                logging.debug("Before filtering: " + str(generated) + " mutations")
            # Remove syntactic duplicates
            mutations = set(mutations)
            if debug:
                logging.debug("After removing duplicates: " + str(len(mutations)) + " mutations")
            exec_mutations = executable_forms(mutations, expandquant, debug, strict=strict)
            # Keep new mutations (not generated in previous iterations)
            # mutations = set(exec_mutations) - all_mutations
            mutations = mw.new_mutations(exec_mutations, generated)
            if debug:
                logging.debug("New mutations retained: " + str(len(mutations)))
        else:
            if debug:
                logging.debug("Last command sequence terminated with None")
        stop = time.time()
        mw.new_time(stop - start)
        if mutations is None:
            if outoftime:
                mw.timeout(timeout=tov)
                po.printout("No mutations generated within a timeout of " + str(tov) + " seconds")
            else:
                mw.new_mutations(set())
                po.printout( "No mutations returned" )
        else:
            # Union
            all_mutations = mw.all_mutations
            shared = set([f[1].text() for f in mw.found()])
            # shared = how_many_found([x[1] for x in mutations], unique_invariants)
            po.printout( "-----> Generated " + str(generated) + \
                         " of which " + str(len(mutations)) + " new and valid. " + \
                         "Found " + str(len(shared)) + " new invariants (syntactically)" + \
                         " out of " + str(len(unique_invariants)) + \
                         " in " + str(stop - start) )
            for s in shared:
                po.printout( s )
            po.printout( "" )
            if printall:
                po.printout( "-----> All mutations follow" )
                for s in mutations:
                    if fullyqualified:
                        if debug:
                            logging.debug("Replacing names for fully qualified names")
                        rn = jml2exec.ReplaceNames(s[0], theoryFullNames)
                        exec_text = rn.result.text()
                        if debug:
                            logging.debug("After name replacement: " + exec_text)
                            if len(rn.errors) == 0:
                                logging.debug("Name replacement produced without errors")
                            else:
                                logging.debug("Name replacement produced with errors:")
                                en = 0
                                for err in rn.errors:
                                    logging.debug("Error #" + str(en) + " " + err)
                                    en += 0
                    else:
                        exec_text = s[0].text()
                    po.printout( exec_text + '   <--exec|JML-->  ' + s[1].text() )
        if serialize:
            mw.dump()
    if po_open:
        po.done()
    return outoftime    
    

def process_with_commands(class_names, theorylocation, method_only=None, printall=False, \
                          fullyqualified=False, expandquant=False, PrintOut=None, \
                          debug=False, timeout=0, only=0, serialize=False, \
                          skip=False, special=None, strict=True):
    import os
    files = [f for f in os.listdir(theorylocation) if os.path.isfile(os.path.join(theorylocation, f))]
    theoryclasses = [f[:-len(extension)] for f in files if f.startswith('T') and f.endswith(extension)]
    
    methods, theories = get_methods_and_theories(class_names, theoryclasses, theorylocation)
    theoryFullNames = {}
    if fullyqualified:
        for th in theories:
            if th.package != '':
                theoryFullNames[th.name] = th.package + '.' + th.name
            else:
                theoryFullNames[th.name] = th.name
    if method_only is None:
        keep_methods = methods
    else:
        keep_methods = []
        for m in methods:
            if m[0].name == method_only:
                keep_methods.append(m)
    for method, postconditions, invariants, preconditions in keep_methods:
        res = process_method(method, postconditions, invariants, theories, theoryFullNames, \
                             PrintOut, expandquant, fullyqualified, printall, only, timeout, serialize, debug, \
                             skip=skip, special=special, preconditions=[parseString(p) for p in preconditions], \
                             strict=strict)
    if len(keep_methods) > 0:
        return res
    else:
        if debug:
            logging.debug("Did not find a method: " + method_only)
        return None



class PlainPrinter(object):

    name = None
    iteration = None
    ext = '.txt'

    def __init__(self, name, iteration):
        self.name = name
        self.iteration = iteration
    
    def printout(self, s):
        print s

    def done(self):
        pass


class FilePrinter(PlainPrinter):

    def __init__(self, name, iteration):
        super(FilePrinter, self).__init__(name, iteration)
        v = 0
        done = False
        while not done:
            v += 1
            vt = '.' + "v" + str(v) if v > 1 else ""
            filename = self.name + "." + "iteration_" + str(self.iteration) + "." + self.ext + vt
            if os.path.exists(filename):
                continue
            else:
                self.file = open(filename, 'w')
                done = True

    def printout(self, s):
        self.file.write(s + '\n')
        
    def done(self):
        self.file.close()



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate mutations of postconditions.')
    parser.add_argument('-T', dest='theory_directory', action='store', default='.', \
                        help='directory with theory info files (default: %(default)s)')
    parser.add_argument('-m', dest='method_name', action='store',
                        help='process only method with name %(dest)s')
    parser.add_argument('--tofile', dest='ext', action='store',
                        help='write results to separate files with extension %(dest)s')
    parser.add_argument('--command', dest='cmd', action='store',
                        help='mutate using only command sequence %(dest)s')
    parser.add_argument('--fullnames', dest='fullyqualified', action='store_true',
                        help='use fully qualified method names (default: %(default)s)')
    parser.add_argument('-q', '--explicitquantifiers', dest='explicitquant', action='store_true',
                        help='expand quantifiers in JML form in all mutations (default: %(default)s)')
    parser.add_argument('-a', '--all', dest='print_all', action='store_true',
                        help='print all mutations (not just those that match the invariants)')
    parser.add_argument('--serialize', dest='serialize', action='store_true', 
                        help='serialize mutations, and start from deserialized predecessors if they are available')
    parser.add_argument('--only', dest='only', action='store', type=int, 
                        help='execute at most %(dest)s mutation sequences per method')
    parser.add_argument('--to', '--timeout', dest='timeout', action='store', type=int, 
                        help='terminate a mutation sequence after %(dest)s seconds')
    parser.add_argument('--nonstrict', dest='strict', action='store_false', 
                        help='keep mutations without theory predicates')
    parser.add_argument('--debug', dest='debug_file', action='store',
                        help='output debug information to file %(dest)s (overwritten)')
    parser.add_argument('classes', nargs='+', metavar='C', action='append', \
                        help='a class (without extension) to be processed; file %(dest)s' + \
                            extension + ' must exist')
    args = parser.parse_args()
    if args.ext is not None:
        # Changing class attribute value!
        PlainPrinter.ext = property(lambda self: args.ext)
        po = FilePrinter
    else:
        po = PlainPrinter
    if args.debug_file is not None:
        debug=True
        logging.basicConfig(filename=args.debug_file, filemode='w', level=logging.DEBUG)
    else:
        debug=False
    to = 0 if args.timeout is None else args.timeout
    nn = 0 if args.only is None else args.only
    special=None
    if args.cmd is not None:
        special = [[c.strip() for c in args.cmd.split(';')]]
    process_with_commands(args.classes[0], args.theory_directory, \
                          method_only=args.method_name, printall=args.print_all, \
                              fullyqualified=args.fullyqualified, \
                              expandquant=args.explicitquant, \
                              PrintOut=po, debug=debug, timeout=to, only=nn, serialize=args.serialize, \
                              strict=args.strict, special=special)
