#!/usr/bin/python

from JML_ast import *
from JMLformulasParser import parseString
from jmlvisitor import Substitutions
import argparse
import copy
import sys


methods_with_Integer_arrays = ['java3.util3.Arrays.fillInteger0', 'java3.util3.Arrays.fillInteger1', 'java3.util3.Arrays.hashCodeInteger', 'java3.util3.Arrays.insertionSort0', 'java3.util3.Arrays.merge0', 'java3.util3.Vector.indexOf1', 'java3.util3.Vector.lastIndexOf1']


class FixIntValue(Substitutions):

    errors = []

    def __init__(self, nexp, root):
        super(FixIntValue, self).__init__(nexp, root)
        self.failed = False

    def fix_intValue(self, targett, callt, args):
        known_calls = ['less', 'lesseq', 'grt', 'grteq', 'eq', 'neq']
        res = [a.text() for a in args]
        if targett == 'TArrays' and len(args) == 4 and callt in known_calls:
            res[3] = res[3].replace('.intValue()', '')
            args[3] = parseString(res[3] + ';')
            assert args[3] is not None

    def MethodCall_visit(self, node):
        super(FixIntValue, self).MethodCall_visit(node)
        if node.target is not None:
            targett = node.target.text()
            if node.call is not None:
                callt = '.'.join([c.text() for c in node.call])
                self.fix_intValue(targett, callt, node.args)


class PredicateExpansion(Substitutions):

    errors = []

    Integer_arrays = None
    preconditions = None

    def __init__(self, nexp, root):
        super(PredicateExpansion, self).__init__(nexp, root)
        self.errors = list()
        self.Integer_arrays = True
        self.preconditions = False
        self.failed = False
        # self.fixIntValues = False

    def use_int_arrays(self):
        self.Integer_arrays = False

    def use_preconditions(self):
        self.preconditions = True

    # def use_fixIntValues(self):
    #     self.fixIntValues = True

    precondition_predicates = { \
        "TArrays": { \
            "sorted": lambda a, fromIndex, toIndex: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "in": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "less": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "lesseq": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "grt": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "grteq": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "eq": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "eqnull": lambda a, fromIndex, toIndex: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "neq": lambda a, fromIndex, toIndex, key: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
            "nonnull": lambda a, fromIndex, toIndex: \
                "(" + a + " != null" + ")" + \
                " && " + \
                "(0 <= " + fromIndex + " && " + fromIndex + " <= " + toIndex + " && " + toIndex + " <= " + a + ".length)", \
                } \
            }

    precondition_Integer_predicates = { \
        "TArrays": { \
            "sorted": lambda a, fromIndex, toIndex: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)", \
            "less": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)" + " && (" + key + " != null)", \
            "lesseq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)" + " && (" + key + " != null)", \
            "grt": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)" + " && (" + key + " != null)", \
            "grteq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)" + " && (" + key + " != null)", \
            "eq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)" + " && (" + key + " != null)", \
            "neq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null)" + " && (" + key + " != null)", \
                } \
            }

    pred2post = { \
        "TArrays": { \
            "sorted": lambda a, fromIndex, toIndex: \
                "(\\forall int jj1, jj2; " + fromIndex + " <= jj1 && jj1 < jj2 && jj2 < " + toIndex + \
                " ==> " + a + "[jj1].intValue() <= " + a + "[jj2].intValue());", \
            "in": lambda a, fromIndex, toIndex, key: \
                "(\\exists int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " && " + a + "[jj].intValue() == " + key + ".intValue());", \
            "less": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj].intValue() < " + key + ".intValue());", \
            "lesseq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj].intValue() <= " + key + ".intValue());", \
            "grt": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj].intValue() > " + key + ".intValue());", \
            "grteq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj].intValue() >= " + key + ".intValue());", \
            "eq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj].intValue() == " + key + ".intValue());", \
            "eqnull": lambda a, fromIndex, toIndex: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] == null);", \
            "neq": lambda a, fromIndex, toIndex, key: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj].intValue() != " + a + key + ".intValue());", \
            "nonnull": lambda a, fromIndex, toIndex: \
                "(\\forall int jj; " + fromIndex + " <= jj && jj < " + toIndex + \
                " ==> " + a + "[jj] != null);" \
                } \
            }

    def ConditionalExpression_visit(self, node):
        super(PredicateExpansion, self).ConditionalExpression_visit(node)
        if isinstance(node.condition, MethodCall) and node.condition._sub is not None:
            node.condition = node.condition._sub
        if isinstance(node.then_exp, MethodCall) and node.then_exp._sub is not None:
            node.then_exp = node.then_exp._sub
        if isinstance(node.else_exp, MethodCall) and node.else_exp._sub is not None:
            node.else_exp = node.else_exp._sub

    def BinaryExpression_visit(self, node):
        super(PredicateExpansion, self).BinaryExpression_visit(node)
        if isinstance(node.left, MethodCall) and node.left._sub is not None:
            node.left = node.left._sub
        if isinstance(node.right, MethodCall) and node.right._sub is not None:
            node.right = node.right._sub

    def UnaryExpression_visit(self, node):
        super(PredicateExpansion, self).UnaryExpression_visit(node)
        if isinstance(node.item, MethodCall) and node.item._sub is not None:
            node.item = node.item._sub

    # def fix_intValue(self, targett, callt, args):
    #     known_calls = ['less', 'lesseq', 'grt', 'grteq', 'eq', 'neq']
    #     res = [a.text() for a in args]
    #     if targett == 'TArrays' and len(args) == 4 and callt in known_calls:
    #         res[3] = res[3].replace('.intValue()', '')
    #         args[3] = parseString(res[3] + ';')
    #         assert args[3] is not None
    #     return res

    def MethodCall_visit(self, node):
        super(PredicateExpansion, self).MethodCall_visit(node)
        node._sub = None
        if node.target is not None:
            targett = node.target.text()
            if self.pred2post.has_key(targett):
                tr = self.pred2post[targett]
                if node.call is not None:
                    callt = '.'.join([c.text() for c in node.call])
                    if tr.has_key(callt) and tr[callt].func_code.co_argcount == len(node.args):
                        # if self.fixIntValues and self.Integer_arrays:
                        #     argtext = self.fix_intValue(targett, callt, node.args)
                        # else:
                        argtext = [a.text() for a in node.args]
                        res = tr[callt](*argtext)
                        if not self.Integer_arrays:
                            res = res.replace('.intValue()', '')
                        if self.preconditions:
                            pre = ""
                            if self.precondition_predicates.has_key(targett) \
                                   and self.precondition_predicates[targett].has_key(callt):
                                pre = self.precondition_predicates[targett][callt](*argtext)
                            if self.Integer_arrays and \
                               self.precondition_Integer_predicates.has_key(targett) \
                               and self.precondition_Integer_predicates[targett].has_key(callt):
                                pre += " && " + self.precondition_Integer_predicates[targett][callt](*argtext)
                            if pre != "":
                                res = pre + " && " + res
                        node._sub = parseString(res + ';')
                        if node._sub is None:
                            self.failed = True


triviality = { \
    "TArrays": { \
        "grt": lambda args: args[1] == args[2] or args[1] == args[3] or args[2] == args[3], \
            "grteq": lambda args: args[1] == args[2] or args[1] == args[3] or args[2] == args[3], \
            "less": lambda args: args[1] == args[2] or args[1] == args[3] or args[2] == args[3], \
            "lesseq": lambda args: args[1] == args[2] or args[1] == args[3] or args[2] == args[3], \
            "sorted": lambda args: args[1] == args[2], \
            }, \
    "TLists": { \
        "eq": lambda args: args[1] == args[2] or args[1] == args[3] or args[2] == args[3], \
            "neq": lambda args: args[1] == args[2] or args[1] == args[3] or args[2] == args[3], \
            "sorted": lambda args: args[1] == args[2], \
            "deqWithin": lambda args: args[1] == args[2] or args[2] == args[3] or args[1] == args[3], \
            "deqEq": lambda args: args[1] == args[2] or args[2] == args[3] or args[1] == args[3], \
            "deqHas": lambda args: args[1] == args[2] or args[2] == args[3] or args[1] == args[3], \
            "deqHasNot": lambda args: args[1] == args[2] or args[2] == args[3] or args[1] == args[3], \
            }, \
    "QTArrays": { \
        "QEq": lambda args: args[0] == args[1] or (args[2] == args[5] and args[3] == args[6] and args[4] == args[7]), \
            "QEqK": lambda args: args[0] == args[1], \
            }, \
    "QTLists": { \
        "QEq": lambda args: args[0] == args[1] or \
            (len(args) == 10 and args[2] == args[6] and args[3] == args[7] and args[4] == args[8] and args[5] == args[9]), \
            }, \
    "TSorting": {}, \
    "TArithmetic": {}, \
        }


def is_trivial(mcall):
    if len(mcall.call) != 1 or mcall.args is None:
        return False
    call = mcall.call[0].text()
    target = mcall.target.text()
    args = [a.text() for a in mcall.args]
    return triviality.has_key(target) and \
        triviality[target].has_key(call) and \
        triviality[target][call](args)

def is_theorypred(mcall):
    target = mcall.target.text()
    return triviality.has_key(target)



class RemoveTrivial(Substitutions):
    """Sets self.trivial to True iff all method calls are not trivial, and there is at least one theory predicate"""

    trivial = None

    theory_preds = 0

    _in_pred = None

    def __init__(self, nexp, strict=True):
        super(RemoveTrivial, self).__init__(nexp, None)
        # Marked by false if a theory predicate is found
        self.trivial = False
        self._in_pred = False
        if strict:
            self.theory_preds = 0
        else:
            self.theory_preds = 1
        self.nexp.accept(self)
        if self.theory_preds == 0:
            self.trivial = True

    def MethodCall_visit(self, node):
        if is_theorypred(node):
            self.theory_preds += 1
            istp = True
        else:
            istp = False
        oldip = self._in_pred
        if self._in_pred:
            if istp:
                # No theory predicates nested inside other predicates
                self.trivial = True
                return
        else:
            self._in_pred = True
        super(RemoveTrivial, self).MethodCall_visit(node)
        self._in_pred = oldip
        if not self.trivial:
            self.trivial = is_trivial(node)



class ReplaceNames(Substitutions):

    fullNamesMapping = None

    result = None

    def __init__(self, nexp, fullNamesMapping):
        a_copy = copy.deepcopy(nexp)
        super(ReplaceNames, self).__init__(a_copy, a_copy)
        self.errors = list()
        self.fullNamesMapping = fullNamesMapping
        self.fullNamesMapping['QTArrays'] = 'java3.theories.QTArrays'
        self.fullNamesMapping['QTLists'] = 'java3.theories.QTLists'
        a_copy.accept(self)
        self.result = a_copy

    def MethodCall_visit(self, node):
        super(ReplaceNames, self).MethodCall_visit(node)
        if node.target is not None:
            targett = node.target.text()
            if self.fullNamesMapping.has_key(targett):
                node.target = parseString(self.fullNamesMapping[targett] + ';')



class ExecutableSubstitution(Substitutions):

    errors = []


    def __init__(self, nexp, root):
        super(ExecutableSubstitution, self).__init__(nexp, root)
        self.errors = list()


    def _offset(self, node, qvar):
        if not isinstance(node, AdditiveExpression):
            return ("", False) if node.text() == qvar else (node.text(), False)
        else:
            off1, neg1 = self._offset(node.left, qvar)
            off2, neg2 = self._offset(node.right, qvar)
            if isinstance(node, AddExpression):
                if off2 == "":
                    s = off1
                    neg = False
                else:
                    s = ' + '.join([off1, off2])
                    neg = neg1 or neg2
            elif isinstance(node, SubExpression):
                if off2 == "":
                    s = off1
                    neg = True
                else:
                    s = ' - '.join([off1, off2])
                    neg = neg1 or neg2
            return (s, neg)


    def _expr(self, node):
        if hasattr(node, 'args') and len(node.args) > 0:
            if len(node.args) != 1:
                self.errors.append("Warning: more than one argument expression in " + node.text())
            return node.args[0]
        else:
            self.errors.append("Warning: node has no args (leaving it unchanged): " + node.text())
            return node


    def _barecall(self, call):
        res = call.text()
        if isinstance(call, MethodCall):
            p = res.find('(')
        elif isinstance(call, ArrayReference):
            p = res.find('[')
        else:
            p = len(res)
        return res[:p]


    def unold(self, node): 
        assert isinstance(node, OldExpression)
        return parseString("old_" + node.args[0].text() + ';')

    
    def argwrap(self, old, expr, offset, neg, reflect):
        old_str = "old_" if old else ""
        expr_str = self._barecall(expr)
        if reflect:
            if hasattr(expr, 'target') and expr.target is not None:
                ct = ['"' + expr_str[expr_str.find('.') + 1:] + '"', old_str + expr.target.text()]
            else:
                self.errors.append("Cannot get arguments of: " + expr.text())
                ct = [old_str + expr_str]
        else:
            ct = [old_str + expr_str]
        return ct + ["0" if offset == "" else offset, "true" if neg else "false"]

    def get_bound(self, relex, qvar):
        if isinstance(relex, GreaterExpression) or isinstance(relex, GreaterEqualExpression):
            left, right = relex.right, relex.left
        else:
            left, right = relex.left, relex.right
        if isinstance(relex, GreaterExpression) or isinstance(relex, LessExpression):
            strict = True
        else:
            strict = False
        qv = qvar.text()
        ltxt, rtxt = left.text(), right.text()
        if rtxt == qv and ltxt != qv:
            bound = left
            lower = True
        elif ltxt == qv and rtxt != qv:
            bound = right
            lower = False
        else:
            self.errors.append("Malformed bound in: " + relex.text())
            bound = None
            lower = None
        return bound, lower, strict


    def BinaryExpression_visit(self, node):
        super(ExecutableSubstitution, self).BinaryExpression_visit(node)
        if isinstance(node.left, QuantifiedExpression) or isinstance(node.left, ImpliesExpression):
            node.left = node.left._sub
        elif isinstance(node.left, OldExpression):
            node.left = self.unold(node.left)
        if isinstance(node.right, QuantifiedExpression) or isinstance(node.right, ImpliesExpression):
            node.right = node.right._sub
        elif isinstance(node.right, OldExpression):
            node.right = self.unold(node.right)


    def UnaryExpression_visit(self, node):
        super(ExecutableSubstitution, self).UnaryExpression_visit(node)
        if isinstance(node.item, QuantifiedExpression) or isinstance(node.item, ImpliesExpression):
            node.item = node.item._sub
        elif isinstance(node.item, OldExpression):
            node.item = self.unold(node.item)


    def ImpliesExpression_visit(self, node):
        self.BinaryExpression_visit(node)
        node._sub = None
        try:
            node._sub = parseString('(' + '(' + node.left.text() + ')' + ' ? ' + \
                                        '(' + node.right.text() + ')' + ' : ' + \
                                        'true' + ')' + ';')
        except:
            self.errors.append("Dropping malformed implies expression")


    def QuantifiedExpression_visit(self, node):
        node._sub = None
        if node.quantifier != node.FORALL:
            self.errors.append("Existential quantification is not supported")
            return
        if len(node.variables) != 1:
            self.errors.append("Cannot handle quantifications over multiple variables")
            return
        qvar = node.variables[0].val
        scope = node.scope
        if not isinstance(scope, ImpliesExpression):
            self.errors.append("Cannot find implies expression of quantified expression: " + scope.text())
            return

        antec = scope.left
        conseq = scope.right

        if not isinstance(antec, AndExpression):
            self.errors.append("The antecedent is not a conjunction: " + antec.text())
            return

        leftmost = antec.left
        rightmost = antec.right
        if not isinstance(leftmost, RelationalExpression):
            self.errors.append("The leftmost antecedent expression doesn't define a lower bound: " + \
                               leftmost.text())
            return
        if not isinstance(rightmost, RelationalExpression):
            self.errors.append("The rightmost antecedent expression doesn't define a lower bound: " + \
                               leftmost.text())
            return

        lbound, llower, lstrict = self.get_bound(leftmost, qvar)
        rbound, rlower, rstrict = self.get_bound(rightmost, qvar)
        if lbound is None or rbound is None:
            self.errors.append("Giving up because of malformed bound")
            return
        if llower == rlower:
            self.errors.append("Two bounds of the same type in: " + antec.text())
            return
        if llower:
            # Left bound is lowerbound, and hence right bound is upperbound
            if not lstrict:
                low = lbound
            else:
                low = parseString(lbound.text() + ' + 1' + ';')
            if rstrict:
                high = rbound
            else:
                high = parseString(rbound.text() + ' + 1' + ';')
        else:
            # Right bound is lowerbound, and hence left bound is upperbound
            if not rstrict:
                low = rbound
            else:
                low = parseString(rbound.text() + ' + 1' + ';')
            if lstrict:
                high = lbound
            else:
                high = parseString(lbound.text() + ' + 1' + ';')

        if isinstance(conseq, EqualsExpression):
            pred = 'QEq'
        elif isinstance(conseq, NEqualsExpression):
            pred = 'QNeq'
        elif isinstance(conseq, LessExpression):
            pred = 'QLess'
        elif isinstance(conseq, LessEqualExpression):
            pred = 'QLessEq'
        elif isinstance(conseq, GreaterExpression):
            pred = 'QGrt'
        elif isinstance(conseq, GreaterEqualExpression):
            pred = 'QGrtEq'
        else:
            self.errors.append("I cannot translate consequent: " + conseq.text())
            return

        le, re = conseq.left, conseq.right

        if isinstance(le, OldExpression):
            lold = True
            while isinstance(le, OldExpression):
                # Nested old expressions
                le = le.args[0]
        else:
            lold = False
        if isinstance(re, OldExpression):
            rold = True
            while isinstance(re, OldExpression):
                # Nested old expressions
                re = re.args[0]
        else:
            rold = False

        switch = False
        kk = False
        if isinstance(le, ArrayReference) and isinstance(re, ArrayReference):
            theory = 'QTArrays'
            lrefl = False
            rrefl = False
        elif isinstance(le, ArrayReference) and isinstance(re, AttributeCall):
            theory = 'QTArrays'
            lrefl = False
            rrefl = False
            kk = True
        elif isinstance(le, AttributeCall) and isinstance(re, ArrayReference):
            theory = 'QTArrays'
            lrefl = False
            rrefl = False
            kk = True
            switch = True
        elif isinstance(le, MethodCall) and isinstance(re, MethodCall):
            theory = 'QTLists'
            lrefl = True
            rrefl = True
        elif isinstance(le, MethodCall) and isinstance(re, ArrayReference):
            theory = 'QTLists'
            lrefl = True
            rrefl = False
        elif isinstance(le, ArrayReference) and isinstance(re, MethodCall):
            theory = 'QTLists'
            lrefl = True
            rrefl = False
            switch = True
        else:
            self.errors.append("I cannot handle this combination of operands: " + \
                               conseq.text())
            return

        if not kk and (pred == 'QLess' or pred == 'QLessEq' or pred == 'QGrt' or pred == 'QGrtEq'):
            self.errors.append("I cannot translate relational operators " + \
                               " unless they are to a constant expression: " + conseq.text())
            return
        if kk and (pred == 'QEq' or pred == 'QNeq'):
            pred = pred + 'K'

        left_offset, left_neg = self._offset(self._expr(le), qvar.text())
        right_offset, right_neg = self._offset(self._expr(re), qvar.text())

        if switch:
            # Switching left and right
            lold, le, left_offset, left_neg, lrefl, \
                  rold, re, right_offset, right_neg, rrefl  =  \
                  rold, re, right_offset, right_neg, rrefl, \
                  lold, le, left_offset, left_neg, lrefl
            self.errors.append("Warning: switching left and right in " + conseq.text())
        
        if pred == 'QEq' or pred =='QNeq':
            args = [low.text(), high.text()] + \
                   self.argwrap(lold, le, left_offset, left_neg, lrefl) + \
                   self.argwrap(rold, re, right_offset, right_neg, rrefl)
        else:
            args = [low.text(), high.text()] + \
                   self.argwrap(lold, le, left_offset, left_neg, lrefl) + \
                   [("old_" if rold else "") + re.text()]

        sub_string = theory + '.' + pred + '(' + ', '.join(args) + ')' + ';'
        node._sub = parseString(sub_string)
        


class ExecutableJML(object):

    original_node = None

    node = None

    errors = None

    def __init__(self, JML_string, parsed=False):
        self.errors = list()
        if parsed:
            # self.original_node = parseString(JML_string.text() + ';')
            # Faster than reparsing
            self.original_node = JML_string
            self.node = self.original_node
            if not self.is_executable():
                # Transformations will not alter original
                self.original_node = copy.deepcopy(self.original_node)
        else:
            self.original_node = parseString(JML_string)
        self.node = self.original_node

    def get_node(self):
        return self.node

    def get_text(self):
        if self.is_executable():
            return self.node.text()
        else:
            return ""

    def is_executable(self):
        if self.node is None:
            return False
        try:
            t = self.node.text()
            if t.find('\\forall') == -1 and t.find('\\old') == -1 and t.find('\\result') == -1 and t.find('==>') == -1:
                return True
            else:
                return False
        except:
            return False

    def make_executable(self):
        if self.is_executable():
            # Already executable: nothing to do
            return
        t = self.original_node.text()
        if t.find('\\result') != -1:
            # Result expressions cannot be evaluated mid method: conversion impossible
            self.node = None
        else:
            e = ExecutableSubstitution(self.node, self.node)
            self.node.accept(e)
            self.errors = e.errors
            if isinstance(self.node, QuantifiedExpression) or isinstance(self.node, ImpliesExpression):
                self.node = self.node._sub
            if not self.is_executable():
                # Try removing remaining old expressions
                try:
                    ns = self.node.text().replace('\\old(', '(old_')
                    self.node = parseString(ns + ';')
                    self.errors.append("Warning: syntactic removal of remaining old expressions in " + \
                                       self.node.text())
                except:
                    self.errors.append("Corrupted node: cannot make executable")
                
    def make_quantification(self, IntegerArrays=True, preconditions=False, fixIntValues=False):
        pe = PredicateExpansion(self.node, self.node)
        if not IntegerArrays:
            pe.use_int_arrays()
        if preconditions:
            pe.use_preconditions()
        # if fixIntValues:
        #     pe.use_fixIntValues()
        self.node.accept(pe)
        self.errors = pe.errors
        if pe.failed:
            return False
        if isinstance(self.node, MethodCall) and self.node._sub is not None:
            self.node = self.node._sub
        # Simple conditional expressions go back to implications
        if isinstance(self.node, ConditionalExpression) and self.node.else_exp.text() == 'true':
            impl = ImpliesExpression()
            impl.left = self.node.condition
            impl.right = self.node.then_exp
            self.node = impl
        return True



def basic_run(line, command, errors, classname, methodname):
    ex = ExecutableJML(line)
    if command == 'exec':
        ex.make_executable()
        print ex.get_text()
    elif command == 'quant':
        ex.make_quantification()
        print ex.node.text()
    elif command == 'dynamate':
        if classname + '.' + methodname in methods_with_Integer_arrays:
            success = ex.make_quantification(IntegerArrays=True, preconditions=True)
        else:
            success = ex.make_quantification(IntegerArrays=False, preconditions=True)
        if ex.node is None or not success:
            print "FAILURE:"
        else:
            print "SUCCESS:", ex.node.text()
    if errors:
        if len(ex.errors) > 0:
            print "== Errors: =="
            for err in ex.errors:
                print err
        else:
            print "== No errors =="



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Convert a JML formula to executable form or to explicit quantification.')
    parser.add_argument('command', action='store', choices=['exec', 'quant', 'dynamate'], \
                             help='convert to executable form (exec) or introduce quantifiers (quant)')
    parser.add_argument('-f', dest='input_file', action='store_true', \
                        help='read input from file with name F')
    parser.add_argument('--class', dest='classname', action='store', \
                        help='fully qualified class name')
    parser.add_argument('--method', dest='methodname', action='store', \
                        help='method name')
    parser.add_argument('-e', dest='show_errors', action='store_true', \
                        help='print errors and warnings')
    parser.add_argument('-t', dest='def_test', action='store_true', \
                        help='run standard regression tests (F is ignored)')
    parser.add_argument('fmlas', nargs='+', metavar='F', action='append', \
                        help='a JML formula or a file name to be processed')
    args = parser.parse_args()
    if args.def_test:
        args.input_file = True
        if args.command == 'exec':
            args.fmlas = [['quantifiers.test']]
        elif args.command == 'quant':
            args.fmlas = [['predicates.test']]
    if args.command == 'dynamate':
        if args.classname is None or args.methodname is None:
            print "error: command 'dynamate' requires class and method names"
            sys.exit()
    if args.input_file:
        for fn in args.fmlas[0]:
            print "=== " + fn + " ==="
            with open(fn) as f:
                for l in f.readlines():
                    basic_run(l, args.command, args.show_errors, args.classname, args.methodname)
    else:
        for l in args.fmlas[0]:
            basic_run(l, args.command, args.show_errors, args.classname, args.methodname)
