#! /usr/bin/python

import jml2exec
from JMLformulasParser import parseString
from jmlvisitor import Substitutions
import JML_ast
import argparse



marker = "-----> All mutations follow"
separator = '   <--exec|JML-->  '
pkg = "java3.util3"
classes = ['Arrays']
classes = ['Collections']
methods_with_Integer = ['fillInteger0', 'fillInteger1'] #['fillInteger0', 'fillInteger1', 'hashCodeInteger', 'insertionSort0', 'merge0']
methods_with_int = [] # ['binarySearch0', 'equals0', 'hashCode0', 'binarySearch0', 'fill1', 'fill0', 'quicksortPartition', 'insertionSort1', 'vecswap']
methods = methods_with_Integer + methods_with_int
methods = ['reverse']
itmarker = 'iteration_'
iterations = range(1, 16 + 1)
# iterations = [1]
extension = 'mutations'



class CheckQuantifiedFormula(Substitutions):
    """Sets self.wellformed to True iff all quantified expressions are expressible as QTLists.QEq"""

    wellformed = None

    def __init__(self, nexp):
        super(CheckQuantifiedFormula, self).__init__(nexp, None)
        self.wellformed = True

    def has_variable(self, expr, var):
        if isinstance(expr, JML_ast.BinaryExpression):
            return self.has_variable(expr.left, var) or self.has_variable(expr.right, var)
        if isinstance(expr, JML_ast.UnaryExpression):
            return self.has_variable(expr.item, var)
        return expr.text().strip() == var.text().strip()

    def QuantifiedExpression_visit(self, node):
        super(CheckQuantifiedFormula, self).QuantifiedExpression_visit(node)
        if self.wellformed:
            # Will assign back to True iff all checks are passed
            self.wellformed = False
            if len(node.variables) != 1:
                return
            qv = node.variables[0].val
            if not isinstance(node.scope, JML_ast.ImpliesExpression):
                return
            eq = node.scope.right
            if not isinstance(eq, JML_ast.EqualsExpression):
                return
            lt = eq.left
            if isinstance(lt, JML_ast.OldExpression):
                lt = lt.args[0]
            rt = eq.right
            if isinstance(rt, JML_ast.OldExpression):
                rt = rt.args[0]
            if not isinstance(lt, JML_ast.FeatureCall) or not isinstance(rt, JML_ast.FeatureCall):
                return
            if not hasattr(lt, 'args') or not hasattr(rt, 'args'):
                return
            if len(lt.args) != 1 or len(rt.args) != 1:
                return
            # Quantified variable must exist in text
            if not self.has_variable(lt.args[0], qv) or not self.has_variable(rt.args[0], qv):
                return
            # All checks passed: reset to True
            self.wellformed = True



def wellformed_QT(pred, formula):
    if pred.find('QTLists.QEq') == -1:
        # Doesn't check QTArrays
        return True
    f = parseString(formula + ';')
    cqf = CheckQuantifiedFormula(f)
    f.accept(cqf)
    return cqf.wellformed



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Reprocess mutation files to add quantifiers.')
    parser.add_argument('--QTcheck', dest='QTcheck', action='store_true',
                        help='only perform a well-formedness check of mutants involving QT-predicates (default: %(default)s)')
    args = parser.parse_args()
    for c in classes:
        for m in methods:
            for i in iterations:
                it = itmarker + str(i)
                filename = pkg + '.' + c + '.' + m + '.' + it + '.' + extension
                with open(filename + '.' + 'new', 'w') as outf:
                    with open(filename) as f:
                        content = False
                        for line in f.readlines():
                            if line.strip() == marker:
                                content = True
                                outf.write(line)
                                continue
                            if not content:
                                outf.write(line)
                            else:
                                pred, quant = line.strip().split(separator)
                                if pred.find('java3.theories.QT') != -1:
                                    if args.QTcheck:
                                        if wellformed_QT(pred, quant):
                                            outf.write(pred + separator + quant + '\n')
                                    continue
                                if args.QTcheck:
                                    outf.write(pred + separator + quant + '\n')
                                    continue
                                newpred = pred.replace('java3.theories.', '')
                                isIntegerArray = m in methods_with_Integer
                                if isIntegerArray:
                                    parsedpred = parseString(newpred + ';')
                                    fint = jml2exec.FixIntValue(parsedpred, parsedpred)
                                    parsedpred.accept(fint)
                                    newpred = parsedpred.text()
                                    pred = newpred.replace('TArrays', 'java3.theories.TArrays')
                                if quant.find("\old") == -1:
                                    ex = jml2exec.ExecutableJML(newpred)
                                else:
                                    ex = jml2exec.ExecutableJML(quant)
                                success = ex.make_quantification(IntegerArrays=isIntegerArray, \
                                                                 preconditions=True)
                                if not success:
                                    continue
                                outline = pred + separator + ex.node.text() + '\n'
                                outf.write(outline)
