#!/usr/bin/python

from JML_ast import OldExpression
from JMLformulasParser import parseString
from jmlvisitor import Substitutions
import argparse
import sys



class OldRemover(Substitutions):

    failed = False
    old_level = None

    def __init__(self, expr):
        """A parsed JML expression `expr', to be converted by changing \old( arg ) expressions into old_arg expressions."""
        self.nexp = expr
        self.failed = False
        self.errors = list()
        self.old_level = 0

    def remove(self):
        self.nexp.accept(self)
        if not self.failed:
            return self.nexp.text()
        else:
            return None

    def OldExpression_visit(self, node):
        self.old_level += 1
        super(OldRemover, self).MethodCall_visit(node)
        if isinstance(node.args[0], OldExpression):
            node.args[0] = node.args[0]._sub
        if self.old_level > 1:    # For nested application: \old is idempotent
            node._sub = node.args[0]
        else:
            arg = node.args[0].text()
            failed = False
            try:
                newnode = parseString("old_" + arg + ";")
            except:
                self.errors.append("Failed to reparse the expression: " + "old_" + arg)
                failed = True
            if failed or newnode is None:
                self.failed = True
                node._sub = node
            else:
                node._sub = newnode
        self.old_level -=1

    def ConditionalExpression_visit(self, node):
        super(OldRemover, self).ConditionalExpression_visit(node)
        if isinstance(node.condition, OldExpression):
            node.condition = node.condition._sub
        if isinstance(node.then_exp, OldExpression):
            node.then_exp = node.then_exp._sub
        if isinstance(node.else_exp, OldExpression):
            node.else_exp = node.else_exp._sub

    def BinaryExpression_visit(self, node):
        super(OldRemover, self).BinaryExpression_visit(node)
        if isinstance(node.left, OldExpression):
            node.left = node.left._sub
        if isinstance(node.right, OldExpression):
            node.right = node.right._sub

    def UnaryExpression_visit(self, node):
        super(OldRemover, self).UnaryExpression_visit(node)
        if isinstance(node.item, OldExpression):
            node.item = node.item._sub

    def FeatureCall_visit(self, node):
        super(OldRemover, self).FeatureCall_visit(node)
        if isinstance(node.target, OldExpression):
            node.target = node.target._sub
        if node.call is not None:
            for i in range(len(node.call)):
                if isinstance(node.call[i], OldExpression):
                    node.call[i] = node.call[i]._sub

    def MethodCall_visit(self, node):
        super(OldRemover, self).MethodCall_visit(node)
        if node.args is not None:
            for i in range(len(node.args)):
                if isinstance(node.args[i], OldExpression):
                    node.args[i] = node.args[i]._sub

    def QuantifiedExpression_visit(self, node):
        super(OldRemover, self).QuantifiedExpression_visit(node)
        if isinstance(node.scope, OldExpression):
            node.scope = node.scope._sub



def unold(expr, debug=False):
    """Given a string `expr' representing a JML expression, return it with \old( arg ) expressions changed into old_arg expressions.
    Return None if the change failed (i.e., did not produce a parsable JML expression)."""
    try:
        pexpr = parseString(expr + ';')
    except:
        if debug:
            print >> sys.stderr, "Exception in parsing input expression:", expr
            raise
        pexpr = None
    if pexpr is not None:
        rem = OldRemover(pexpr)
        res = rem.remove()
        if debug:
            for e in rem.errors:
                print >> sys.stderr, e
        return res
    else:
        if debug:
            print >> sys.stderr, "Cannot parse input expression", expr
        return None



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Convert \old subexpressions into old_ form.')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='turn on debug information')
    parser.add_argument('--JP', dest='JP', action='store_true',
                        help='special output format')
    parser.add_argument('fmla', nargs=1, metavar='F', action='store', \
                        help='a JML formula to be converted')
    args = parser.parse_args()
    res = unold(args.fmla[0], debug=args.debug)
    outcome = "FAILURE:" if res is None else "SUCCESS:"
    if args.JP:
        print outcome, res
    else:
        print res
