'''
Created on Sep 10, 2013

@author: galeotti
'''
import datetime
import time
import os
import shutil
from tempfile import mkstemp
from shutil import move
from os import remove, close
from datetime import timedelta
import glob
from comment_invariant_checks import delete_invariants
import re
import subprocess
from queue import Queue
import fcntl
import platform

#=====================================================================================
# CWD should be loopmate-escj root folder
CWD = os.getcwd()

LOOPMATE_DIR = CWD

GINPINK_DIR = CWD + "/ginpink"
ESCJ_DIR = CWD + "/tools/escj"
JUNIT_JAR = CWD + "/tools/junit/junit-4.10.jar"
EVOSUITE_JAR = CWD + "/tools/evosuite/evosuite-20130619.jar"
DAIKON_JAR = CWD + "/tools/loopmate-daikon/dist/daikon.jar"
GINPINK_JAR = CWD + "/tools/ginpink/dist/ginpink.jar"

DAIKON_PREDICATES_JAR = CWD + "/tools/loopmate-daikon/dist/daikon-predicates.jar"

JAVA_CMD = "java"
JAVAC_CMD = "javac"
ESCJ_CMD = ESCJ_DIR + "/escj"

NO_TIMEOUT = -1

#=====================================================================================
class LoopmateException(Exception):
    pass
#=====================================================================================
class DaikonTimeout(LoopmateException):
    pass
#=====================================================================================
class GinpinkTimeout(LoopmateException):
    pass
#=====================================================================================
class EvoSuiteRestartLimit(LoopmateException):
    pass
#=====================================================================================
class TrivialVerification(LoopmateException):
    pass
#=====================================================================================
class LoopmateError(Exception):
    pass
#=====================================================================================
class EvoSuiteError(LoopmateError):
    pass
#=====================================================================================
class GinpinkError(LoopmateError):
    pass
#=====================================================================================
class ConfigurationError(LoopmateError):
    pass
#=====================================================================================
class FileNotFoundError(LoopmateError):
    pass
#=====================================================================================
class ESCJError(LoopmateError):
    pass
#=====================================================================================
class DaikonError(LoopmateError):
    pass
#=====================================================================================

class DaikonConfig:
    enable_more_invariants = True
    enable_more_derived_variables = False

#=====================================================================================
class VerifiedInvariants:
    
    verified_daikon_invariants_filenames = list()
    verified_ginpink_invariants_filenames = list()

    verified_invariants = dict()
    
    @staticmethod
    def is_empty():
        if len(VerifiedInvariants.verified_invariants) == 0:
            return True
        else:
            return False

    @staticmethod
    def all_invariants_are_verified(daikon_invariants, ginpink_invariants):
        return VerifiedInvariants.all_daikon_invariants_are_verified(daikon_invariants) \
            and VerifiedInvariants.all_ginpink_invariants_are_verified(ginpink_invariants)

    @staticmethod
    def all_ginpink_invariants_are_verified(ginpink_invariants):
        for loop_header in ginpink_invariants.keys():
            invariants = ginpink_invariants[loop_header]
            if len(invariants) > 0:
                loop_header_found = False
                for this_loop_header in VerifiedInvariants.verified_invariants.keys():
                    if this_loop_header.loop_id == loop_header.loop_id:
                        loop_header_found = True
                        verified_invariants = VerifiedInvariants.verified_invariants[this_loop_header]
                        for invariant in invariants:
                            verified_invariant_found = False
                            for verified_invariant in verified_invariants:
                                if invariant.jml_string == verified_invariant [0] or invariant.java_string == verified_invariant[1]:
                                    verified_invariant_found = True
                            
                            if verified_invariant_found == False:
                                return False
                        
                if loop_header_found == False:
                    return False
                
        return True
    
    @staticmethod
    def all_daikon_invariants_are_verified(daikon_invariants):
        for loop_header in daikon_invariants.keys():
            invariants = daikon_invariants[loop_header]
            if len(invariants) > 0:
                found = False
                for this_loop_header in VerifiedInvariants.verified_invariants.keys():
                    if this_loop_header.loop_id == loop_header.loop_id:
                        found = True
                        verified_invariants = VerifiedInvariants.verified_invariants[this_loop_header]
                        if not invariants.issubset(verified_invariants):
                            return False
                        
                if found == False:
                    return False
                
        return True
        
    @staticmethod
    def read_verified_invariants_file(daikon_invariant_filename):
        VerifiedInvariants.verified_invariants = read_invs_escj_plus_java(daikon_invariant_filename)
    
    @staticmethod
    def add_verified_daikon_invariants_filename(daikon_invariants_filename):
        VerifiedInvariants.verified_daikon_invariants_filenames.append(daikon_invariants_filename)

    @staticmethod
    def add_verified_ginpink_invariants_filename(ginpink_invariants_filename):
        VerifiedInvariants.verified_ginpink_invariants_filenames.append(ginpink_invariants_filename)

    @staticmethod
    def is_verified_invariant(loop_id, jml_invariant_str):
        curr_loop_header = None
        for loop_header in VerifiedInvariants.verified_invariants.keys():
            if loop_header.loop_id == loop_id:
                curr_loop_header = loop_header
                break
            
        if curr_loop_header == None:
            return False
        else:
            for  jml_part, java_part in VerifiedInvariants.verified_invariants[curr_loop_header]:
                if "::" in jml_part:
                    jml_str = jml_part.split("::")[1]
                else:
                    jml_str = jml_part
                if jml_str.strip() == jml_invariant_str.strip():
                    return True
            return False
    
    @staticmethod
    def add_new_verified_daikon_invariants(daikon_invariants):
        
        for loop_header, jml_java_pair_set in daikon_invariants.items():
            this_loop_header = None
            for curr_loop_header in VerifiedInvariants.verified_invariants.keys():
                if curr_loop_header.loop_id == loop_header.loop_id:
                    this_loop_header = curr_loop_header
                    break
            
            
            if this_loop_header == None:
                new_set = set(jml_java_pair_set)
                print("Loop invariant verification: Adding " + str(len(new_set)) + " verified loop invariants to " + loop_header.loop_id)
                VerifiedInvariants.verified_invariants[loop_header] = new_set
            else:
                loop_id = loop_header.loop_id
                old_invs_count = len(VerifiedInvariants.verified_invariants[this_loop_header])

                set_union = VerifiedInvariants.verified_invariants[this_loop_header] | jml_java_pair_set
                new_invs_count = len(set_union)
                if old_invs_count == new_invs_count:
                    print("Loop invariant verification: No new verified loop invariants for " + loop_id)
                else:
                    print("Loop invariant verification: Verified " + str(new_invs_count - old_invs_count) + " new loop invariants for " + loop_id)
                    print("Loop invariant verification: Number of loop invariants verified for " + loop_id + " grew from " + str(old_invs_count) + " to " + str(new_invs_count))
            
                VerifiedInvariants.verified_invariants[this_loop_header] = set_union    


    @staticmethod
    def add_new_verified_ginpink_invariants(ginpink_invariants):
        
        for loop_header, ginpink_invariant_set in ginpink_invariants.items():
            this_loop_header = None
            for curr_loop_header in VerifiedInvariants.verified_invariants.keys():
                if curr_loop_header.loop_id == loop_header.loop_id:
                    this_loop_header = curr_loop_header
                    break
            
            
            if this_loop_header == None:
                new_set = set(ginpink_invariant_set)
                print("Loop invariant verification: Adding " + str(len(new_set)) + " verified loop invariants to " + loop_header.loop_id)
                VerifiedInvariants.verified_invariants[loop_header] = new_set
            else:
                loop_id = loop_header.loop_id
                old_invs_count = len(VerifiedInvariants.verified_invariants[this_loop_header])

                for ginpink_invariant in ginpink_invariant_set:
                    new_tuple = ginpink_invariant.jml_string, ginpink_invariant.java_string
                    VerifiedInvariants.verified_invariants[this_loop_header].add(new_tuple)

                new_invs_count = len(VerifiedInvariants.verified_invariants[this_loop_header])
                if old_invs_count == new_invs_count:
                    print("Loop invariant verification: No new verified loop invariants for " + loop_id)
                else:
                    print("Loop invariant verification: Verified " + str(new_invs_count - old_invs_count) + " new loop invariants for " + loop_id)
                    print("Loop invariant verification: Number of loop invariants verified for " + loop_id + " grew from " + str(old_invs_count) + " to " + str(new_invs_count))
            
    @staticmethod
    def get_number_of_verified_loop_invariants():
        ret_val = 0
        for loop_header in VerifiedInvariants.verified_invariants.keys():
            ret_val += len(VerifiedInvariants.verified_invariants[loop_header])
        
        return ret_val

#=====================================================================================

class LoopmateConfig:
    # General settings
    work_dir = None
    target_class_name = None
    target_method_name = None
    target_class_path = None
    target_java_source_path = None
    target_escj_source_path = None

    purity_filename = None
    
    orig_path = None
    java6_home = None
    java5_home = None

    max_number_of_iterations = 100
       
    # Evosuite settings
    dse_rate = 0
    evosuite_global_timeout = 240  # default is 60
    search_budget = 50  # unused
    minimization_timeout = 10  # unused



    local_search_rate = 20
    min_initial_tests = 20
    max_initial_tests = 50
    max_array = 20
    instrument_context = "false"
    object_reuse_probability = 0.5
    null_probability = 0.2
    
    # ESCJ settings
    escj_hard_timeout = 180 * 2
    escj_vc_limit = 10000000

    chicory_hard_timeout = 600
    daikon_hard_timeout = 600
    ginpink_hard_timeout = 600

    
    # verbosity 0,1,2
    verbosity = 2
    
    # severity options
    stop_if_escj_reports_error = True
    
    max_number_of_evosuite_restarts_in_single_iteration = 5
    max_evosuite_split_depth = 10
    
    # enable gin-pink
    ginpink_is_enabled = True
    use_frame_loop_invariants = True
    
    evosuite_tauto_is_enabled = False
    remove_ginpink_candidates_if_vc_too_big = False
    remove_ginpink_candidates_using_new_verified_loop_invariants = False
    remove_method_calls_within_specs = False
    
#=====================================================================================

class Stats:
    evosuite_time = timedelta()
    chicory_time = timedelta()
    daikon_time = timedelta()
    ginpink_time = timedelta()
    escj_time = timedelta()
    total_time = timedelta()
    
    nrOfDaikonIterations = 0
    nrOfGinpinkIterations = 0
    highestGinpinkWave = 0

    nrOfRestartedIterations = 0
    
    
    initial_number_of_warnings = -1
    final_number_of_warnings = -1
    minimum_number_of_warnings = -1
    
    @staticmethod
    def notify_new_verified_escj_output(escj_output_str):
        
        new_number_of_warnings = ESCJResult(escj_output_str).get_warnings()
        if new_number_of_warnings < Stats.minimum_number_of_warnings:
            Stats.minimum_number_of_warnings = new_number_of_warnings 
        
    @staticmethod
    def set_initial_number_of_warnings(number_of_warnings):
        Stats.initial_number_of_warnings = number_of_warnings
        Stats.minimum_number_of_warnings = number_of_warnings

#=====================================================================================

def nonBlockRead(output):
    fd = output.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
    try:
        read_val = output.read()
        return read_val
    except:
        return None

#=====================================================================================

def run_subprocess_timeout(command, timeout=10, stop_flag_keyword=None, use_shell=False):
    proc = subprocess.Popen(command, bufsize=0, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=use_shell)
    poll_seconds = .250
    deadline = time.time() + timeout
    
    stdout = ""
    stderr = ""
   
    while (time.time() < deadline or timeout == -1)and proc.poll() == None:
        std_out_read = nonBlockRead(proc.stdout)
        std_err_read = nonBlockRead(proc.stderr)
        
        if std_out_read != None:
            new_stdout_str = std_out_read.decode("utf-8")
            stdout += new_stdout_str
            print(new_stdout_str)

        if std_err_read != None:
            new_stderr_str = std_err_read.decode("utf-8")
            print(new_stderr_str)
            stderr += new_stderr_str
        
        if stop_flag_keyword != None and stop_flag_keyword in stdout:
            break
        else:
            time.sleep(poll_seconds)

    timeout_ret_code = False
    if proc.poll() == None:
        proc_pid = proc.pid
        
        if platform.system() == 'Linux':
            print("Linux platform detected")
            print("Listing all processes launched by " + str(proc_pid))
            cmd = ['ps', '-eo', 'pid,ppid'] 
            ps_proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            ps_stdout, ps_stderr = ps_proc.communicate()
            ps_stdout_str = ps_stdout.decode("utf-8") 
            ps_stdout_str_lines = ps_stdout_str.split('\n')[1:]  # skip the header
            childs_of_pid = dict()
            for row in ps_stdout_str_lines:
                if not row: 
                    continue 
                child_pid_str, parent_pid_str = row.split() 
                child_pid = int(child_pid_str)
                parent_pid = int(parent_pid_str)
                
                if not parent_pid in childs_of_pid.keys():
                    childs_of_pid[parent_pid] = list() 
                
                childs_of_pid[parent_pid].append(child_pid)
                   
            killed_pids = list()
            killing_queue = Queue() 
            killing_queue.put(proc_pid)
            while not killing_queue.empty():
                pid = killing_queue.get()
                if pid in killed_pids:
                    continue
                else:
                    if pid in childs_of_pid.keys():
                        for child_pid in childs_of_pid[pid]:
                            killing_queue.put(child_pid)
                
                    if pid == proc_pid:
                        print("Killing master process " + str(pid))
                    else:
                        print("Killing descendant process " + str(pid))
                
                    os.system("kill -9 " + str(pid))
                    killed_pids.append(pid)
        else:
            print("Non-Linux platform detected:" + platform.system())
            os.system("kill -9 " + str(proc_pid))
        
        timeout_ret_code = True
        
        kill_start_timestamp = datetime.datetime.now()
        
        retcode = proc.poll()
        while retcode == None:
            time.sleep(poll_seconds)
            retcode = proc.poll()

        kill_stop_timestamp = datetime.datetime.now()
        
        

        print("Process " + str(proc_pid) + " terminated with ret_code " + str(retcode))
        killing_time = kill_stop_timestamp - kill_start_timestamp
        
        print("Total killing time: " + str(killing_time))
        
    
    std_out_read = nonBlockRead(proc.stdout)
    std_err_read = nonBlockRead(proc.stderr)
        
    if std_out_read != None:
        new_stdout_str = std_out_read.decode("utf-8")
        stdout += new_stdout_str
        print(new_stdout_str)

    if std_err_read != None:
        new_stderr_str = std_err_read.decode("utf-8")
        print(new_stderr_str)
        stderr += new_stderr_str

    return stdout, stderr, timeout_ret_code

#=====================================================================================

def javac(class_path, class_folder, source_files):
    print ("Compiling " + source_files)
    javac_cmd = JAVAC_CMD + " -cp " + class_path + " -d " + class_folder + " -g " + " -source 1.6 " + " -target 1.6" 
    if LoopmateConfig.verbosity == 2:
        javac_cmd += " -verbose "
    
    javac_cmd += source_files
    print ("Executing command " + javac_cmd)
    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(javac_cmd, timeout= -1, use_shell=True)

    return std_out_str, std_err_str

#=====================================================================================
def build_evosuite_setup_cmd(class_path, evosuite_work_dir_name):
    evosuite_setup_cmd = JAVA_CMD + " -cp " + EVOSUITE_JAR + " org.evosuite.EvoSuite -setup " + class_path + " -base_dir " + evosuite_work_dir_name
    return evosuite_setup_cmd
#=====================================================================================

def build_evosuite_cmd(class_path, evosuite_work_dir_name, target_method_prefix):
    evosuite_cmd = JAVA_CMD + " -Xmx2g -cp " + EVOSUITE_JAR + " org.evosuite.EvoSuite " 
    evosuite_cmd += " -cp " + class_path 
    evosuite_cmd += " -generateSuite " 
    evosuite_cmd += " -Dshow_progress=false " 
    evosuite_cmd += " -Dminimize=true "  
    evosuite_cmd += " -Dcoverage=false " 
    evosuite_cmd += " -Dassertions=false "
    evosuite_cmd += " -Dinstrument_context=false " 
    evosuite_cmd += " -mem 2000 "
    evosuite_cmd += " -Dstopping_condition=MaxTime " 
    evosuite_cmd += " -Dmax_array=" + str(LoopmateConfig.max_array) + " "
    evosuite_cmd += " -Dobject_reuse_probability=" + str(LoopmateConfig.object_reuse_probability) + " " 
    evosuite_cmd += " -Dlocal_search_rate=" + str(LoopmateConfig.local_search_rate) + " "
    evosuite_cmd += " -Dsearch_budget=" + str(LoopmateConfig.evosuite_global_timeout) + " "
#    evosuite_cmd += " -Dsearch_budget=" + str(LoopmateConfig.search_budget) + " "
#    evosuite_cmd += " -Dminimization_timeout=" + str(LoopmateConfig.minimization_timeout) + " "
    evosuite_cmd += " -Dnull_probability=" + str(LoopmateConfig.null_probability) + " "
    evosuite_cmd += " -base_dir " + evosuite_work_dir_name + " "
    evosuite_cmd += " -class " + LoopmateConfig.target_class_name + " "
    evosuite_cmd += " -Dtarget_method_prefix=" + target_method_prefix + " "

    return evosuite_cmd

#=====================================================================================

def split_evosuite_problem(src_java_filename, base_dir1, base_dir2):
    
    src_folder1 = base_dir1 + "/" + "src"
    src_folder2 = base_dir2 + "/" + "src"
    
    ensure_dir(src_folder1)
    ensure_dir(src_folder2)
    
    source_filename1 = src_folder1 + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    source_filename2 = src_folder2 + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"

    dirname1 = os.path.dirname(source_filename1)
    dirname2 = os.path.dirname(source_filename2)

    ensure_dir(dirname1)
    ensure_dir(dirname2)

    source_file1 = open(source_filename1, 'w')
    source_file2 = open(source_filename2, 'w')
    
    line_counter = 0
    lines = open(src_java_filename).readlines()

    predicate_length = 14
    
    while line_counter < len(lines):

        if "private static void " + LoopmateConfig.target_method_name + "_loop_invariant_" in lines[line_counter] or "public static void " + LoopmateConfig.target_method_name + "_loop_invariant_" in lines[line_counter]:
            method_decl = lines[line_counter]
            source_file1.write(method_decl)
            source_file2.write(method_decl)
            
            line_counter += 1
            bracket_line = lines[line_counter]
            source_file1.write(bracket_line)
            source_file2.write(bracket_line)            
            
            predicate_count = 0
            
            line_counter += 1
            
            while not "methodThrewSomething_instrument" in lines[line_counter]:
                if lines[line_counter].strip() == "":
                    source_file1.write(lines[line_counter])
                    source_file2.write(lines[line_counter])
                    line_counter += 1
                elif "/*ginpink invariants*/" in lines[line_counter]:
                    source_file1.write(lines[line_counter])
                    source_file2.write(lines[line_counter])
                    line_counter += 1
                elif "/*daikon invariants*/" in lines[line_counter]:
                    source_file1.write(lines[line_counter])
                    source_file2.write(lines[line_counter])
                    line_counter += 1
                elif "/*Begin Loop Invariants Checks*/"in lines[line_counter]:
                    while True:
                        source_file1.write(lines[line_counter])
                        source_file2.write(lines[line_counter])
                        line_counter += 1
                        if "/*End Loop Invariants Checks*/" in lines[line_counter - 1]:
                            break
                else:
                    for pred_offset in range(predicate_length):
                        if line_counter >= len(lines):
                            print("Error")
                            
                        if predicate_count % 2 == 0:
                            source_file1.write(lines[line_counter])
                        else:
                            source_file2.write(lines[line_counter])
                        line_counter += 1
                
                    predicate_count += 1

            # Print end of method
            methodThrewSomething_instrument = lines[line_counter]
            line_counter += 1
            internal_method_call = lines[line_counter]
            
            line_counter += 1
            closing_bracket = lines[line_counter]
            
            source_file1.write(methodThrewSomething_instrument)
            source_file2.write(methodThrewSomething_instrument)

            source_file1.write(internal_method_call)
            source_file2.write(internal_method_call)

            source_file1.write(closing_bracket)
            source_file2.write(closing_bracket)

        else:
            source_file1.write(lines[line_counter])
            source_file2.write(lines[line_counter])
            
        line_counter += 1
            
        
        
    source_file1.close()
    source_file2.close()
    
    class_folder1 = base_dir1 + "/" + "bin"
    class_folder2 = base_dir2 + "/" + "bin"

    return source_filename1, class_folder1, source_filename2, class_folder2

#=====================================================================================
class EvoSuiteTask:
    evosuite_setup_cmd = None
    evosuite_cmd = None
    source_java_file = None
    base_dir = None
    junit_file_name = None
    depth = None
    task_can_be_splitted = None
    restart_counter = 0
#=====================================================================================

def  merge_junit_test_suites(junit_file_names, merge_evosuite_tests_dir):
    if len(LoopmateConfig.target_class_name.split(".")) > 1:
        test_package = "/".join(LoopmateConfig.target_class_name.split(".")[:-1])
    else:
        test_package = ""

    dst_folder = merge_evosuite_tests_dir + "/" + test_package
    junit_file_name = dst_folder + "/" + "Test" + LoopmateConfig.target_class_name.split(".")[-1] + ".java"
    if len(junit_file_names) == 1:
        ensure_dir(dst_folder)
        if junit_file_names[0] != junit_file_name:
            shutil.copy(junit_file_names[0], junit_file_name)        
    else:
        
        import_set = set()
        test_case_list = list()
        
        test_case_counter = -1
        for junit_filename in junit_file_names:
            junit_class_name = get_junit_class_name(junit_filename)
            junit_lines = open(junit_filename, 'r').readlines()
            line_counter = 0
            while line_counter < len(junit_lines):
                curr_line = junit_lines[line_counter]
                if curr_line.startswith("import"):
                    import_set.add(curr_line)
                
                elif curr_line.startswith("public class Test" + LoopmateConfig.target_class_name.split(".")[-1]):
                    line_counter += 1
                    curr_line = junit_lines[line_counter]
                    while not curr_line.startswith("}"):

                        if "//Test case number:" in curr_line:
                            test_case_counter += 1
                        
                        if "public void test" in curr_line:
                            curr_line = curr_line.replace("public void test", "public void test_" + str(test_case_counter) + "_")
                        
                        test_case_list.append(curr_line)
                        line_counter += 1
                        curr_line = junit_lines[line_counter]
                        
                line_counter += 1
                    
        if len(LoopmateConfig.target_class_name.split(".")[:-1]) > 0:
            package_name = ".".join(LoopmateConfig.target_class_name.split(".")[:-1])
        else:
            package_name = None
                
        junit_str = ""
        junit_str += "/*" + "\n"
        junit_str += " * This file was automatically generated by LoopMate" + "\n"
        junit_str += " */" + "\n"
            
        if package_name != None:
            junit_str += "package " + package_name + ";" + "\n"
        
        junit_str += "\n"
        junit_str += "".join(import_set) + "\n"
            
        junit_str += "@RunWith(EvoSuiteRunner.class)" + "\n"
        junit_str += "public class " + "Test" + LoopmateConfig.target_class_name.split(".")[-1] + " {" + "\n"
        junit_str += "".join(test_case_list) + "\n"
        junit_str += "}" + "\n"
            
        junit_folder = os.path.dirname(junit_file_name)
        ensure_dir(junit_folder)
        junit_file = open(junit_file_name, "w")
        junit_file.write(junit_str)
        junit_file.close()
        
    return junit_file_name

#=====================================================================================
# Spawns a new evosuite process and kills it if the hard_timeout is reached
#=====================================================================================

def spawn_evosuite_process(evosuite_cmd, evosuite_hard_timeout, evosuite_base_dir):

    
    print("Compiling target for EvoSuite")
    java_class_path = DAIKON_PREDICATES_JAR + ":" + LoopmateConfig.target_class_path
    javac_target_folder = evosuite_base_dir + "/" + "bin"
    ensure_dir(javac_target_folder)
    javac_source_files = evosuite_base_dir + "/" + "src" + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    javac_std_out, javac_std_err = javac(java_class_path, javac_target_folder, javac_source_files)

    if "code too large for try statement" in javac_std_err:
        print("javac: code too large for try statement")
        timeout_ret_code = True
        return timeout_ret_code
    else:
        print("javac: no code too large found")
        print("Checking successful compilation")
        target_class_file_name = javac_target_folder + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".class"

        assertFile(target_class_file_name)
    
    print("Executing command " + evosuite_cmd)

    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(evosuite_cmd.split(" ") , timeout=evosuite_hard_timeout)

    evosuite_out_filename = evosuite_base_dir + "/" + "evosuite_out.txt"
    evosuite_err_filename = evosuite_base_dir + "/" + "evosuite_err.txt"
    
    evosuite_out_file = open(evosuite_out_filename, 'w')
    evosuite_out_file.write(std_out_str)
    evosuite_out_file.close()

    evosuite_err_file = open(evosuite_err_filename, 'w')
    evosuite_err_file.write(std_err_str)
    evosuite_err_file.close()

    if timeout_ret_code == True:
        print("EvoSuite timeout!")
    else:
        print("EvoSuite finished within time threshold")

    print("EvoSuite return code: " + str(timeout_ret_code))
    
    return timeout_ret_code

#=====================================================================================

def split_evosuite_task(evosuite_task, target_method_prefix):
    task_source_file = evosuite_task.source_java_file
    base_dir = evosuite_task.base_dir
    
    print("Splitting task: " + task_source_file)
    base_dir1 = base_dir + ".0"
    base_dir2 = base_dir + ".1"
                
    source_file1, class_folder1, source_file2, class_folder2 = split_evosuite_problem(task_source_file, base_dir1, base_dir2)
    class_path1 = class_folder1 + ":" + LoopmateConfig.target_class_path + ":" + DAIKON_PREDICATES_JAR 
    class_path2 = class_folder2 + ":" + LoopmateConfig.target_class_path + ":" + DAIKON_PREDICATES_JAR 
                
    evosuite_task1 = EvoSuiteTask()
    evosuite_task1.base_dir = base_dir1
    evosuite_task1.source_java_file = source_file1
    evosuite_task1.evosuite_setup_cmd = build_evosuite_setup_cmd(class_path1, base_dir1)
    evosuite_task1.evosuite_cmd = build_evosuite_cmd(class_path1, base_dir1, target_method_prefix)
    evosuite_task1.depth = evosuite_task.depth + 1
    evosuite_task1.task_can_be_splitted = True
                
    evosuite_task2 = EvoSuiteTask()
    evosuite_task2.base_dir = base_dir2
    evosuite_task2.source_java_file = source_file2
    evosuite_task2.evosuite_setup_cmd = build_evosuite_setup_cmd(class_path2, base_dir1)
    evosuite_task2.evosuite_cmd = build_evosuite_cmd(class_path2, base_dir2, target_method_prefix)
    evosuite_task2.depth = evosuite_task.depth + 1
    evosuite_task2.task_can_be_splitted = True    
    
    print("Creating new task " + source_file1)
    print("New task depth is " + str(evosuite_task1.depth))
    
    print("Creating new task " + source_file2)
    print("New task depth is " + str(evosuite_task2.depth))
    
    return evosuite_task1, evosuite_task2
#=====================================================================================
def execute_evosuite(src_target_java_file_name, evosuite_work_dir_name, iterCount):
    class_folder = evosuite_work_dir_name + "/" + "bin"
    class_path = class_folder + ":" + LoopmateConfig.target_class_path + ":" + DAIKON_PREDICATES_JAR 
    
    print("Evosuite is enabled")

    evosuite_setup_cmd = build_evosuite_setup_cmd(class_path, evosuite_work_dir_name) 
    if iterCount == 0:
        target_method_prefix = LoopmateConfig.target_method_name
    else:
        target_method_prefix = LoopmateConfig.target_method_name + "_loop_invariant_"

    evosuite_cmd = build_evosuite_cmd(class_path, evosuite_work_dir_name, target_method_prefix)
        
    evosuite_task = EvoSuiteTask()
    evosuite_task.base_dir = evosuite_work_dir_name
    evosuite_task.source_java_file = src_target_java_file_name
    evosuite_task.evosuite_setup_cmd = evosuite_setup_cmd
    evosuite_task.evosuite_cmd = evosuite_cmd
    evosuite_task.depth = 0
        
    if iterCount == 0:
        evosuite_task.task_can_be_splitted = False
    else:
        evosuite_task.task_can_be_splitted = True
            

    work_queue = Queue()
    work_queue.put(evosuite_task)
        
    completed_tasks = list()
    junit_file_names = list() 

    print("Running EvoSuite on " + LoopmateConfig.target_class_name)
    while not work_queue.empty():
            
        evosuite_task = work_queue.get()

        start_timestamp = datetime.datetime.now()
            
        evosuite_setup_cmd = evosuite_task.evosuite_setup_cmd
        print("Executing " + evosuite_cmd)
        os.system(evosuite_setup_cmd)
            
        evosuite_cmd = evosuite_task.evosuite_cmd
        print("Executing " + evosuite_cmd)
        evosuite_base_dir = evosuite_task.base_dir
        evosuite_hard_timeout = spawn_evosuite_process(evosuite_cmd, LoopmateConfig.evosuite_global_timeout * 2, evosuite_base_dir)

        if evosuite_hard_timeout == True:
            print("A hard timeout has reached executing EvoSuite.")

        end_timestamp = datetime.datetime.now()
        execution_time = end_timestamp - start_timestamp
        print("Evosuite execution time " + str(execution_time))
        Stats.evosuite_time += (execution_time)
    
        print("Checking succesful EvoSuite execution")
        base_dir = evosuite_task.base_dir
        evosuite_tests_dir_name = base_dir + "/" + "evosuite-tests"
        test_dirname = os.path.dirname(evosuite_tests_dir_name + "/" + LoopmateConfig.target_class_name.replace(".", "/"))
        test_package = LoopmateConfig.target_class_name.split(".")[:-1]
        test_name = "Test" + LoopmateConfig.target_class_name.split(".")[-1]
        junit_file_name = test_dirname + "/" + test_name + ".java"
        
        task_source_file = evosuite_task.source_java_file
        
        if os.path.exists(junit_file_name):
            print("EvoSuite JUnit Test Suite was generated for current task: " + task_source_file)
            evosuite_task.junit_file_name = junit_file_name
            junit_file_names.append(junit_file_name)
            completed_tasks.append(evosuite_task)
                
        elif evosuite_task.task_can_be_splitted == False:
            print("Loopmate cannot split this EvoSuite problem because the task was flagged as unsplittable. ")
            print("No Test JUnit Test Suite was created for current task: " + task_source_file)
            evosuite_task.restart_counter += 1
            if  evosuite_task.restart_counter < LoopmateConfig.max_number_of_evosuite_restarts_in_single_iteration:
                print("Since the number of restarts for the current task is " + str(evosuite_task.restart_counter) + " less than the maximum number of restarts. We put the task back to the work queue")
                work_queue.put(evosuite_task)
            else:
                print("The task is no enqueded again since it has already " + str(evosuite_task.restart_counter) + " restarts")
            
                
        elif evosuite_task.depth == LoopmateConfig.max_evosuite_split_depth:
            print("Loopmate cannot split this EvoSuite problem because maximum depth was reached: " + str(LoopmateConfig.max_evosuite_split_depth))
            print("No Test JUnit Test Suite was created for current task: " + task_source_file)
                
        else:
            print("Loopmate can split this EvoSuite problem because maximum depth was not reached yet. ")
            print("current task depth " + str(evosuite_task.depth))
            split_depth = evosuite_task.depth

            evosuite_task1, evosuite_task2 = split_evosuite_task(evosuite_task, target_method_prefix)
            
            tasks = list()
            tasks.append(evosuite_task1)
            tasks.append(evosuite_task2)
                
                # Split other tasks
            while not work_queue.empty():
                evosuite_task = work_queue.get()
                if evosuite_task.task_can_be_splitted == True and evosuite_task.depth == split_depth:
                    evosuite_task1, evosuite_task2 = split_evosuite_task(evosuite_task, target_method_prefix)
                    tasks.append(evosuite_task1)
                    tasks.append(evosuite_task2)
                else:
                    tasks.append(evosuite_task)

                # Populate queue again
            for evosuite_task in tasks:
                work_queue.put(evosuite_task)
                
                
        
    if len(completed_tasks) == 0:
        print("No EvoSuite task was able to generate a Test Suite properly")
        print("Loopmate is exitting because there are no new tests to process")
        raise EvoSuiteError()
            
    merge_evosuite_tests_dir = evosuite_work_dir_name + "/" + "merged-tests"
    junit_file_name = merge_junit_test_suites(junit_file_names, merge_evosuite_tests_dir)
    assertFile(junit_file_name)    
    
    
    new_class_name = test_name + str(iterCount)
        
    print("renaming generated test suite to " + new_class_name)
    pattern = "public class " + test_name
    subst = "public class " + new_class_name
    replace_in_file(junit_file_name, pattern, subst)
    
    old_file_name = junit_file_name;
    old_dirname = os.path.dirname(old_file_name)
    new_file_name = old_dirname + "/" + new_class_name + ".java"
    os.rename(old_file_name, new_file_name)
    
    return new_file_name

#=====================================================================================

def get_junit_class_name(junit_filename):
    class_name_plus_java = junit_filename.split("/")[-1]
    junit_class_name = class_name_plus_java.split(".")[0]
    if len(LoopmateConfig.target_class_name.split(".")) > 1:
        package_name = ".".join(LoopmateConfig.target_class_name.split(".")[:-1])
        junit_class_name = package_name + "." + junit_class_name

    return junit_class_name

#=====================================================================================

def is_java_method_modifier(token):
    return token == "private" or token == "public" or token == "protected" or token == "static" or token == "abstract" or token == "final" or token == "native" or token == "synchronized" or token == "strictfp"

#=====================================================================================

def parse_purity_file(purity_filename):
    pure_method_list = list()
    for purity_line in open(purity_filename, 'r').readlines():
        purity_tokens = purity_line.strip().split()
        
        if is_java_method_modifier(purity_tokens[0]):
            method_name = purity_tokens[2]
        else:
            method_name = purity_tokens[1]
        
        method_name = method_name.split("(")[0].strip()
        pure_method_list.append(method_name)
    
    return pure_method_list

#=====================================================================================

def add_unreachable_code_methods(source_path):
    return "java3.lang3.Integer.unreachable_method|java3.util3.ArrayDeque.unreachable_method|java3.util3.ArrayList.unreachable_method|java3.util3.Vector.unreachable_method|java3.util3.Arrays.unreachable_method|java3.util3.Collections.unreachable_method|mbc.List.unreachable_method"

#=====================================================================================

def execute_chicory(daikon_work_dir_name, curr_dir_bin, junit_file_name, std_visibility, purity_filename):
    
    print("Executing Chicory on " + junit_file_name)
    
    start_timestamp = datetime.datetime.now()
    
    chicory_class_path = DAIKON_PREDICATES_JAR + ":" + DAIKON_JAR + ":" + JUNIT_JAR + ":" + curr_dir_bin + ":" + EVOSUITE_JAR + ":" + LoopmateConfig.target_class_path
    os.chdir(daikon_work_dir_name)
    
    new_dtrace_filename = get_junit_class_name(junit_file_name) + ".dtrace.gz"

    chicory_ppt_select_pattern = LoopmateConfig.target_method_name + "_" + "loop_invariant" 
    if purity_filename != None:
        pure_method_list = parse_purity_file(purity_filename)
        if len(pure_method_list) > 0:
            new_ppt_select_pattern = "|".join(pure_method_list) + "|" + chicory_ppt_select_pattern
            chicory_ppt_select_pattern = new_ppt_select_pattern


    chicory_ppt_select_pattern += "|" + add_unreachable_code_methods(LoopmateConfig.target_java_source_path)

    chicory_cmd = JAVA_CMD 
    chicory_cmd += " -Xmx2g " 
    chicory_cmd += "-cp " + chicory_class_path 
    chicory_cmd += " daikon.Chicory " 
    chicory_cmd += " --premain=" + DAIKON_JAR + " " 
    chicory_cmd += " --ppt-select-pattern='" + chicory_ppt_select_pattern + "'  " 
    chicory_cmd += " --ppt-omit-pattern='.*loop_invariant_.*:::EXIT.*'  " 
    chicory_cmd += " --dtrace-file=" + new_dtrace_filename + " "
    chicory_cmd += " --output-dir=" + daikon_work_dir_name + " "
    if std_visibility == True:
        chicory_cmd += " --std-visibility "
       
        
    if purity_filename != None:
        chicory_cmd += " --purity-file=" + purity_filename + " "
        
    chicory_cmd += " org.junit.runner.JUnitCore "
    chicory_cmd += " " + get_junit_class_name(junit_file_name)
        
    print("Executing Chicory command " + chicory_cmd)
    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(chicory_cmd, timeout= -1 , use_shell=True)
    
    chicory_std_out_filename = daikon_work_dir_name + "/" + "chicory_out.txt"
    chicory_std_err_filename = daikon_work_dir_name + "/" + "chicory_err.txt"
    
    chicory_out_file = open(chicory_std_out_filename, 'w')
    chicory_out_file.write(std_out_str)
    chicory_out_file.close()

    chicory_err_file = open(chicory_std_err_filename, 'w')
    chicory_err_file.write(std_err_str)
    chicory_err_file.close()
    
    new_dtrace_file = daikon_work_dir_name + "/" + new_dtrace_filename
    assertFile(new_dtrace_file)

    end_timestamp = datetime.datetime.now()
    execution_time = end_timestamp - start_timestamp

    Stats.chicory_time += execution_time
    print("Chicory execution time " + str(execution_time))

    return new_dtrace_file

#=====================================================================================

def execute_daikon(daikon_work_dir_name, new_dtrace_file, dtrace_files, junit_file_name):
    
    print("Executing Daikon on " + junit_file_name)
    
    start_timestamp = datetime.datetime.now()
    inv_filename = daikon_work_dir_name + "/" + get_junit_class_name(junit_file_name) + ".inv.gz"

    daikon_ppt_select_pattern = LoopmateConfig.target_method_name + "_" + "loop_invariant" 

    daikon_cmd = JAVA_CMD + " -Xmx2g -cp " + DAIKON_JAR + " daikon.Daikon "
    daikon_cmd += "--config_option daikon.derive.Derivation.disable_derived_variables=false "
    daikon_cmd += "--config_option daikon.PptSliceEquality.set_per_var=true "
    daikon_cmd += "--config_option daikon.Daikon.use_dynamic_constant_optimization=true "
    daikon_cmd += "--nohierarchy "
    daikon_cmd += " --ppt-select-pattern='" + daikon_ppt_select_pattern + "'  " 
    daikon_cmd += " --var-omit-pattern='old_.*' " 

    
    if DaikonConfig.enable_more_invariants == True:
        daikon_cmd += "--config_option daikon.inv.ternary.threeScalar.FunctionBinary.enabled=true "
        daikon_cmd += "--config_option daikon.inv.unary.sequence.SeqIndexIntEqual.enabled=true "
        
    if DaikonConfig.enable_more_derived_variables == True:
        daikon_cmd += "--config_option daikon.derive.binary.SequenceScalarSubsequence.enabled=true "  # array[..sclr] or array[sclr...]
        daikon_cmd += "--config_option daikon.derive.ternary.SequenceScalarArbitrarySubsequence.enabled=true "  # array[sclr1..sclr2]
    
    if LoopmateConfig.verbosity >= 2:
        daikon_cmd += "--config_option daikon.Daikon.quiet=false "
    else:
        daikon_cmd += "--config_option daikon.Daikon.quiet=true "
        daikon_cmd += "--no_text_output "
    

        
    daikon_cmd += " -o " + inv_filename + " "
    
    
    daikon_cmd_new_dtrace = daikon_cmd + new_dtrace_file + " "
    print("Executing Daikon on latest dtrace file " + daikon_cmd_new_dtrace)
    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(daikon_cmd_new_dtrace, timeout=LoopmateConfig.daikon_hard_timeout, use_shell=True)

    daikon_new_std_out_filename = daikon_work_dir_name + "/" + "daikon_new_out.txt"
    daikon_new_std_err_filename = daikon_work_dir_name + "/" + "daikon_new_err.txt"
    
    daikon_new_out_file = open(daikon_new_std_out_filename, 'w')
    daikon_new_out_file.write(std_out_str)
    daikon_new_out_file.close()

    daikon_new_err_file = open(daikon_new_std_err_filename, 'w')
    daikon_new_err_file.write(std_err_str)
    daikon_new_err_file.close()
    
    
    
    if timeout_ret_code == True:
        print("Normal Termination: Timeout")
        print("Daikon timeout was reached analyzing new dtrace file")
        print("Loopmate is exiting because Daikon timeout has reached in " + str(LoopmateConfig.daikon_hard_timeout) + " secs.")
        raise DaikonTimeout()
    
    if not os.path.exists(inv_filename):
        print("No Invariant File was found for latest dtrace file: " + inv_filename)
        print ("We asume no program points were covered. Restart iteration.")
        inv_ret_val = None

    else:
        
        if len(dtrace_files) == 0:
            print("No other dtrace files to merge")
            inv_ret_val = inv_filename

        else:

            daikon_cmd_all_dtraces = daikon_cmd
            daikon_cmd_all_dtraces += " ".join(dtrace_files) 
    
            daikon_cmd_all_dtraces += " " + new_dtrace_file + " "
            print("Executing Diakon command on all dtrace files " + daikon_cmd_all_dtraces)
    
            stdout, stderr, timeout_ret_code = run_subprocess_timeout(daikon_cmd_all_dtraces, timeout=LoopmateConfig.daikon_hard_timeout, use_shell=True)

            daikon_all_std_out_filename = daikon_work_dir_name + "/" + "daikon_all_out.txt"
            daikon_all_std_err_filename = daikon_work_dir_name + "/" + "daikon_all_err.txt"
    
            daikon_all_out_file = open(daikon_all_std_out_filename, 'w')
            daikon_all_out_file.write(stdout)
            daikon_all_out_file.close()

            daikon_all_err_file = open(daikon_all_std_err_filename, 'w')
            daikon_all_err_file.write(stderr)
            daikon_all_err_file.close()
            
            if timeout_ret_code == True:
                print("Normal Termination: Timeout")
                print("Daikon timeout was reached analyzing all dtrace files")
                print("Loopmate is exiting because Daikon timeout has reached in " + str(LoopmateConfig.daikon_hard_timeout) + " secs.")
                raise DaikonTimeout()

            # If the inv file does not exists, it means an ERROR happened
            assertFile(inv_filename)
            
            inv_ret_val = inv_filename


    end_timestamp = datetime.datetime.now()
    execution_time = end_timestamp - start_timestamp
    Stats.daikon_time += execution_time
    print("Daikon execution time " + str(execution_time))
    
    return new_dtrace_file, inv_ret_val

#=====================================================================================

def ensure_dir(d):
    if not os.path.exists(d):
        os.makedirs(d)
#=====================================================================================

def assertFile(f):
    if not os.path.exists(f):
        print("Terminal Error")
        print ("File " + f + " was not found.")
        print("Loopmate is exiting because an a required file was missing.")

        raise FileNotFoundError()
    else:
        print("Existence of file " + f + " checked")

#=====================================================================================

def prepare_evosuite(src_target_java_file_name, evosuite_work_dir_name):
    curr_dir_src = evosuite_work_dir_name + "/" + "src"
    curr_dir_bin = evosuite_work_dir_name + "/" + "bin"

    ensure_dir(evosuite_work_dir_name)     
    ensure_dir(curr_dir_src) 
    ensure_dir(curr_dir_bin)     
        
    dst_target_java_file_name = curr_dir_src + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"

    print("Copying java files to evosuite src folder")
    dst_target_dir_name = os.path.dirname(dst_target_java_file_name)
    ensure_dir(dst_target_dir_name)
    shutil.copy(src_target_java_file_name, dst_target_java_file_name)
    
    uncommented_verified_loop_invariants_filename = dst_target_java_file_name + "x"
    shutil.copy(dst_target_java_file_name, uncommented_verified_loop_invariants_filename)
    delete_invariants(VerifiedInvariants.verified_invariants, uncommented_verified_loop_invariants_filename, dst_target_java_file_name)
    


#=====================================================================================

def prepare_daikon(daikon_work_dir_name, junit_file_names, iterCount):
    curr_dir_src = daikon_work_dir_name + "/" + "src"
    curr_dir_bin = daikon_work_dir_name + "/" + "bin"

    ensure_dir(daikon_work_dir_name)     
    ensure_dir(curr_dir_src) 
    ensure_dir(curr_dir_bin)     

    src_target_java_file_name = LoopmateConfig.target_java_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dst_target_java_file_name = curr_dir_src + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dst_target_dir_name = os.path.dirname(dst_target_java_file_name)
    ensure_dir(dst_target_dir_name)

    shutil.copy(src_target_java_file_name, dst_target_java_file_name)
    
    print("Gathering JUnit test suites from previous iteration")
    if iterCount > 0:
        previous_iter_src = daikon_work_dir_name + "/" + ".." + "/" + "iter" + str(iterCount - 1) + "/" + "src"
        for test_case_file in glob.glob(previous_iter_src + "/" + "Test" + LoopmateConfig.target_class_name.replace(".", "/") + "*" + ".java"):
            shutil.copy(test_case_file, dst_target_dir_name)
    
    for junit_file_name in junit_file_names:
        shutil.copy(junit_file_name, dst_target_dir_name)
        
    print("Preparing JUnit test runner")
    test_suite_dir_name = os.path.dirname(dst_target_java_file_name)
    class_names = get_junit_class_names(test_suite_dir_name, curr_dir_src)
    
    print("Compiling target/helper for Daikon")
    class_path = DAIKON_JAR + ":" + JUNIT_JAR + ":" + EVOSUITE_JAR + ":" + LoopmateConfig.target_class_path
    class_folder = daikon_work_dir_name + "/" + "bin"
    source_files = test_suite_dir_name + "/" + "*.java"
    javac(class_path, class_folder, source_files)
    
    print("Checking successful compilation")
    curr_dir_bin = daikon_work_dir_name + "/" + "bin"
    target_class_file_name = curr_dir_bin + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".class"
    assertFile(target_class_file_name)

    for class_name in class_names:
        junit_class_file_name = curr_dir_bin + "/" + class_name.replace(".", "/") + ".class"
        assertFile(junit_class_file_name)

#=====================================================================================

def replace_in_file(file, pattern, subst):
    # Create temp file
    fh, abs_path = mkstemp()
    new_file = open(abs_path, 'w')
    old_file = open(file)
    for line in old_file:
        new_file.write(line.replace(pattern, subst))
    # close temp file
    new_file.close()
    close(fh)
    old_file.close()
    # Remove original file
    remove(file)
    # Move new file
    move(abs_path, file)    

#=====================================================================================

def print_java_invariants(daikon_work_dir_name, inv_file_name):
    cmd = JAVA_CMD + " -cp " + DAIKON_JAR + " daikon.PrintInvariants " + " --format Java " + inv_file_name
    os.system(cmd)

#=====================================================================================

# Writes the inv_filename into daikon_inv

#=====================================================================================

def print_daikon_invariants_to_file(daikon_inv_filename, inv_file_name):
    cmd = JAVA_CMD + " -cp " + DAIKON_JAR + " daikon.PrintInvariants " + " --format Daikon " + inv_file_name
    
    if LoopmateConfig.verbosity >= 2:
        cmd += " | tee " + daikon_inv_filename
    else:
        cmd += " > " + daikon_inv_filename
    
    
    print("Executing " + cmd)
    os.system(cmd)


#=====================================================================================

def print_java_invariants_to_file(daikon_work_dir_name, java_inv_filename, inv_file_name):
    cmd = JAVA_CMD + " -cp " + DAIKON_JAR + " daikon.PrintInvariants " + " --format Java " + inv_file_name
    
    if LoopmateConfig.verbosity >= 2:
        cmd += " | tee " + java_inv_filename
    else:
        cmd += " > " + java_inv_filename
    
    
    print("Executing " + cmd)
    os.system(cmd)

#=====================================================================================

def print_jml_invariants_to_file(jml_inv_filename, inv_gz_file_name):
    cmd1 = JAVA_CMD + " -cp " + DAIKON_JAR + " daikon.PrintInvariants " + " --format ESCJ_PLUS_JAVA " + inv_gz_file_name 
    print("Executing " + cmd1)
    
    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(cmd1, timeout=NO_TIMEOUT , use_shell=True)    

    jml_inv_file = open(jml_inv_filename, "w")
    jml_inv_file.write(std_out_str)
    jml_inv_file.close()

    if "Exception in thread" in std_err_str:
        print("An error occurred while printing Daikon's ESCJ invariants to a file")
        print("Loopmate is exiting because the list of invariants is not complete")
        raise DaikonError()


#=====================================================================================

def create_runtime_predicate(ginpink_invariant):
    runtime_predicate_lines = list()
    runtime_predicate_lines.append("  try")
    runtime_predicate_lines.append("  {")
    runtime_predicate_lines.append("    if ( !(" + ginpink_invariant.strip() + ") )")
    runtime_predicate_lines.append("  {")
    runtime_predicate_lines.append("    daikon2.Runtime.signalViolation();")
    runtime_predicate_lines.append("  }")
    runtime_predicate_lines.append("  }")
    runtime_predicate_lines.append("  catch (ThreadDeath t_instrument)")
    runtime_predicate_lines.append("  {")
    runtime_predicate_lines.append("    throw t_instrument;")
    runtime_predicate_lines.append("  }")
    runtime_predicate_lines.append("  catch (Throwable t_instrument)")
    runtime_predicate_lines.append("  {")
    runtime_predicate_lines.append("  }")
        
    runtime_predicate_str = "\n".join(runtime_predicate_lines)
    return runtime_predicate_str

#=====================================================================================
def add_predicates_to_internal_loop_invariant(java_file_name, loop_id, runtime_predicates_to_add_str):
    
    new_lines = list()
    java_file = open(java_file_name, "r")
    java_lines = java_file.readlines()
    java_file.close()
    
    loop_signature_prefix = "private static void " + loop_id + "("
    
    i = 0    
    while i < len(java_lines):
        java_line = java_lines[i]
        new_lines.append(java_line)
        i += 1
        if loop_signature_prefix in java_line:
            java_line = java_lines[i]
            i += 1 
            new_lines.append(java_line)
            new_lines.append("  /*ginpink invariants*/" + "\n")
            new_lines.append(runtime_predicates_to_add_str + "\n")
            new_lines.append("  /*daikon invariants*/")
            
    
    new_java_str = "".join(new_lines)
    
    java_file = open(java_file_name, "w")
    java_file.write(new_java_str)
    java_file.close()


#=====================================================================================

def instrument_invariants(daikon_work_dir_name, inv_file_name, ginpink_dir, ginpink_inv_filenames):
    daikon_src_dir = daikon_work_dir_name + "/" + "src"
    target_java_file_name = daikon_src_dir + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    class_path = DAIKON_PREDICATES_JAR + ":" + DAIKON_JAR + ":" + LoopmateConfig.target_class_path

    print("Instrumenting Daikon invariants")
    os.chdir(daikon_work_dir_name)
    exec_cmd = JAVA_CMD + " -cp " + class_path + " daikon.tools.runtimechecker.Main " + " instrument " + inv_file_name + " " + target_java_file_name
    print("Executing: " + exec_cmd)

    os.system(exec_cmd)
    print("Checking successful Daikon instrumentation")
    instrumented_target_java_file_name = daikon_work_dir_name + "/" + "instrumented-classes" + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    assertFile(instrumented_target_java_file_name)


    if len(ginpink_inv_filenames) > 0 and ginpink_dir != None:
        print("Checking if conpilation is succesfful")
    
        print("Instrumenting ginpink invariants")
        ginpink_inv_filename = ginpink_inv_filenames[-1]

        ginpink_instrumented_java_file_name = ginpink_dir + "/" + "instrumented-src" + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
        
        ginpink_instrumented_java_file_dir_name = os.path.dirname(ginpink_instrumented_java_file_name)
        ensure_dir(ginpink_instrumented_java_file_dir_name)
        shutil.copy(instrumented_target_java_file_name, ginpink_instrumented_java_file_name)
        
        ginpink_invariants_by_loop = read_ginpink_java_invariants(ginpink_inv_filename)
        for loop_id in ginpink_invariants_by_loop.keys():
            
            runtime_predicates = list()
            ginpink_invariants = ginpink_invariants_by_loop[loop_id]
            
            for ginpink_invariant in ginpink_invariants:
                runtime_predicate_str = create_runtime_predicate(ginpink_invariant)
                runtime_predicates.append(runtime_predicate_str)
    
            runtime_predicates_str = "\n".join(runtime_predicates)
            add_predicates_to_internal_loop_invariant(ginpink_instrumented_java_file_name, loop_id, runtime_predicates_str)


        instrumented_target_java_file_name = ginpink_instrumented_java_file_name
    
            
    return instrumented_target_java_file_name


#=====================================================================================
# Reads a daikon invariant filename (in daikon format)
# and populates a dictionary loop_id->set(invariants)
#=====================================================================================

def read_daikon_invariants(daikon_invs_filename):
    daikon_invariants_dict = dict()
    lines = open(daikon_invs_filename, 'r').readlines()
    for line in lines:
        if line.strip().endswith(":::ENTER"):
            loop_id = line.split(":::")[0].split("(")[0].split(".")[-1]
            invariants = set()
            daikon_invariants_dict[loop_id] = invariants
            continue 
        
        if line.strip().startswith("========="):
            continue
        
        daikon_inv = line
        invariants.add(daikon_inv)
    return daikon_invariants_dict


#=====================================================================================

def read_jml_invs(invs_filename):
    invs_by_index = dict()
    lines = open(invs_filename, 'r').readlines()
    for line in lines:
        if line.strip().endswith(":::ENTER"):
            loop_invariant_index = line.split(":::")[0].split("(")[0].split(".")[-1]
            invs = set()
            invs_by_index[loop_invariant_index] = invs
            continue 
        
        if line.strip().startswith("========="):
            continue
        
        jml_part = line.split("->")[0]
        invs.add(jml_part)
    return invs_by_index

#=====================================================================================

def get_junit_class_names(test_suite_dir_name, curr_dir_src):
    class_names = list()
    class_name_no_package = LoopmateConfig.target_class_name.split(".")[-1]
    for filename in glob.glob(test_suite_dir_name + "/" + "Test" + class_name_no_package + "*" + ".java"):
        class_name_plus_java = filename[len(curr_dir_src + "/"):]
        class_name = class_name_plus_java.split(".")[0].replace("/", ".")
        class_names.append(class_name)
    
    return class_names

#=====================================================================================

def read_invs_escj_plus_java(escj_plus_java_filename):
    daikon_escj_java_dict = dict()
    lines = open(escj_plus_java_filename, 'r').readlines()
    for line in lines:
        if line.strip().endswith(":::ENTER"):
            loop_id = line.split(":::")[0].split("(")[0].split(".")[-1]
            invariants = set()
            loop_header = LoopHeader()
            loop_header.loop_id = loop_id
            loop_header.signature_string = line.strip()
            daikon_escj_java_dict[loop_header] = invariants
            continue 
        
        if line.strip().startswith("========="):
            continue
        
        if line.strip().startswith("warning"):
            continue
        
        if line.strip() == "":
            continue
        
        escj_part = line.split("->")[0].strip()
        java_part = line.split("->")[1].strip()
        t = escj_part, java_part
        invariants.add(t)
        
    return daikon_escj_java_dict
        
        
        
#=====================================================================================

def isFixPoint(daikon_inv_filenames, ginpink_inv_filenames):
    if len(daikon_inv_filenames) < 2:
        return False
    else:
        
        number_of_filenames = len(daikon_inv_filenames)
        prev_daikon_invs_filename = daikon_inv_filenames[number_of_filenames - 2]
        curr_daikon_invs_filename = daikon_inv_filenames[number_of_filenames - 1]

        prev_daikon_invs = read_daikon_invariants(prev_daikon_invs_filename)
        curr_daikon_invs = read_daikon_invariants(curr_daikon_invs_filename)
        
        if prev_daikon_invs != curr_daikon_invs:
            return False
        
    if len(ginpink_inv_filenames) > 2:

        number_of_filenames = len(ginpink_inv_filenames)
        prev_ginpink_invs_filename = ginpink_inv_filenames[number_of_filenames - 2]
        curr_ginpink_invs_filename = ginpink_inv_filenames[number_of_filenames - 1]
        
        prev_ginpink_invs = read_ginpink_jml_invariants(prev_ginpink_invs_filename)
        curr_ginpink_invs = read_ginpink_jml_invariants(curr_ginpink_invs_filename)
        
        
        if prev_ginpink_invs != curr_ginpink_invs:
            return False

    return True

#=====================================================================================
        
def write_escj_file(escj_template_filename, escj_work_dir_name, daikon_loop_invariants, ginpink_loop_invariants):
    jml_str = " ".join(open(escj_template_filename, "r").readlines())
    for loop_id in daikon_loop_invariants:
        jml_loop_inv_annotations_str = daikon_loop_invariants[loop_id]
        jml_loop_inv_annotations = jml_loop_inv_annotations_str.split("\n")

        daikon_annotations = list()
        for jml_loop_inv_annotation in jml_loop_inv_annotations:
            if len(jml_loop_inv_annotation.strip()) > 0:
                jml_loop_inv_annotation = jml_loop_inv_annotation + " // " + loop_id
                daikon_annotations.append(jml_loop_inv_annotation)

        # loop invariant candidates are sorted to avoid non-deterministic choices of triggers at ESCJ level (i.e. bubble_sort)
        daikon_annotations.sort()
        
        ginpink_annotations = list()
        if loop_id in ginpink_loop_invariants.keys():
            ginpink_loop_invariants_str = ginpink_loop_invariants[loop_id]
            ginpink_loop_invariants_lines = ginpink_loop_invariants_str.split("\n")
            for ginpink_loop_invariant in ginpink_loop_invariants_lines:
                if len(ginpink_loop_invariant.strip()) > 0:
                    ginpink_annotation = ginpink_loop_invariant.strip() + " // " + loop_id
                    ginpink_annotations.append(ginpink_annotation)
        
            ginpink_annotations.sort()
        
        annotations = list()
        annotations.extend(daikon_annotations)
        annotations.extend(ginpink_annotations)
            
        jml_loop_inv_annotations_str = "\n".join(annotations)
        
        jml_str_lines = jml_str.split("\n") 
        new_jml_str = ""
        for jml_str_line in jml_str_lines:
            pattern = "//$$" + loop_id + "$$"
            if pattern in jml_str_line: 
                new_jml_str += jml_loop_inv_annotations_str + '\n'
            else:
                new_jml_str += jml_str_line + '\n'
        
        jml_str = new_jml_str
        
    # remove missing $$ markers in C file
    jml_str = re.sub(re.escape("//$$" + LoopmateConfig.target_method_name + "_loop_invariant_") + "(.*)" + re.escape("$$"), "", jml_str)
    
    escj_filename = escj_work_dir_name + "/" + "src" + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    escj_dir_name = os.path.dirname(escj_filename)
    ensure_dir(escj_dir_name)
    escj_file = open(escj_filename, "w")
    escj_file.write(jml_str)
    escj_file.close()
    return escj_filename


#=====================================================================================

def execute_escj(escj_work_dir_name, class_to_check, routine_to_check, Typecheck=False, only_loop_inv_check=False):
    
    start_timestamp = datetime.datetime.now()

    escj_source_path = escj_work_dir_name + "/" + "src"
    
    cmd = ESCJ_CMD 
    cmd += " -SourcePath " + escj_source_path + ":" + LoopmateConfig.target_escj_source_path
    cmd += " -LoopSafe " 
    cmd += " -Routine " + routine_to_check
    cmd += " -class " + class_to_check
    cmd += " -VCLimit " + str(LoopmateConfig.escj_vc_limit)
    if Typecheck == True:
        cmd += " -Typecheck "
    if only_loop_inv_check == True:
        cmd += " -NoWarn All "
        cmd += " -Warn LoopInv "
    
    escj_out_file = escj_work_dir_name + "/" + "escj_output.txt"
    
    print("Changing JVM to Java 1.5")
    changeJVM(LoopmateConfig.java5_home)

    print("Executing command " + cmd)
    stdout, stderr, timeout_ret_code = run_subprocess_timeout(cmd.split(" ") , timeout=LoopmateConfig.escj_hard_timeout, stop_flag_keyword="ESCJ has finished")

    print("Restoring JVM to Java 1.6")
    changeJVM(LoopmateConfig.java6_home)

    escj_output_file = open(escj_out_file, "w")
    escj_output_file.write(stdout)
    escj_output_file.write(stderr)
    if timeout_ret_code == True:
        escj_output_file.write("ESCJ timeout reached.\n")

    escj_output_file.close()
 
    escj_output_str = "\n".join(open(escj_out_file, "r").readlines())

    end_timestamp = datetime.datetime.now()
    execution_time = end_timestamp - start_timestamp
    Stats.escj_time += execution_time
    print("ESCJ execution time " + str(execution_time))
       
            
    if Typecheck == False and LoopmateConfig.stop_if_escj_reports_error == True and ESCJResult(escj_output_str).is_error() and not ESCJResult(escj_output_str).is_loop_inv_fail():
        print("An error was reported by ESCJ/Java2 in output " + escj_output_str)
        print("Loopmate is exiting because an Error was reported in the ESC/Java2 output.")
        raise ESCJError()            
            
    return escj_out_file, escj_output_str
#=====================================================================================
class GinpinkInvariant:
    java_string = None
    jml_string = None
    
#=====================================================================================
class LoopHeader:
    signature_string = None
    loop_id = None

#=====================================================================================

def read_ginpink_invariants(ginpink_inv_filename):
    ginpink_invariants = dict()
    lines = open(ginpink_inv_filename, 'r').readlines()
    for line in lines:
        if line.strip().endswith(":::ENTER"):
            loop_id = line.split(":::")[0].split("(")[0].split(".")[-1]
            invariants = list()
            loop_header = LoopHeader()
            loop_header.signature_string = line.strip()
            loop_header.loop_id = loop_id
            ginpink_invariants[loop_header] = invariants
            continue 
        
        if line.strip().startswith("========="):
            continue
        
        if line.strip() == "":
            continue
        
        java_string = line.split("<--exec|JML-->")[0].strip()
        jml_string = line.split("<--exec|JML-->")[1].strip()
        
        ginpink_invariant = GinpinkInvariant()
        ginpink_invariant.java_string = java_string
        ginpink_invariant.jml_string = jml_string
        
        invariants.append(ginpink_invariant)
        
    return ginpink_invariants

#=====================================================================================
def remove_old_expression(jml_string):
    
    if "PYTHONPATH" in os.environ.keys():
        old_python_path = os.environ["PYTHONPATH"]
        print("current PYTHONPATH=" + old_python_path)
        print("Cleaning PYTHONPATH for python2.7 execution")
        os.environ["PYTHONPATH"] = ""
        restore_python_path = True
    else:
        restore_python_path = False
        
    cmd = "python2.7 " + GINPINK_DIR + "/unolder.py --JP " + "\"" + jml_string + "\"" 
    print("Executing command: " + cmd)
    stdout, stderr, timeout_ret_code = run_subprocess_timeout(cmd, NO_TIMEOUT, use_shell=True)
    
    if restore_python_path:
        print("Restoring PYTHONPATH to " + old_python_path)
        os.environ["PYTHONPATH"] = old_python_path
        
    for line in stdout.split("\n"):
        if "SUCCESS:" in line:
            ret_val_jml_string = line.split("SUCCESS:")[1].strip()
            return ret_val_jml_string
        elif "FAILURE:" in line:
            print("Warning: failure unoldering expression " + jml_string)
            return jml_string

    return jml_string

#=====================================================================================

def read_ginpink_jml_invariants(ginpink_inv_filename):
    invs_by_index = dict()
    lines = open(ginpink_inv_filename, 'r').readlines()
    for line in lines:
        if line.strip().endswith(":::ENTER"):
            loop_invariant_index = line.split(":::")[0].split("(")[0].split(".")[-1]
            invs = set()
            invs_by_index[loop_invariant_index] = invs
            continue 
        
        if line.strip().startswith("========="):
            continue
        
        java_string = line.split("<--exec|JML-->")[0]
        jml_string = line.split("<--exec|JML-->")[1]
        
        invs.add(jml_string)
        
    return invs_by_index

#=====================================================================================

def read_ginpink_java_invariants(ginpink_inv_filename):
    invs_by_index = dict()
    lines = open(ginpink_inv_filename, 'r').readlines()
    for line in lines:
        if line.strip().endswith(":::ENTER"):
            loop_invariant_index = line.split(":::")[0].split("(")[0].split(".")[-1]
            invs = set()
            invs_by_index[loop_invariant_index] = invs
            continue 
        
        if line.strip().startswith("========="):
            continue
        
        java_string = line.split("<--exec|JML-->")[0]
        jml_string = line.split("<--exec|JML-->")[1]
        
        invs.add(java_string)
        
    return invs_by_index
#=====================================================================================

def prepare_escj(daikon_work_dir_name, escj_work_dir_name, iterCount, daikon_inv_file_name, ginpink_inv_filenames):
    ensure_dir(escj_work_dir_name)     
    
    jml_invs_filename = escj_work_dir_name + "/" + "daikon_invs" + str(iterCount) + ".txt"
    print_jml_invariants_to_file(jml_invs_filename, daikon_inv_file_name)
    daikon_jml_loop_invs_dict = read_jml_invs(jml_invs_filename)

    ignored_loop_invs = list()

    jml_loop_invariant_counter = 0

    if LoopmateConfig.verbosity >= 2:
        print("")
        print("Printing inferred JML loop invariants")
        print("=====================================")
    
    daikon_jml_loop_invs_by_index = dict()
    for jml_inv_index in daikon_jml_loop_invs_dict:
        if LoopmateConfig.verbosity >= 2:
            print("Loop invariants for program point:" + jml_inv_index)
            print("===================================================")
        
        jml_loop_invs = daikon_jml_loop_invs_dict[jml_inv_index]
        jml_loop_inv_annotations = ""
            
        for jml_loop_inv_with_prefix in jml_loop_invs:
            if jml_loop_inv_with_prefix.strip().startswith("warning:"):
                continue
               
            # pow function not supported by JML
            if "java.lang.Math.pow(" in jml_loop_inv_with_prefix:
                ignored_loop_invs.append(jml_loop_inv_with_prefix)
                continue
            

            jml_loop_inv_no_prefix = jml_loop_inv_with_prefix.split()[1:]
            jml_loop_inv = ' '.join(jml_loop_inv_no_prefix)
            jml_loop_inv_annotation = "  //@loop_invariant " + jml_loop_inv + ";"
            
            if VerifiedInvariants.is_verified_invariant(jml_inv_index, jml_loop_inv):
                jml_loop_inv_annotation += "//@nowarn"
            
            jml_loop_inv_annotations += jml_loop_inv_annotation + "\n"
            
            if LoopmateConfig.verbosity >= 2:
                print(jml_loop_inv_annotation)
            
            jml_loop_invariant_counter += 1

        daikon_jml_loop_invs_by_index[jml_inv_index] = jml_loop_inv_annotations
                        
    if LoopmateConfig.verbosity >= 2:
        if len(ignored_loop_invs) > 0:
            print("Ignored loop invariants")
            print("=======================")
            for ignored_loop_inv in ignored_loop_invs:
                print(ignored_loop_inv)
    
    print("Number of JML loop invariants: " + str(jml_loop_invariant_counter))
    print("Number of ignored JML loop invariants: " + str(len(ignored_loop_invs)))
            
    if len(ginpink_inv_filenames) > 0:
        ginpink_inv_filename = ginpink_inv_filenames[-1]
        ginpink_jml_loop_invs_by_index = read_ginpink_jml_invariants(ginpink_inv_filename)
        ginpink_loop_invariants = dict()
        
        for loop_id in ginpink_jml_loop_invs_by_index.keys():
            ginpink_invariants = ginpink_jml_loop_invs_by_index[loop_id]
            ginpink_annotations = ""
            for ginpink_invariant in ginpink_invariants:
                ginpink_loop_inv_annotation = "  //@loop_invariant " + ginpink_invariant.strip() + ";"

                if VerifiedInvariants.is_verified_invariant(loop_id, ginpink_invariant.strip()):
                    ginpink_loop_inv_annotation += "//@nowarn"

                ginpink_annotations += ginpink_loop_inv_annotation + "\n"
            
            ginpink_loop_invariants[loop_id] = ginpink_annotations
                
        
        
    else:
        ginpink_loop_invariants = dict()
            
    
    print("Creating JML input file")
    jml_template_file = LoopmateConfig.target_escj_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    jml_filename = write_escj_file(jml_template_file, escj_work_dir_name, daikon_jml_loop_invs_by_index, ginpink_loop_invariants) 

    if LoopmateConfig.use_frame_loop_invariants == True:
        print("Enabling frame loop invariants (if any)")
        pattern = "//frame_loop_invariant "
        subst = "//@loop_invariant "
        replace_in_file(jml_filename, pattern, subst)
    
    return jml_invs_filename, daikon_jml_loop_invs_by_index, jml_filename

#=====================================================================================

class ESCJResult:
    
    escj_result_str = None
    
    def __init__(self, escj_out_str):
        self.escj_result_str = escj_out_str
        pass
    
    # Cautions: VC too big, PROVER_CC_LIMIT limit reached
    def has_caution(self):
        if "Caution:"in self.escj_result_str or "caution" in self.escj_result_str:
            return True
        else:
            return False
    
    def is_timeout(self):
        if "ESCJ timeout reached" in self.escj_result_str or "Prover error: Unexpected prover response: timeout" in self.escj_result_str or "completely because too much time required" in self.escj_result_str:
            return True
        else:
            return False        

    def is_vc_too_big(self):
        if "VC too big" in self.escj_result_str or "java.lang.OutOfMemoryError: Java heap space" in self.escj_result_str:
            return True
        else:
            return False

    def is_success(self):
        if not "warning" in  self.escj_result_str and not "error" in  self.escj_result_str and "passed" in self.escj_result_str and not "VC too big" in self.escj_result_str :
            return True
        else:
            return False

    def is_loop_inv_fail(self):
        if "(LoopInv)" in self.escj_result_str:
            return True
        else:
            return False

    def is_error(self):
        if "error" in self.escj_result_str and not ("1 error" in self.escj_result_str and "Fatal error: Unexpected exit by Simplify subprocess" in self.escj_result_str):
            return True
        elif "error" in self.escj_result_str and "unexpectedly missing Simplify output" in self.escj_result_str:
            return True
        else:
            return False
        
    def is_prover_cc_limit(self):
        if "completely because warning limit (PROVER_CC_LIMIT) reached" in self.escj_result_str:
            return True
        else:
            return False
    
    def is_loop_failure(self):
        if "(IndexNegative)" in self.escj_result_str or "(IndexTooBig)" in self.escj_result_str or "(Pre)" in  self.escj_result_str:
            return True
        else:
            return False
        
    def get_warnings(self):
        return self.escj_result_str.count("Warning:")
    
#=====================================================================================
def print_jml_invs_dict():
    verified_loop_invariants_dic = VerifiedInvariants.verified_invariants
    jml_loop_invs_dict = discard_java_part(verified_loop_invariants_dic)
    print_jml_invs_real_dict(jml_loop_invs_dict)
    
#=====================================================================================

def print_jml_invs_file(jml_invs_filename):

    verified_loop_invariants_dict = read_invs_escj_plus_java(jml_invs_filename)
    jml_loop_invs_dict = discard_java_part(verified_loop_invariants_dict)
    print_jml_invs_real_dict(jml_loop_invs_dict)

#=====================================================================================

def print_jml_invs_real_dict(jml_invariants_dict):
    
    jml_loop_invariant_counter = 0

    print("Printing last inferred ESCJ loop invariants")
    print("=====================================")
    for loop_header in jml_invariants_dict:
        print("Loop invariants for program point:" + loop_header.loop_id)
        print("===================================================")
        jml_loop_invs = jml_invariants_dict[loop_header]
        print(jml_loop_invs)
        jml_loop_invariant_counter += jml_loop_invs.count("//@loop_invariant")
            
    
    print("Final number of inferred ESCJ loop invariants: " + str(jml_loop_invariant_counter))
    print("")

#=====================================================================================
def has_more_ginpink_mutations(curr_ginpink_mutation):
    next_ginpink_mutation = curr_ginpink_mutation + 1

    source_path = LoopmateConfig.target_java_source_path
    class_name = LoopmateConfig.target_class_name
    
    if "." in  class_name:
        package_name = ".".join(class_name.split(".")[:-1])
        package_dir = source_path + "/" + package_name.replace(".", "/") 
    else:
        package_name = ""
        package_dir = source_path 
    
    ginpink_mutation_filename = package_dir + "/" + class_name + "." + LoopmateConfig.target_method_name + ".iteration_" + str(next_ginpink_mutation) + ".mutations"

    if os.path.exists(ginpink_mutation_filename):
        return True
    else:
        return False

#=====================================================================================

def print_stats():
    print("")
    print("Statistics")
    print("==========")
    print("initial_number_of_warnings=" + str(Stats.initial_number_of_warnings))
    print("number_of_verified_loop_invariants=" + str(VerifiedInvariants.get_number_of_verified_loop_invariants()))
    print("minimum_number_of_warnings=" + str(Stats.minimum_number_of_warnings))
    print("----------")
    total_time = Stats.evosuite_time + Stats.daikon_time + Stats.ginpink_time + Stats.escj_time
    print("Number of Daikon iterations= " + str(Stats.nrOfDaikonIterations))
    print("Number of Ginpink iterations= " + str(Stats.nrOfGinpinkIterations))
    print("Highest Ginpink Wave= " + str(Stats.highestGinpinkWave))
    print("----------")
    print("Total time= " + str(Stats.total_time))
    print("Execution time= " + "{0:.0f}%".format(float(total_time.total_seconds() / Stats.total_time.total_seconds()) * 100))
    print("Evosuite time= " + "{0:.0f}%".format(float(Stats.evosuite_time.total_seconds() / total_time.total_seconds()) * 100))
    print("Daikon time= " + "{0:.0f}%".format(float(Stats.daikon_time.total_seconds() / total_time.total_seconds()) * 100))
    print("Ginpink time= " + "{0:.0f}%".format(float(Stats.ginpink_time.total_seconds() / total_time.total_seconds()) * 100))
    print("ESCJ time= " + "{0:.0f}%".format(float(Stats.escj_time.total_seconds() / total_time.total_seconds()) * 100)) 
    print("==========")
    

#=====================================================================================

def clean_previous_dirs():
    if os.path.exists(LoopmateConfig.work_dir + "/" + "evosuite"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "evosuite")
        
    if os.path.exists(LoopmateConfig.work_dir + "/" + "daikon"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "daikon")

    if os.path.exists(LoopmateConfig.work_dir + "/" + "escj"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "escj")

    if os.path.exists(LoopmateConfig.work_dir + "/" + "ginpink"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "ginpink")

    if os.path.exists(LoopmateConfig.work_dir + "/" + "ginpink-new"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "ginpink-new")

    if os.path.exists(LoopmateConfig.work_dir + "/" + "evosuite-tauto"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "evosuite-tauto")

    if os.path.exists(LoopmateConfig.work_dir + "/" + "escj-tauto"):
        shutil.rmtree(LoopmateConfig.work_dir + "/" + "escj-tauto")

#=====================================================================================


def discard_java_part(verified_loop_invariants_dic):
    jml_invariants_dict = dict()
    for loop_header, tuples_escj_plus_java in verified_loop_invariants_dic.items():
        escj_annotations_str = ""
        for pair_escj_plus_java in tuples_escj_plus_java:
            escj_part = pair_escj_plus_java[0]
            if "::" in escj_part:
                escj_loop_inv_no_prefix = escj_part.split("::")[1].strip()
            else:
                escj_loop_inv_no_prefix = escj_part.strip()
            
            escj_loop_inv = escj_loop_inv_no_prefix
            escj_loop_inv_annotation = "  //@loop_invariant " + escj_loop_inv + ";"
            escj_annotations_str += escj_loop_inv_annotation + "\n"
        
        jml_invariants_dict[loop_header] = escj_annotations_str

    return jml_invariants_dict

#=====================================================================================

def execute_houdini(escj_work_dir_name, daikon_invs_dict, daikon_invs_filename, ginpink_inv_filenames, original_jml_output_str):
    
    loop_invariant_verification_counter = -1
    
    daikon_invariants = read_invs_escj_plus_java(daikon_invs_filename)
    
    if len(ginpink_inv_filenames) > 0:
        ginpink_invariants = read_ginpink_invariants(ginpink_inv_filenames[-1])
    else:
        ginpink_invariants = dict()
        
    if len(ginpink_inv_filenames) > 0:
        ginpink_inv_filename = ginpink_inv_filenames[-1]
    else:
        ginpink_inv_filename = None
        
    if VerifiedInvariants.all_invariants_are_verified(daikon_invariants, ginpink_invariants):
        print("Loop invariant verification: All invariants are known to be verifiable before entering the loop. ESCJ fixpoint reached.")
        # we know all verified invariants are not enough for verification and we reached a fix-point
        return original_jml_output_str, True, dict(), dict()

    
    escj_output_filename = escj_work_dir_name + "/" + "escj_output.txt"
    src_java_input_filename = escj_work_dir_name + "/" + "src" + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    escj_output_str = original_jml_output_str
    
    print("Loop invariant verification: " + escj_output_str)
    
    loop_iteration = 0
    escj_timeout_reached = False
    while True:            
        if ESCJResult(escj_output_str).is_success():
            
            print("Loop invariant verification: We magically need to end everything")
            return escj_output_str, False, daikon_invariants, ginpink_invariants

        elif ESCJResult(escj_output_str).is_error():
            print("Loop invariant verification: ESCJ Error.")
            return escj_output_str, False, dict(), dict()
            
        elif ESCJResult(escj_output_str).is_vc_too_big():
            print("Loop invariant verification: VC too big.")
            return escj_output_str, False, dict(), dict()
        
        elif ESCJResult(escj_output_str).is_timeout() and not ESCJResult(escj_output_str).is_loop_inv_fail(): 
            print("Loop invariant verification: ESCJ timeout and no loop failures.")
            return escj_output_str, False, dict(), dict()
        
        elif not ESCJResult(escj_output_str).is_loop_inv_fail() \
            and not ESCJResult(escj_output_str).is_prover_cc_limit() \
            and not ESCJResult(escj_output_str).is_timeout() \
            and not ESCJResult(escj_output_str).is_vc_too_big():
            
            print("Loop invariant verification: No LoopInv and only Postcondition failures + No Timeouts")
            is_fixpoint = loop_iteration == 0
            return escj_output_str, is_fixpoint, daikon_invariants, ginpink_invariants
    
        else:
            
            if not ESCJResult(escj_output_str).is_loop_inv_fail():
                print("No Loop invariants to remove. ESCJ fixporint reached")
                return escj_output_str, False, dict(), dict()
                
            loop_invariant_verification_counter += 1
            escj_loop_verification_work_dir_name = escj_work_dir_name + "." + str(loop_invariant_verification_counter)
            ensure_dir(escj_loop_verification_work_dir_name)
            
            new_src_java_file = escj_loop_verification_work_dir_name + "/" + "src" + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
            new_daikon_invs_filename = escj_loop_verification_work_dir_name + "/" + "daikon_invs" + str(loop_invariant_verification_counter) + ".txt"
            
            if ginpink_inv_filename == None:
                new_ginpink_invs_filename = None
            else:
                new_ginpink_invs_filename = escj_loop_verification_work_dir_name + "/" + "ginpink_invs" + str(loop_invariant_verification_counter) + ".txt"

            print("Loop invariant verification: Removing candidates that did not verify")    
            # returns a new dicitionary with the remaining invariants
            daikon_invariants, ginpink_invariants = remove_unproved_invariants(src_java_input_filename, \
                                                                               daikon_invs_filename, \
                                                                               escj_output_filename, \
                                                                               new_src_java_file, \
                                                                               new_daikon_invs_filename, \
                                                                               LoopmateConfig.target_method_name, \
                                                                               ginpink_inv_filename, \
                                                                               new_ginpink_invs_filename)
            
            daikon_invs_filename = new_daikon_invs_filename  # new daikon_insXX.txt
            ginpink_inv_filename = new_ginpink_invs_filename  # new ginpink_invsXX.txt
            src_java_input_filename = new_src_java_file  # new case_study.java 

            # if all invariants are known to be verified return without another call to ESCJ - return empty dictionary - no new invariants to add
            if VerifiedInvariants.all_invariants_are_verified(daikon_invariants, ginpink_invariants) and not (escj_timeout_reached == True):
                print("Loop invariant verification: All invariants are known to be verifiable. Skipping ESCJ execution.")
                # we definitely know that we can not verify the postcondition because we tried before
                return original_jml_output_str, False, dict(), dict()
    
            print("Loop invariant verification: Executing ESCJ once more")
            escj_output_filename, escj_output_str = execute_escj(escj_loop_verification_work_dir_name, class_to_check=LoopmateConfig.target_class_name, routine_to_check=LoopmateConfig.target_method_name)

            
            print("Loop invariant verification: " + escj_output_str)
            escj_timeout_reached = ESCJResult(escj_output_str).is_timeout
            
        loop_iteration = loop_iteration + 1

#=====================================================================================
def get_line_from_file(filename, line_number):
    file = open(filename, "r")
    lines = file.readlines()
    line = lines[line_number - 1]
    return line
#=====================================================================================

def escj_typecheck_ginpink_mutations2(ginpink_work_dir_name, ginpink_inv_filename):
    
    escj_temp_dir = ginpink_work_dir_name + "/" + "src"
     
    orig_escj_filename = LoopmateConfig.target_escj_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dest_escj_filename = escj_temp_dir + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dest_dir = os.path.dirname(dest_escj_filename)
    ensure_dir(dest_dir)
    
    ginpink_invariants = read_ginpink_invariants(ginpink_inv_filename)

    while True:
        shutil.copy(orig_escj_filename, dest_dir)
        for loop_header in ginpink_invariants.keys():
            curr_loop_ginpink_invariants = ginpink_invariants[loop_header]
            pattern = "//$$" + loop_header.loop_id + "$$"
            subst = list()
            for i in range(len(curr_loop_ginpink_invariants)):
                ginpink_invariant = curr_loop_ginpink_invariants[i]
                subst.append("//@loop_invariant " + ginpink_invariant.jml_string + "; //loop_id:" + loop_header.loop_id + " position:" + str(i))
            
            replace_in_file(dest_escj_filename, pattern, "\n".join(subst))
            
        print("Typechecking JML invariants ")
        escj_out_file, escj_output_str = execute_escj(ginpink_work_dir_name, Typecheck=True, class_to_check=LoopmateConfig.target_class_name, routine_to_check=LoopmateConfig.target_method_name)
            
            
        if ESCJResult(escj_output_str).is_error():
            to_remove = dict()
            escj_output_lines = escj_output_str.split("\n")
            for escj_output_line in escj_output_lines:
                if ":Error :" in escj_output_line or ": Fatal error:" in escj_output_line or " Error:" in escj_output_line:
                    error_line_number = int(escj_output_line.split(":")[1])
                    error_line = get_line_from_file(dest_escj_filename, error_line_number)
                    loop_id_plus_position = error_line.split("; //loop_id:")[1]
                    loop_id = loop_id_plus_position.split(" position:")[0]
                    print("Ginpink Invariant " + error_line + " does not typecheck at loop: " + loop_id)
                    if not loop_id in to_remove.keys():
                        to_remove[loop_id] = list()
                    position = int(loop_id_plus_position.split(" position:")[1])
                    if not position in to_remove[loop_id]:
                        to_remove[loop_id].append(position)
                
            if len(to_remove.keys()) == 0:
                break
                
            for loop_id in to_remove.keys():
                to_remove[loop_id] = sorted(to_remove[loop_id])
                to_remove[loop_id].reverse()
                
                for loop_header in ginpink_invariants.keys():
                    if loop_header.loop_id == loop_id:
                        curr_loop_ginpink_invariants = ginpink_invariants[loop_header]
                        for del_index in to_remove[loop_id]:
                            print("Deleting index " + str(del_index) + " in loop invariant " + loop_id)
                            print("Current ginpink invariant " + curr_loop_ginpink_invariants[del_index].jml_string)
                            del curr_loop_ginpink_invariants[del_index]
                            print("New number of loop invariants for " + loop_id + " is " + str(len(curr_loop_ginpink_invariants)))
                
        else:
            print("All remaining ginpink invariants typechecked!")
            break

    return ginpink_invariants


#=====================================================================================

def escj_typecheck_ginpink_mutations(ginpink_work_dir_name, ginpink_inv_filename):
    
    escj_temp_dir = ginpink_work_dir_name + "/" + "src"
     
    orig_escj_filename = LoopmateConfig.target_escj_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dest_escj_filename = escj_temp_dir + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dest_dir = os.path.dirname(dest_escj_filename)
    ensure_dir(dest_dir)
    
    ginpink_invariants = read_ginpink_invariants(ginpink_inv_filename)
    typechecked_ginpink_invariants = dict()
    for loop_header in ginpink_invariants.keys():
        curr_loop_ginpink_invariants = ginpink_invariants[loop_header]
        typechecked_ginpink_invariants[loop_header] = list()
        for ginpink_invariant in curr_loop_ginpink_invariants:
            shutil.copy(orig_escj_filename, dest_dir)
            pattern = "//$$" + loop_header.loop_id + "$$"
            subst = "//@loop_invariant " + ginpink_invariant.jml_string + ";"
            replace_in_file(dest_escj_filename, pattern, subst)
            
            print("Typechecking JML invariant " + ginpink_invariant.jml_string)
            escj_out_file, escj_output_str = execute_escj(ginpink_work_dir_name, Typecheck=True, class_to_check=LoopmateConfig.target_class_name, routine_to_check=LoopmateConfig.target_method_name)
            
            if ESCJResult(escj_output_str).is_error():
                print("Ginpink Invariant " + ginpink_invariant.jml_string + " does not typecheck at loop: " + loop_header.loop_id)
            else:
                typechecked_ginpink_invariants[loop_header].append(ginpink_invariant)

    print("ESCJ typechecking")
    print("-----------------")
    for loop_header in ginpink_invariants.keys():
        orig_len = len(ginpink_invariants[loop_header])
        typecheck_len = len(typechecked_ginpink_invariants[loop_header])
        diff_str = "All" if orig_len == typecheck_len else "None" if typecheck_len == 0 else "Diff=" + str(orig_len - typecheck_len)
        print(loop_header.loop_id + " : " + str(orig_len) + "->" + str(typecheck_len) + " (" + diff_str + ")")
    
    return typechecked_ginpink_invariants


#=====================================================================================

def print_java_version():
    print("JVM Version:")
    os.system(JAVA_CMD + " -version")
    print("")
    
    print("Java Compiler Version:")
    os.system(JAVAC_CMD + " -version")
    print("")


def print_iteration(iterCount):
    print("================================================================================================================")
    print ("Entering iteration: " + str(iterCount))
    print("================================================================================================================")

    
def get_iteration_name(iterCount):
    return "iter" + str(iterCount)

#=====================================================================================
def create_invocation_from_declaration(method_declaration):
    ret_val = method_declaration.replace("[] ", "")
    ret_val = ret_val.replace("int", "")
    ret_val = ret_val.replace("boolean", "")
    ret_val = ret_val.replace("java3.lang3.Integer", "")
    ret_val = ret_val.replace("private", "")
    ret_val = ret_val.replace("static", "")
    ret_val = ret_val.replace("void", "")
    ret_val = ret_val.replace("ArrayList", "")
    ret_val = ret_val.replace("Vector", "")
    ret_val = ret_val.replace("ArrayDeque", "")
    ret_val = ret_val.replace("{", "")
    return ret_val
    
#=====================================================================================
def add_verified_invariants_checks(java_file_name):
    java_out_lines = list()
    java_in_lines = open(java_file_name, "r").readlines()
    line_number = 0
    while line_number < len(java_in_lines):
        java_line = java_in_lines[line_number]
        line_number += 1        
        java_out_lines.append(java_line)
        for loop_header in VerifiedInvariants.verified_invariants.keys():
            loop_id = loop_header.loop_id
            if "private static void " + loop_id + "(" in java_line:
                java_out_lines = java_out_lines[:-1]
                method_declaration = java_line
                while not "{"  in java_line:
                    java_line = java_in_lines[line_number]
                    line_number += 1
                    method_declaration += java_line
                    
                checked_method_declaration = method_declaration.replace("private", "public").replace(loop_id, "external$" + loop_id)
                java_out_lines.append(checked_method_declaration)                
                java_out_lines.append("   /*Begin Loop Invariants Checks*/" + "\n")
                for loop_invariant in VerifiedInvariants.verified_invariants[loop_header]:
                    java_part = loop_invariant[1].strip()
                    java_out_lines.append("  if (!(" + java_part + ")) {" + "\n")
                    java_out_lines.append("    return;" + "\n")
                    java_out_lines.append("  }" + "\n")

                java_out_lines.append("   /*End Loop Invariants Checks*/" + "\n")

                loop_method_invocation = create_invocation_from_declaration(method_declaration)
                java_out_lines.append("   " + loop_method_invocation.strip() + ";" + "\n")
                java_out_lines.append("   }" + "\n")
                java_out_lines.append("\n")

                java_out_lines.append(method_declaration)
            

    java_file = open(java_file_name, "w")
    for java_line in java_out_lines:
        java_file.write(java_line)
        
    java_file.close()
    
    pass

#=====================================================================================

def remove_evosuite_tautologies(work_dir_name, daikon_inv_gz_filename, ginpink_invariant_filename, iterCount):
    ensure_dir(work_dir_name)
    
    ginpink_invariants = read_ginpink_invariants(ginpink_invariant_filename)

    ginpink_inv_filenames = list()
    ginpink_inv_filenames.append(ginpink_invariant_filename)
    
    work_src_dir = work_dir_name + "/" + "src"
    target_java_file_name = work_src_dir + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    
    dir_name = os.path.dirname(target_java_file_name)
    ensure_dir(dir_name)
    
    shutil.copy(LoopmateConfig.target_java_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java", target_java_file_name)
    
    src_target_java_file_name = instrument_invariants(work_dir_name, daikon_inv_gz_filename, work_dir_name, ginpink_inv_filenames)
    shutil.copy(src_target_java_file_name, target_java_file_name)
    
    print("Adding verified loop invariants as runtime checks to instrumenting methods")
    add_verified_invariants_checks(target_java_file_name)
    
    junit_file_name = execute_evosuite(target_java_file_name, work_dir_name, iterCount)
    
    print("Compiling target/helper for Daikon")
    evosuite_tauto_bin_dir = work_dir_name + "/" + "bin"
    class_path = evosuite_tauto_bin_dir + ":" + DAIKON_JAR + ":" + JUNIT_JAR + ":" + EVOSUITE_JAR + ":" + LoopmateConfig.target_class_path
    javac(class_path, evosuite_tauto_bin_dir, junit_file_name)    
    
    print("Executing Chicory for removing EvoSuite tautologies (no std-visibility) and no purity filename")
    ginpink_dtrace_file = execute_chicory(work_dir_name, evosuite_tauto_bin_dir, junit_file_name, False, None)
    ginpink_dtrace_filenames = list()
    ginpink_dtrace_filenames.append(ginpink_dtrace_file)
                
    print("Executing Ginpink Checker on new ginpink candidates")
    ginpink_inv_filename = execute_ginpink(work_dir_name, ginpink_invariant_filename, ginpink_dtrace_filenames, iterCount, print_falsified_candidates=True)

    return ginpink_inv_filename
    
    
#=====================================================================================
def find_loop_method_declaration(loop_id):
    java_src_filename = LoopmateConfig.target_java_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    java_lines = open(java_src_filename, "r").readlines()
    for line_number in range(len(java_lines)):
        java_line = java_lines[line_number]
        if "private static void " + loop_id + "(" in java_line:
            if "{" in java_line:
                return java_line
            else:
                ret_val = java_line
                while not "{"  in java_line:
                    line_number += 1
                    java_line = java_lines[line_number]
                    ret_val += java_line
                return ret_val
        
    return None
#=====================================================================================
def get_import_list(escj_filename):
    import_list = list()
    for java_line in open(escj_filename, "r").readlines():
        if java_line.startswith("import "):
            import_list.append(java_line.strip())
    
    return import_list

#=====================================================================================

def remove_escj_tautologies(ginpink_work_dir_name, ginpink_invariant_filename, iterCount):
    ginpink_invariants = read_ginpink_invariants(ginpink_invariant_filename)
    
    filtered_invariants = dict()
    total_read_invariants = 0
    removeCounter = 0
    ignore_counter = 0
    
    new_ginpink_invariant_filename = ginpink_work_dir_name + "/" + "ginpink_invs" + str(iterCount) + ".txt"

        
    for loop_header in ginpink_invariants.keys():
        curr_ginpink_invariants = ginpink_invariants[loop_header]
        filtered_invariants[loop_header] = list()
        for ginpink_invariant in curr_ginpink_invariants:

            if VerifiedInvariants.is_verified_invariant(loop_header.loop_id, ginpink_invariant.jml_string):
                print("Ignoring already verified loop invariant " + ginpink_invariant.jml_string)
                filtered_invariants[loop_header].append(ginpink_invariant)
                ignore_counter += 1
                continue

            total_read_invariants += 1

            if "." in  LoopmateConfig.target_class_name:
                package_name = ".".join(LoopmateConfig.target_class_name.split(".")[:-1])
            else:
                package_name = None

            tchecker_src = list()
            if package_name != None:
                tchecker_src.append("package " + package_name + ";")
            
            escj_filename = LoopmateConfig.target_escj_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
            import_list = get_import_list(escj_filename)
            
            for import_stmt in import_list:
                tchecker_src.append(import_stmt)
            
            tchecker_src.append("")
            tchecker_src.append("public class TChecker {")
            
            loop_method_declaration = find_loop_method_declaration(loop_header.loop_id)
            tchecker_src.append(loop_method_declaration)
            
            for verified_loop_header in VerifiedInvariants.verified_invariants.keys():
                if verified_loop_header.loop_id == loop_header.loop_id:
                    loop_invariants = VerifiedInvariants.verified_invariants[verified_loop_header]
                    for loop_invariant in loop_invariants:
                        name_colon_jml_part = loop_invariant[0]
                        if "::" in name_colon_jml_part:
                            jml_part = name_colon_jml_part.split("::")[1]
                        else:
                            jml_part = name_colon_jml_part
                            
                        if "\\old" in jml_part:
                            jml_part = remove_old_expression(jml_part)
                            
                        tchecker_src.append("    //@assume  " + jml_part + ";")
                
            tchecker_src.append("")
            
            jml_part = ginpink_invariant.jml_string
            if "\\old" in jml_part:
                jml_part = remove_old_expression(jml_part)
                
            tchecker_src.append("    //@assert " + jml_part + ";")
            tchecker_src.append("    ")
            tchecker_src.append("  }")
            tchecker_src.append("}")

            check_work_dir_name = ginpink_work_dir_name + "/" + "check" + str(total_read_invariants)
            if "." in  LoopmateConfig.target_class_name:
                package_dir = check_work_dir_name + "/" + "src" + "/" + package_name.replace(".", "/") 
            else:
                package_dir = check_work_dir_name + "/" + "src"

            ensure_dir(package_dir)
            tchecker_file = open(package_dir + "/" + "TChecker.java", "w")
            tchecker_file.write("\n".join(tchecker_src))
            tchecker_file.close()
            
            if package_name == None:
                checker_classname = "TChecker"
            else:
                checker_classname = package_name + "." + "TChecker"

            escj_out_file, escj_output_str = execute_escj(check_work_dir_name, class_to_check=checker_classname, routine_to_check=loop_header.loop_id)
            
            if not "passed" in escj_output_str:
                filtered_invariants[loop_header].append(ginpink_invariant)
            else:
                print("Removing tautology " + ginpink_invariant.jml_string)
                removeCounter += 1
                print(str(removeCounter) + " removed out of " + str(total_read_invariants) + " read invariants (" + "{0:.0f}%".format(float(removeCounter / total_read_invariants) * 100) + ") " + str(ignore_counter) + " candidates already verified loop invariants")
    
    write_ginpink_invariants(filtered_invariants, new_ginpink_invariant_filename)
    return new_ginpink_invariant_filename
            
#=====================================================================================
def write_daikon_invariants(daikon_invariants, daikon_escj_java_filename):
    dir_name = os.path.dirname(daikon_escj_java_filename)
    ensure_dir(dir_name)
    daikon_escj_java_file = open(daikon_escj_java_filename, "w")
    for loop_header in daikon_invariants.keys():
        daikon_escj_java_file.write("===========================================================================")
        daikon_escj_java_file.write("\n")
        daikon_escj_java_file.write(loop_header.signature_string)
        daikon_escj_java_file.write("\n")
        invariants = daikon_invariants[loop_header]
        for jml_string, java_string in invariants:
            daikon_escj_java_file.write(jml_string)
            daikon_escj_java_file.write("->")
            daikon_escj_java_file.write(java_string)
            daikon_escj_java_file.write("\n")
            
    daikon_escj_java_file.close()
#=====================================================================================

#=====================================================================================
def write_ginpink_invariants(ginpink_invariants, ginpink_inv_filename):
    dir_name = os.path.dirname(ginpink_inv_filename)
    ensure_dir(dir_name)
    ginpink_inv_file = open(ginpink_inv_filename, "w")
    for loop_header in ginpink_invariants.keys():
        ginpink_inv_file.write("===========================================================================")
        ginpink_inv_file.write("\n")
        ginpink_inv_file.write(loop_header.signature_string)
        ginpink_inv_file.write("\n")
        invariants = ginpink_invariants[loop_header]
        for invariant in invariants:
            ginpink_inv_file.write(invariant.java_string)
            ginpink_inv_file.write("<--exec|JML-->")
            ginpink_inv_file.write(invariant.jml_string)
            ginpink_inv_file.write("\n")
            
    ginpink_inv_file.close()
#=====================================================================================
def read_ginpink_generation_time(mutations_filename):
    mutations_file = open(mutations_filename, "r")
    for line in mutations_file:
        if line.startswith("-----> Generated "):
            seconds_str = line.split(' ')[-1]
            seconds_float = float(seconds_str)
            return timedelta(seconds=seconds_float)
        elif line.startswith("No mutations generated within a timeout of"):
            seconds_str = line.split(' ')[-2]
            seconds_float = float(seconds_str)
            return timedelta(seconds=seconds_float)
        
    return timedelta(seconds=0)
#=====================================================================================
def read_ginpink_mutations(ginpink_work_dir_name, curr_ginpink_mutation):
    
    ensure_dir(ginpink_work_dir_name)

    source_path = LoopmateConfig.target_java_source_path
    class_name = LoopmateConfig.target_class_name
    
    if "." in  class_name:
        package_name = ".".join(class_name.split(".")[:-1])
        package_dir = source_path + "/" + package_name.replace(".", "/") 
    else:
        package_name = ""
        package_dir = source_path 
    
    mutations_filename = package_dir + "/" + class_name + "." + LoopmateConfig.target_method_name + ".iteration_" + str(curr_ginpink_mutation) + ".mutations"

    
    if not os.path.exists(mutations_filename) :
        print("Unable to find mutation filename  " + mutations_filename + " when ginpink is enabled")
        print("Since we can add ginpink invariants, ginpink feature is useless.")
        print("Loopmate is exiting because ginpink was missing a requited file.")
        raise GinpinkError()        
    
    ginpink_generation_time = read_ginpink_generation_time(mutations_filename)
    print("Ginpink Wave: " + str(curr_ginpink_mutation) + " Ginpink generation time=" + str(ginpink_generation_time))
    Stats.ginpink_time += ginpink_generation_time
    
    start_timestamp = datetime.datetime.now()
    target_class_path = LoopmateConfig.target_class_path
    ginpink_inv_filename = ginpink_work_dir_name + "/" + "ginpink_invs_java_typechecked" + ".iteration_" + str(curr_ginpink_mutation) + ".txt"
    
    ginpink_cmd = JAVA_CMD 
    ginpink_cmd += " -Xmx2g " 
    ginpink_cmd += " -cp " + GINPINK_JAR + ":" + DAIKON_JAR + ":" + target_class_path + " "
    ginpink_cmd += " org.ginpink.ReadMutations " 
    ginpink_cmd += " --class_name " + LoopmateConfig.target_class_name + " "
    ginpink_cmd += " --method_name " + LoopmateConfig.target_method_name + " "
    ginpink_cmd += " --ginpink_inv_file " + ginpink_inv_filename + " "
    ginpink_cmd += " " + mutations_filename + " "

    print("Reading mutations filename " + mutations_filename + " using classpath " + target_class_path)
    print("Saving typechecked invariants to file " + ginpink_inv_filename)
    
    ensure_dir(ginpink_work_dir_name)

    print("Executing Ginpink command " + ginpink_cmd)
    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(ginpink_cmd, timeout=LoopmateConfig.ginpink_hard_timeout , use_shell=True)    

    ginpink_out_filename = ginpink_work_dir_name + "/" + "ginpink_out.txt"
    ginpink_err_filename = ginpink_work_dir_name + "/" + "ginpink_err.txt"
    
    evosuite_out_file = open(ginpink_out_filename, 'w')
    evosuite_out_file.write(std_out_str)
    evosuite_out_file.close()

    evosuite_err_file = open(ginpink_err_filename, 'w')
    evosuite_err_file.write(std_err_str)
    evosuite_err_file.close()

    if timeout_ret_code == True:
        print("Normal Termination: Timeout")
        print("GinPink timeout was reached")
        print("Loopmate is exiting because GinPink timeout has reached in " + str(LoopmateConfig.ginpink_hard_timeout) + " secs.")
        raise GinpinkTimeout()

    assertFile(ginpink_inv_filename)
    
    print("Checking JML format for each new mutation")
    typeched_ginpink_invariants = escj_typecheck_ginpink_mutations2(ginpink_work_dir_name, ginpink_inv_filename)
    typechecked_ginpink_inv_filename = ginpink_work_dir_name + "/" + "ginpink_invs_escj_typechecked" + ".iteration_" + str(curr_ginpink_mutation) + ".txt"
    write_ginpink_invariants(typeched_ginpink_invariants, typechecked_ginpink_inv_filename)
    
    end_timestamp = datetime.datetime.now()
    execution_time = end_timestamp - start_timestamp
    Stats.ginpink_time += execution_time
    print("Ginpink execution time " + str(execution_time))

    
    return typechecked_ginpink_inv_filename
    
#=====================================================================================

def execute_ginpink(ginpink_work_dir_name, previous_ginpink_inv_filename, dtrace_filenames, iterCount, print_falsified_candidates=False):
    
    start_timestamp = datetime.datetime.now()

    if not os.path.exists(previous_ginpink_inv_filename):
        print("Unable to find ginpink invariants filename " + previous_ginpink_inv_filename + " when ginpink is enabled")
        print("Since we can not check ginpink invariants, ginpink feature is useless.")
        print("Loopmate is exiting because a required ginpink file was missing.")
        raise GinpinkError()        
    
    target_class_path = LoopmateConfig.target_class_path
    new_ginpink_inv_filename = ginpink_work_dir_name + "/" + "ginpink_invs" + str(iterCount) + ".txt"
    
    ginpink_cmd = JAVA_CMD 
    ginpink_cmd += " -Xmx2g " 
    ginpink_cmd += " -cp " + GINPINK_JAR + ":" + DAIKON_JAR + ":" + target_class_path + " "
    ginpink_cmd += " org.ginpink.CheckInvariants " 
    ginpink_cmd += " --class_name " + LoopmateConfig.target_class_name + " "
    ginpink_cmd += " --method_name " + LoopmateConfig.target_method_name + " "
    ginpink_cmd += " --ginpink_inv_file " + previous_ginpink_inv_filename + " "
    ginpink_cmd += " --checked_ginpink_inv_file " + new_ginpink_inv_filename + " "
    ginpink_cmd += " --ppt_select_pattern " + LoopmateConfig.target_method_name + "_loop_invariant_" + " " 
    
    if print_falsified_candidates == True:
        discarded_ginpink_inv_filename = ginpink_work_dir_name + "/" + "false_ginpink_invs" + str(iterCount) + ".txt"
        ginpink_cmd += " --discarded_ginpink_inv_file " + discarded_ginpink_inv_filename + " "
    
    ginpink_cmd += "  ".join(dtrace_filenames) + " "

    print("Reading ginpink invariants filename " + previous_ginpink_inv_filename + " using classpath " + target_class_path)
    print("Saving checked invariants to file " + new_ginpink_inv_filename)
    
    print("Executing Ginpink command " + ginpink_cmd)
    std_out_str, std_err_str, timeout_ret_code = run_subprocess_timeout(ginpink_cmd, timeout=LoopmateConfig.ginpink_hard_timeout , use_shell=True)    


    ginpink_out_filename = ginpink_work_dir_name + "/" + "ginpink_out.txt"
    ginpink_err_filename = ginpink_work_dir_name + "/" + "ginpink_err.txt"
    
    evosuite_out_file = open(ginpink_out_filename, 'w')
    evosuite_out_file.write(std_out_str)
    evosuite_out_file.close()

    evosuite_err_file = open(ginpink_err_filename, 'w')
    evosuite_err_file.write(std_err_str)
    evosuite_err_file.close()

    if timeout_ret_code == True:
        print("Normal Termination: Timeout")
        print("GinPink timeout was reached")
        print("Loopmate is exiting because GinPink timeout has reached in " + str(LoopmateConfig.ginpink_hard_timeout) + " secs.")
        raise GinpinkTimeout()

    end_timestamp = datetime.datetime.now()
    execution_time = end_timestamp - start_timestamp
    Stats.ginpink_time += execution_time
    print("Ginpink execution time " + str(execution_time))

    
    if print_falsified_candidates == True:
        assertFile(discarded_ginpink_inv_filename)
        return discarded_ginpink_inv_filename
    else:
        assertFile(new_ginpink_inv_filename)
        return new_ginpink_inv_filename
    
#=====================================================================================
def write_verified_ginpink_invariants_to_file(ginpink_invariants, escj_work_dir_name, iterCount):
    verified_invariants_filename = escj_work_dir_name + "/" + "verified_ginpink_invs" + str(iterCount) + ".txt"
    write_ginpink_invariants(ginpink_invariants, verified_invariants_filename)
#=====================================================================================

def write_verified_daikon_invariants_to_file(daikon_invariants, escj_work_dir_name, iterCount):
    verified_invariants_filename = escj_work_dir_name + "/" + "verified_daikon_invs" + str(iterCount) + ".txt"
    verified_invariants_file = open(verified_invariants_filename, 'w')
    for loop_header in daikon_invariants.keys():
        verified_invariants_file.write("===========================================================================")
        verified_invariants_file.write("\n")
        verified_invariants_file.write(loop_header.signature_string)
        verified_invariants_file.write("\n")
        for jml_plus_java_pair in daikon_invariants[loop_header]:
            verified_invariants_file.write(jml_plus_java_pair[0] + "->" + jml_plus_java_pair[1])
            verified_invariants_file.write("\n")

    verified_invariants_file.close()
    return verified_invariants_filename
#=====================================================================================

def execute_escj_first_time():
    escj_init_dir = LoopmateConfig.work_dir + "/" + "escj" + "/" + "init"
    ensure_dir(escj_init_dir)
    
    escj_init_src = escj_init_dir + "/" + "src"
    ensure_dir(escj_init_src)

    if "." in LoopmateConfig.target_class_name:
        package_dir = escj_init_src + "/" + "/".join(LoopmateConfig.target_class_name.split(".")[:-1])
        ensure_dir(package_dir)
        
    jml_template_file = LoopmateConfig.target_escj_source_path + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    dst_source_file = escj_init_src + "/" + LoopmateConfig.target_class_name.replace(".", "/") + ".java"
    
    shutil.copy(jml_template_file, dst_source_file)        
    
    escj_output_filename, escj_output_str = execute_escj(escj_init_dir, class_to_check=LoopmateConfig.target_class_name, routine_to_check=LoopmateConfig.target_method_name)

    print("ESCJ first time:")
    print(escj_output_str)
    
    escj_result = ESCJResult(escj_output_str)
    
    if escj_result.is_success():
        print("The verification was successfull without loop invariants")
        print("Loopmate is exiting because there is nothing else to be done.")
        raise TrivialVerification()
        
    elif escj_result.is_timeout():
        print("The verification was unsuccessfull because of timeout")
        print("Since no invariants were added, it is most likely this problem is untractable for ESCJ.")
        print("Loopmate is exiting because ESCJ is unable to work with the current input.")
        raise ESCJError()

    elif escj_result.is_vc_too_big():
        print("The verification was unsuccessfull because of VC too big")
        print("Since no invariants were added, it is most likely this problem is untractable for ESCJ.")
        print("Loopmate is exiting because ESCJ is unable to work with the current input.")
        raise ESCJError()
        
    elif escj_result.is_error():
        print("An error happened during verification")
        print("Since no invariants were added, it is most likely the input file is broken.")
        print("Loopmate is exiting because ESCJ is unable to work with the current input.")
        raise ESCJError()

    initial_number_of_warnings = escj_result.get_warnings()
    if initial_number_of_warnings == 0:
        print("First verification led to no warnings")
        print("Since no invariants were added, and verification was flagged as unsuccesfull, it is most likely ESCJ is unable to deal with the input.")
        print("Loopmate is exiting because ESCJ no initial warnings were found.")
        raise ESCJError()

    Stats.set_initial_number_of_warnings(initial_number_of_warnings)
    print("Initial Number of Warnings: " + str(initial_number_of_warnings))

#=====================================================================================

def changeJVM(new_java_home):
    print("Changing JVM")
    
    if not "JAVA_HOME" in os.environ:
        print("No JAVA_HOME environment variable declared")
    else:
        print("Previous JAVA_HOME: " + os.environ["JAVA_HOME"])

    if not "PATH" in os.environ:
        print("No PATH environment variable declared")
    else:
        print("Previous PATH: " + os.environ["PATH"])
    
    print("Setting JAVA_HOME to " + new_java_home)
    os.environ["JAVA_HOME"] = new_java_home
    print("New JAVA_HOME: " + os.environ["JAVA_HOME"])
    
    java_bin = new_java_home + "/" + "bin"
    if LoopmateConfig.orig_path != None:
        new_path = java_bin + ":" + LoopmateConfig.orig_path
    else:
        new_path = java_bin
        
    print("Setting PATH to " + new_path)
    os.environ["PATH"] = new_path
    print("New PATH: " + os.environ["PATH"])

#==========================================================================================
def number_of_invariants(ginpink_inv_dict):
    number_of_invariants = 0
    for loop_id in ginpink_inv_dict.keys():
        number_of_invariants += len(ginpink_inv_dict[loop_id]) 
            
    return number_of_invariants

#==========================================================================================
def is_empty(ginpink_inv_dict):
    if len(ginpink_inv_dict.keys()) == 0:
        return True
    else:
        for loop_id in ginpink_inv_dict.keys():
            if len(ginpink_inv_dict[loop_id]) > 0:
                return False
            
        return True

#=====================================================================================
def merge_ginpink_invariants(ginpink_work_dir_name, left_ginpink_inv_filename, right_ginpink_inv_filename):
    left_invs = read_ginpink_invariants(left_ginpink_inv_filename)
    right_invs = read_ginpink_invariants(right_ginpink_inv_filename)

    merged_ginpink_invariants = dict()
    
    for loop_header in left_invs.keys():
        merged_ginpink_invariants[loop_header] = list(left_invs[loop_header])
    
    for loop_header_right in right_invs.keys():
        found = False
        for loop_header_left in merged_ginpink_invariants.keys():
            if loop_header_right.loop_id == loop_header_left.loop_id:
                found = True
                invariant_list = merged_ginpink_invariants[loop_header_left]
                invariant_list.extend(right_invs[loop_header_right])
                merged_ginpink_invariants[loop_header_left] = invariant_list
                
        if not found:
            merged_ginpink_invariants[loop_header_right] = right_invs[loop_header_right]
            
    merged_ginpink_inv_filename = ginpink_work_dir_name + "/" + "merged_ginpink_invs.txt"
    write_ginpink_invariants(merged_ginpink_invariants, merged_ginpink_inv_filename)
    
    print("Merging ginpink invariants left=" + str(number_of_invariants(left_invs)) + ", right=" + str(number_of_invariants(right_invs)) + ", merged=" + str(number_of_invariants(merged_ginpink_invariants)))
    
    return merged_ginpink_inv_filename


def remove_unproved_invariants(java_src_filename, \
                               daikon_invs_filename, \
                               escj_output_filename, \
                               new_java_src_filename, \
                               new_daikon_invs_filename, \
                               target_method_name, \
                               ginpink_inv_filename, \
                               new_ginpink_inv_filename):
    
    # open files
    escj_output_file = open(escj_output_filename, 'r')
    escj_output = escj_output_file.readlines()
    escj_output_file.close()
    
    dirname = os.path.dirname(new_java_src_filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    
    new_java_file = open(new_java_src_filename, 'w')

    
    java_file = open(java_src_filename, 'r')
    java_file_lines = java_file.readlines()
    java_file.close()
    
    # fill map mapping line number to invariant
    map_line_to_name = {}
    loop_inv_warning = False
    for line in escj_output:
        # split such that one error message in every fragment
        if "Warning: Loop invariant possibly does not hold (LoopInv)" in line:
            
            if loop_inv_warning == True:
                raise Exception
            
            loop_inv_warning = True
        elif loop_inv_warning == True and "Associated declaration is " in line:
            fragments = line.split(",")
            line_number_fragment = fragments[1].strip()
            line_number = int(line_number_fragment[len("line ") - 1:])
            
            invariant = java_file_lines[line_number - 1]
            invariant = invariant.replace("\\\\", "\\")
            begin_pos = invariant.find("loop_invariant ") + len("loop_invariant")
            end_pos = invariant.find("; // " + target_method_name + "_loop_invariant")
            invariant = invariant[begin_pos:end_pos]
            map_line_to_name[line_number] = invariant.strip()
            loop_inv_warning = False
        
        
    # write new cpp file and map each loop method (loop_invariant_1) to the invariants to be deleted    
    map_method_to_inv = {}
    java_line_number = 1
    for line in java_file_lines:
        if java_line_number in map_line_to_name.keys():
            line = line[line.find("//@") + 3:]
            method = line[line.find("//") + 2:]
            method = method.strip()
            inv = map_line_to_name[java_line_number]
            if not method in map_method_to_inv:
                map_method_to_inv[method] = [inv]
            else:
                map_method_to_inv[method].append(inv)
        else:
            new_java_file.write(line)
        java_line_number = java_line_number + 1
    new_java_file.close()

    new_daikon_invariants_dict = dict()
    daikon_invariants_dict = read_invs_escj_plus_java(daikon_invs_filename)
    for loop_header in daikon_invariants_dict.keys():
        new_daikon_invariants_dict[loop_header] = set()
        for invariant in daikon_invariants_dict[loop_header]:
            remove = False
            if loop_header.loop_id in map_method_to_inv:
                for inv in map_method_to_inv[loop_header.loop_id]:
                    jml_part = invariant[0].split("::")[1].strip()
                    if inv.strip() == jml_part or inv[:-1].strip() == jml_part:
                        print("Removing loop invariant that did not verify: " + str(invariant[0]))
                        remove = True
                        break
                    
                        # pow function not supported by JML
            if "java.lang.Math.pow(" in invariant[0]:
                remove = True
            
            if not remove:
                new_daikon_invariants_dict[loop_header].add(invariant)
            else:
                remove = False
                
    write_daikon_invariants(new_daikon_invariants_dict, new_daikon_invs_filename)
    
    new_ginpink_invariants_dict = dict()
    if ginpink_inv_filename != None:
        ginpink_invariants_dict = read_ginpink_invariants(ginpink_inv_filename)
        for loop_header in ginpink_invariants_dict.keys():
            new_ginpink_invariants_dict[loop_header] = list()
            for invariant in ginpink_invariants_dict[loop_header]:
                remove = False
                if loop_header.loop_id in map_method_to_inv:
                    for inv in map_method_to_inv[loop_header.loop_id]:
                        jml_part = invariant.jml_string.strip()
                        if inv.strip() == jml_part or inv[:-1].strip() == jml_part:
                            print("Removing loop invariant that did not verify: " + str(jml_part))
                            remove = True
                            break
                        
                if not remove:
                    new_ginpink_invariants_dict[loop_header].append(invariant)
                else:
                    remove = False
                    
        write_ginpink_invariants(new_ginpink_invariants_dict, new_ginpink_inv_filename)

 
    return new_daikon_invariants_dict, new_ginpink_invariants_dict

