package org.ginpink;

import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.ginpink.eval.EvaluationException;
import org.ginpink.eval.ExprEvaluator;
import org.ginpink.eval.JavaHeap;
import org.ginpink.expr.JExpr;

public final class LoopInvariantMethod implements
		Comparable<LoopInvariantMethod> {

	private final Method reflect_method;

	public LoopInvariantMethod(Method method) {
		this.reflect_method = method;
	}

	public Method getReflectMethod() {
		return reflect_method;
	}

	private final List<Invariant> invariants = new LinkedList<Invariant>();

	private final List<Invariant> discarded_invariants = new LinkedList<Invariant>();

	public void addInvariant(Invariant invariant) {
		invariants.add(invariant);
	}

	public int checkInvariants(JavaHeap javaheap) {
		int remove_count = 0;
		ExprEvaluator evaluator = new ExprEvaluator(javaheap);
		ListIterator<Invariant> listIterator = this.invariants.listIterator();
		while (listIterator.hasNext()) {
			Invariant invariant = listIterator.next();
			JExpr expr = invariant.getJavaExpr();
			boolean removeInvariant;
			try {
				Boolean ret_val = (Boolean) expr.accept(evaluator);
				if (ret_val.booleanValue() == false) {
					if (GinPinkCheckerConfig.isDebugEnabled() == true) {
						String jmlString = invariant.getJmlString();
						String javaString = invariant.getJavaString();
						System.out.println("False evaluation: " + javaString
								+ "<exec|JML>" + jmlString);
						System.out.println("Instrumentation method: "
								+ reflect_method.getName());
						System.out.println("Associated program state "
								+ javaheap.toString());
					}
					removeInvariant = true;
				} else {
					removeInvariant = false;
				}
			} catch (EvaluationException ex) {
				removeInvariant = true;
				if (GinPinkCheckerConfig.isDebugEnabled() == true) {
					String jmlString = invariant.getJmlString();
					String javaString = invariant.getJavaString();
					System.out.println("Evaluation exception: " + javaString
							+ "<exec|JML>" + jmlString);
					System.out.println("Instrumentation method: "
							+ reflect_method.getName());
					System.out.println("Associated program state: "
							+ javaheap.toString());

				}
			}
			if (removeInvariant) {
				listIterator.remove();
				remove_count++;
				this.discarded_invariants.add(invariant);
			}

		}
		return remove_count;
	}

	public List<Invariant> getInvariants() {
		return invariants;
	}

	public String getMethodName() {
		return reflect_method.getName();
	}

	public String getClassName() {
		return reflect_method.getDeclaringClass().getCanonicalName();
	}

	@Override
	public int compareTo(LoopInvariantMethod arg0) {

		LoopInvariantMethod that = (LoopInvariantMethod) arg0;

		String this_class_and_method = this.getClassName() + "."
				+ this.getMethodName();

		String that_class_and_method = that.getClassName() + "."
				+ that.getMethodName();

		return this_class_and_method.compareTo(that_class_and_method);
	}

	public List<Invariant> getDiscardedInvariants() {
		return discarded_invariants;
	}

}
