package org.ginpink.expr;

import java.util.List;

public final class JMethodCall extends JExpr {

	private final JExpr expression;
	private final List<JExpr> arguments;

	public JMethodCall(JExpr expression, List<JExpr> arguments) {
		super();
		this.expression = expression;
		this.arguments = arguments;
	}
	
	@Override
	public <T> T accept(JExprVisitor<T> v) {
		return v.visit(this);
	}

	public JExpr getExpression() {
		return expression;
	}

	public List<JExpr> getArguments() {
		return arguments;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((arguments == null) ? 0 : arguments.hashCode());
		result = prime * result
				+ ((expression == null) ? 0 : expression.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		JMethodCall other = (JMethodCall) obj;
		if (arguments == null) {
			if (other.arguments != null)
				return false;
		} else if (!arguments.equals(other.arguments))
			return false;
		if (expression == null) {
			if (other.expression != null)
				return false;
		} else if (!expression.equals(other.expression))
			return false;
		return true;
	}

}
