package org.ginpink.expr;

import java.util.Arrays;
import java.util.LinkedList;

public final class JStaticMethodCall extends JExpr {

	private final String class_name;
	private final String method_name;
	private final LinkedList<JExpr> arguments;

	public JStaticMethodCall(String class_name, String method_name,
			JExpr... args) {
		this.method_name = method_name;
		this.class_name = class_name;
		this.arguments = new LinkedList<JExpr>(Arrays.asList(args));
	}

	@Override
	public <T> T accept(JExprVisitor<T> v) {
		return v.visit(this);
	}

	public String getMethodName() {
		return method_name;
	}

	public String getClassName() {
		return class_name;
	}

	public LinkedList<JExpr> getArguments() {
		return arguments;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((arguments == null) ? 0 : arguments.hashCode());
		result = prime * result
				+ ((class_name == null) ? 0 : class_name.hashCode());
		result = prime * result
				+ ((method_name == null) ? 0 : method_name.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		JStaticMethodCall other = (JStaticMethodCall) obj;
		if (arguments == null) {
			if (other.arguments != null)
				return false;
		} else if (!arguments.equals(other.arguments))
			return false;
		if (class_name == null) {
			if (other.class_name != null)
				return false;
		} else if (!class_name.equals(other.class_name))
			return false;
		if (method_name == null) {
			if (other.method_name != null)
				return false;
		} else if (!method_name.equals(other.method_name))
			return false;
		return true;
	}
}
